Package: tsitter
Title: Tree-Sitter Parsing Tools
Version: 0.2.0
Authors@R: c(
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com",
        role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
        comment = c(ROR = "03wc8by49")),
    person("Tree-sitter authors", role = "cph", comment = "Tree-sitter C library")
  )
Description: Common tree-sitter parsing tools for R. It is meant to be
    used by other packages that specialize in particular languages and
    file formats.
License: MIT + file LICENSE
Depends: R (>= 3.5.0)
Imports: cli, utils
Suggests: magrittr, pillar, testthat (>= 3.0.0), tsjsonc, tstoml, withr
Config/Needs/check: gaborcsardi/tsjsonc, gaborcsardi/tstoml
Config/Needs/coverage: gaborcsardi/tsjsonc, gaborcsardi/tstoml
Additional_repositories:
        https://github.com/r-lib/tsitter/releases/download
Encoding: UTF-8
URL: https://github.com/r-lib/tsitter, https://r-lib.github.io/tsitter/
BugReports: https://github.com/r-lib/tsitter/issues
Config/testthat/edition: 3
Config/Needs/website: r-lib/asciicast, tidyverse/tidytemplate,
        gaborcsardi/tsjsonc, gaborcsardi/tstoml
Biarch: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-07-29 23:22:10 UTC; gaborcsardi
Author: Gábor Csárdi [aut, cre],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>),
  Tree-sitter authors [cph] (Tree-sitter C library)
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-30 10:50:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-07-30 13:03:17 UTC; unix
Archs: tsitter.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("$",ts_tree)
S3method("[",ts_tree)
S3method("[[",ts_tree)
S3method("[[<-",ts_tree)
S3method("ts_tree_select<-",ts_tree)
S3method(as.character,ts_caller_arg)
S3method(as.character,ts_tree)
S3method(as.raw,ts_tree)
S3method(format,ts_parse_error)
S3method(format,ts_tree)
S3method(print,ts_parse_error)
S3method(print,ts_tree)
S3method(ts_tree_ast,default)
S3method(ts_tree_dom,default)
S3method(ts_tree_mark_selection1,ts_tree)
S3method(ts_tree_new,ts_language)
S3method(ts_tree_query,default)
S3method(ts_tree_select1,default)
S3method(ts_tree_select1,ts_tree.NULL)
S3method(ts_tree_select1,ts_tree.character)
S3method(ts_tree_select1,ts_tree.integer)
S3method(ts_tree_select1,ts_tree.logical)
S3method(ts_tree_select1,ts_tree.numeric)
S3method(ts_tree_select1,ts_tree.ts_tree_selector_default)
S3method(ts_tree_select1,ts_tree.ts_tree_selector_ids)
S3method(ts_tree_select1,ts_tree.ts_tree_selector_regex)
S3method(ts_tree_select1,ts_tree.ts_tree_selector_tsquery)
S3method(ts_tree_sexpr,default)
S3method(ts_tree_write,default)
export("ts_tree_select<-")
export(as_ts_caller_arg)
export(ts_caller_arg)
export(ts_caller_env)
export(ts_check_named_arg)
export(ts_cnd)
export(ts_collapse)
export(ts_list_parsers)
export(ts_parse_error_cnd)
export(ts_tree_ast)
export(ts_tree_delete)
export(ts_tree_deleted)
export(ts_tree_dom)
export(ts_tree_format)
export(ts_tree_insert)
export(ts_tree_mark_selection1)
export(ts_tree_new)
export(ts_tree_query)
export(ts_tree_select)
export(ts_tree_select1)
export(ts_tree_selected_nodes)
export(ts_tree_selection)
export(ts_tree_sexpr)
export(ts_tree_unserialize)
export(ts_tree_update)
export(ts_tree_write)
useDynLib(tsitter, .registration = TRUE, .fixes = "c_")

<!-- README.md is generated from README.Rmd. Please edit that file -->

# tsitter

<!-- badges: start -->

![lifecycle](https://lifecycle.r-lib.org/articles/figures/lifecycle-experimental.svg)
[![R-CMD-check](https://github.com/r-lib/tsitter/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/tsitter/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

Common tree-sitter parsing tools for R.

tsitter is a common interface to
[tree-sitter](https://tree-sitter.github.io/tree-sitter/) parsers,
implemented in other R packages. It has a common API to

- query,
- edit,
- format, and
- unserialize

tree-sitter parse trees.

## Installation

Install tsitter from CRAN:

``` r
install.packages("tsitter")
```

## Documentation

See at
[`https://r-lib.github.io/tsitter/`](https://r-lib.github.io/tsitter/reference/index.html)
and also in the installed package: `help(package = "tsitter")`.

## Parsers that use tsitter

- JSONC (JSON with comments and trailing commas):
  [tsjsonc](https://gaborcsardi.github.io/tsjsonc/).
- TOML: [tstoml](https://gaborcsardi.github.io/tstoml/).

## Quickstart

In this document I show examples with the `tsjsonc` package.

### Create a tree-sitter tree

Create a ts_tree (ts_tree_jsonc) object from a string:

``` r
txt <- r"(
// this is a comment
{
  "a": {
    "a1": [1, 2, 3],
    // comment
    "a2": "string"
  },
  "b": [
    {
      "b11": true,
      "b12": false
    },
    {
      "b21": false,
      "b22": false
    }
  ]
}
)"

json <- tsjsonc::ts_parse_jsonc(txt)
```

Pretty print a ts_tree object:

``` r
json
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/print-json-dark.svg">
<img src="man/figures/print-json.svg" /> </picture>

### Select nodes of a tree

Selecting nodes is the basis of editing and querying tree-sitter trees.

Select element by objects key:

``` r
ts_tree_select(json, "a")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/select-key-dark.svg">
<img src="man/figures/select-key.svg" /> </picture>

Select element inside element:

``` r
ts_tree_select(json, "a", "a1")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/select-select-dark.svg">
<img src="man/figures/select-select.svg" /> </picture>

Select element(s) of an array:

``` r
ts_tree_select(json, "a", "a1", 1:2)
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/select-array-dark.svg">
<img src="man/figures/select-array.svg" /> </picture>

Select multiple keys from an object:

``` r
ts_tree_select(json, "a", c("a1", "a2"))
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/select-multiple-dark.svg">
<img src="man/figures/select-multiple.svg" /> </picture>

Select nodes that match a tree-sitter query:

``` r
json |> ts_tree_select(query = "((pair value: (false) @val))")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/select-query-dark.svg">
<img src="man/figures/select-query.svg" /> </picture>

### Delete elements

Delete selected elements:

``` r
ts_tree_select(json, "a", "a1") |> ts_tree_delete()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/delete-dark.svg">
<img src="man/figures/delete.svg" /> </picture>

### Insert elements

Insert element into an array:

``` r
ts_tree_select(json, "a", "a1") |> ts_tree_insert(at = 2, "new")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/insert-array-dark.svg">
<img src="man/figures/insert-array.svg" /> </picture>

Inserting into an array reformats the array.

Insert element into an object, at the specified key:

``` r
ts_tree_select(json, "a") |>
  ts_tree_insert(key = "a0", at = 0, list("new", "element"))
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/insert-object-dark.svg">
<img src="man/figures/insert-object.svg" /> </picture>

### Update elements

Update existing element:

``` r
ts_tree_select(json, "a", c("a1", "a2")) |> ts_tree_update("new value")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/update-dark.svg">
<img src="man/figures/update.svg" /> </picture>

Inserts the element if some parents are missing:

``` r
json <- ts_parse_jsonc(text = "{ \"a\": { \"b\": true } }")
json
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/update-insert-dark.svg">
<img src="man/figures/update-insert.svg" /> </picture>

``` r
ts_tree_select(json, "a", "x", "y") |> ts_tree_update(list(1,2,3))
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/update-insert-2-dark.svg">
<img src="man/figures/update-insert-2.svg" /> </picture>

### Write out a document

Use `stdout()` to write it to the screen instread of a file:

``` r
json |> ts_tree_write(stdout())
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/write-dark.svg">
<img src="man/figures/write.svg" /> </picture>

### Formatting

Format the whole document:

``` r
json |> ts_tree_format()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/format-dark.svg">
<img src="man/figures/format.svg" /> </picture>

Format part of the document:

``` r
json |> ts_tree_select("a") |>
  ts_tree_format(options = list(format = "compact"))
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/format-part-dark.svg">
<img src="man/figures/format-part.svg" /> </picture>

### Unserializing

Unserialize a whole document:

``` r
json |> ts_tree_unserialize()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/unserialize-dark.svg">
<img src="man/figures/unserialize.svg" /> </picture>

Note that `ts_tree_unserialize()` always returns a list, the first
element of the list is the unserialized document.

Unserialize part(s) of the document:

``` r
json |> ts_tree_select("b") |> ts_tree_unserialize()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/unserialize-part-dark.svg">
<img src="man/figures/unserialize-part.svg" /> </picture>

Again, `ts_tree_unserialize()` returns a list, with one element for each
selected node.

### Exploring a tree-sitter tree

It is often useful to explore the structure of a (JSONC) tree-sitter
tree, to help writing the right selection or tree-sitter queries.

Print the annotated syntax tree:

``` r
ts_tree_ast(json)
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/syntax-tree-dark.svg">
<img src="man/figures/syntax-tree.svg" /> </picture>

Print the document object model:

``` r
ts_tree_dom(json)
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/document-model-dark.svg">
<img src="man/figures/document-model.svg" /> </picture>

Print the structural summary of a tree:

``` r
ts_tree_sexpr(json)
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/structure-dark.svg">
<img src="man/figures/structure.svg" /> </picture>

# License

MIT © Posit Software, PBC

#' Insert a new element into a tree-sitter tree
#'
#' @description
#' Insert a new element into each selected element.
#'
#' @details
#' It is not always possible to insert a new element into a selected
#' element. For example in a JSONC document you can only insert a new
#' element into an array or an object, but not into scalar elements.
#' If the insertion is not possible, an error is raised.
#'
#' If `tree` does not have a selection, the new element is inserted into
#' at the top level.
#'
#' If `tree` has an empty selection, then it is returned unchanged, i.e.
#' no new element is inserted.
#'
#' @param tree
#' A `ts_tree` object.
#' @param new
#' The new element to insert.
#'
#' The type of `new` depends on the parser and the method that implements
#' the insertion. See details in the manual of the specific parser.
#' @param key
#' The key of the new element, if inserting into a keyed element.
#'
#' For example a JSON(C) object or a TOML table are keyed elements.
#' @param at
#' The position to insert the new element at.
#'
#' The interpretation of this argument depends on the method that
#' implements the insertion. Typically the followings are supported:
#' - `0` inserts at the beginning.
#' - `Inf` inserts at the end.
#' - A positive integer `n` inserts _after_ the `n`-th element.
#' - A character scalar inserts _after_ the element with the given key,
#'   in keyed elements.
#'
#' See the details in the manual of the specific parser.
#' @param options
#' A list of options for the insertion.
#'
#' See details in the manual of the specific parser.
#' @param ... Extra arguments for methods.
#'
#' @return
#' A `ts_tree` object representing the modified parse tree.
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{ "a": true, "b": [1, 2, 3] }')
#'
#' ts_tree_insert(ts_tree_select(tree, "b"), 4, at = Inf)
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#'
#' # Create a parse tree with tstoml --------------------------------------
#' tree <- tstoml::ts_parse_toml(r"(
#'   [servers]
#'   alpha = { ip = "127.0.0.1", dc = "eqdc10" }
#'   beta = { ip = "127.0.0.2", dc = "eqdc20" }
#' )")
#'
#' ts_tree_insert(
#'   ts_tree_select(tree, "servers", TRUE),
#'   key = "active",
#'   TRUE
#' )

ts_tree_insert <- function(tree, new, key, at, options, ...) {
  UseMethod("ts_tree_insert")
}

test_that("ts_tree_insert", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
    tree %>% ts_tree_insert(key = "x", list(1, 2, 3)) %>% print(n = 100)
    tree %>% ts_tree_select("b") %>% ts_tree_insert("boo!", at = 1)
  })
})

#' Show the annotated syntax tree of a tree-sitter tree
#'
#' @description
#' `ts_tree_ast()` prints the annotated syntax tree of a ts_tree object.
#' This syntax tree contains all tree-sitter nodes, and it shows the
#' source code associated with each node, along with line numbers.
#'
#' @details
#' ## The syntax tree and the DOM tree
#'
#' This syntax tree contains all nodes of the tree-sitter parse tree,
#' including both named and unnamed nodes and comments. E.g. for a JSON(C)
#' document it includes the pairs, brackets, braces, commas, colons,
#' double quotes and string escape sequences as separate nodes.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' See \code{\link[tsitter:ts_tree_dom]{tsitter::ts_tree_dom()}} for a tree that
#' shows the semantic structure of the parsed document, which may be
#' different from the syntax tree.
#'
#' @param tree
#' A `ts_tree` object.
#'
#' @return
#' Character vector, the formatted annotated syntax tree, line by
#' line. It has class [cli_tree][cli::tree()], from the cli package. It
#' may contain ANSI escape sequences for coloring and hyperlinks.
#'
#' @family ts_tree exploration
#' @family ts_tree generics
#' @seealso [ts_tree_dom()] to show the document object model (DOM) of a
#'   ts_tree object.
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')
#'
#' tree
#'
#' ts_tree_ast(tree)
#'
#' ts_tree_dom(tree)
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#'
#' # Create a parse tree with tstoml --------------------------------------
#' tree <- tstoml::ts_parse_toml(r"(
#'   title = "TOML Example"
#'   [owner]
#'   name = "Tom Preston-Werner"
#'   dob = 1979-05-27T07:32:00-08:00
#' )")
#'
#' tree
#'
#' ts_tree_ast(tree)
#'
#' ts_tree_dom(tree)

ts_tree_ast <- function(tree) {
  UseMethod("ts_tree_ast")
}

#' @export

ts_tree_ast.default <- function(tree) {
  type <- tree$type
  fn <- attr(tree, "file")
  if (cli::ansi_has_hyperlink_support() && !is.null(fn)) {
    type <- cli::style_hyperlink(
      type,
      sprintf(
        "file://%s:%d:%d",
        normalizePath(fn, mustWork = NA),
        tree$start_row + 1L,
        tree$start_column + 1
      )
    )
  }

  linum <- tree$start_row + 1
  linum <- ifelse(duplicated(linum), "", as.character(linum))
  linum <- format(linum, justify = "right")
  # this is the spacer we need to put in for multi-line tokens
  nlspc <- paste0("\n\t", strrep(" ", nchar(linum[1])), "|")
  code <- ifelse(
    is.na(tree$code),
    "",
    paste0(strrep(" ", tree$start_column), tree$code)
  )

  # we put in a \t, and later use it to align the lines vertically
  treetab <- data_frame(
    id = as.character(tree$id),
    children = lapply(tree$children, as.character),
    label = paste0(
      type,
      " (",
      tree$id,
      ")",
      "\t",
      linum,
      "|",
      gsub("\n", nlspc, code, fixed = TRUE)
    )
  )
  tree <- cli::tree(treetab)

  # align lines vertically. the size of the alignment is measured
  # without the ANSI sequences, but then the substitution uses the
  # full ANSI string
  tabpos <- regexpr("\t", cli::ansi_strip(tree), fixed = TRUE)
  maxtab <- max(tabpos)
  tabpos2 <- regexpr("\t", tree, fixed = TRUE)
  regmatches(tree, tabpos2) <- strrep(" ", maxtab - tabpos + 4)

  tree
}

test_that("ts_tree_ast", {
  skip_if_not_installed("tsjsonc")
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
    ts_tree_ast(tree)
  })
})

test_that("ts_tree_ast with hyperlinks", {
  skip_if_not_installed("tsjsonc")
  withr::local_options(cli.hyperlink = TRUE)
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  expect_snapshot(
    {
      writeLines('{"a": 1, "b": [1,2,3]}', con = tmp)
      tree <- tsjsonc::ts_read_jsonc(tmp)
      ts_tree_ast(tree)
    },
    transform = redact_tempfile
  )
})

test_that("get_ranges", {
  skip_if_not_installed("tsjsonc")
  withr::local_seed(13)
  randomstr <- function() {
    l <- as.integer(runif(1) * 3)
    strrep("x", l)
  }
  json <- "[1,2,3]"
  txttab <- rbind(strsplit(json, "")[[1]], replicate(nchar(json), randomstr()))
  txt <- paste(txttab, collapse = "")
  start_col <- cumsum(nchar(txttab))[seq(1, length(txttab), by = 2)]
  end_col <- start_col + nchar(txttab[1, ]) - 1L
  ranges <- data.frame(
    stringsAsFactors = FALSE,
    start_row = rep(1L, ncol(txttab)),
    start_col = start_col,
    end_row = rep(1L, ncol(txttab)),
    end_col = end_col,
    start_byte = start_col,
    end_byte = end_col
  )

  expect_snapshot({
    txt
    tree <- ts_tree_new(
      tsjsonc::ts_language_jsonc(),
      text = txt,
      ranges = ranges
    )
    tree
    ts_tree_ast(tree)
  })

  # invalid ranges
  expect_snapshot(
    error = TRUE,
    {
      ts_tree_new(tsjsonc::ts_language_jsonc(), text = txt, ranges = 1:10)
      ranges2 <- data.frame(a = 0:1, b = 0:1, c = 0:1, d = 0:1, e = 1, f = 0:1)
      ts_tree_new(tsjsonc::ts_language_jsonc(), text = txt, ranges = ranges2)
      ranges3 <- ranges
      ranges3$start_byte[1] <- 100L
      ts_tree_new(tsjsonc::ts_language_jsonc(), text = txt, ranges = ranges3)
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("new_lookahead_sym", {
  skip_if_not_installed("tsjsonc")
  expect_snapshot({
    tree <- ts_tree_new(
      tsjsonc::ts_language_jsonc(),
      text = '[{"a": }]',
      fail_on_parse_error = FALSE
    )
    tree[]
    tree$is_missing
    exp <- tree$expected[[which(tree$is_missing)]]
    do.call(rbind, exp)
  })
})

test_that("ts_xtree_free", {
  skip_if_not_installed("tsjsonc")
  ts_tree_new(tsjsonc::ts_language_jsonc(), text = '{"a": 1}')
  gc()
  gc()
  expect_true(TRUE)
})

test_that("parse", {
  expect_snapshot(
    error = TRUE,
    {
      ts_tree_new(
        structure("jsonc", class = c("ts_language_jsonc", "ts_language")),
        text = '{"a": '
      )
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("check_predicate_eq", {
  skip_if_not_installed("tsjsonc")
  withr::local_options(width = 1000)

  # compare to string
  tree <- ts_tree_new(
    tsjsonc::ts_language_jsonc(),
    text = '{"a": 1, "b": 2}'
  )
  expect_snapshot({
    ts_tree_query(
      tree,
      "((pair key: (string (string_content) @key)) (#eq? @key \"a\"))"
    )
  })
  expect_snapshot({
    ts_tree_query(
      tree,
      "((pair key: (string (string_content) @key)) (#not-eq? @key \"a\"))"
    )
  })

  # compare two captures
  tree <- ts_tree_new(
    tsjsonc::ts_language_jsonc(),
    text = '{"a": "a", "b": "x"}'
  )
  expect_snapshot({
    ts_tree_query(
      tree,
      "((pair key: (string (string_content) @key)
        value: (string (string_content) @value)) @pair
        (#eq? @key @value))"
    )
  })
  expect_snapshot({
    ts_tree_query(
      tree,
      "((pair key: (string (string_content) @key)
        value: (string (string_content) @value))
        (#not-eq? @key @value))"
    )
  })
  # any-eq
  # any-not-eq
})

test_that("r_grepl", {
  skip_if_not_installed("tsjsonc")
  withr::local_options(width = 1000)

  tree <- ts_tree_new(
    tsjsonc::ts_language_jsonc(),
    text = '{"a": 1, "b": 2}'
  )
  expect_snapshot({
    ts_tree_query(
      tree,
      "((pair key: (string (string_content) @key))
        (#match? @key \"^a\"))"
    )
  })
})

test_that("check_predicate_match", {
  #
})

test_that("check_predicate_any_of", {
  #
})

#' Utility functions for ts language implementations (internal)
#'
#' These functions are for packages implementing new parsers based on the ts
#' package. It is very unlikely that you will need to call these functions
#' directly.
#'
#' @param ... Arguments collapsed and interpolated into a condition message.
#' @param class A character vector of classes for the condition.
#' @param call Environment of the call to associate with the condition,
#'   defaults to the caller's environment.
#' @param .envir The environment in which to evaluate the `...` expressions,
#'   defaults to the parent frame.
#' @return `ts_cnd()` returns an error condition object.
#'
#' @details `ts_cnd()` creates a condition object. It interpolates its
#' arguments into a single message.
#'
#' @keywords internal
#' @rdname internal
#' @export

ts_cnd <- function(
  ...,
  class = NULL,
  call = ts_caller_env(),
  .envir = parent.frame()
) {
  call <- frame_get(call, sys.call)
  structure(
    list(message = glue(..., .envir = .envir), call = call),
    class = c(class, "error", "condition")
  )
}

#' @rdname internal
#' @param arg Argument name.
#' @details `ts_caller_arg()` captures the expression used as an argument
#'   to a function, for use in error messages.
#' @return `ts_caller_arg()` returns the captured expression as a
#'   ts_caller_arg object.
#' @export

ts_caller_arg <- function(arg) {
  arg <- substitute(arg)
  expr <- do.call(substitute, list(arg), envir = ts_caller_env())
  structure(list(expr), class = "ts_caller_arg")
}

#' @rdname internal
#' @param x Object to use as a caller argument.
#' @details `as_ts_caller_arg()` converts an object into a caller argument
#'   object. This is useful when referring to parts of the caller argument
#'   in downstream error messages.
#' @return `as_ts_caller_arg()` returns a ts_caller_arg object.
#' @export

as_ts_caller_arg <- function(x) {
  structure(list(x), class = "ts_caller_arg")
}

#' @rdname internal
#' @param x A ts_caller_arg object.
#' @details `as.character.ts_caller_arg()` formats a caller argument
#'   object as a short string for use in error messages. Multi-line
#'   expressions are truncated after the first line.
#' @return `as.character.ts_caller_arg()` returns a short string
#'   representation of the caller argument, a character scalar.
#' @export

as.character.ts_caller_arg <- function(x, ...) {
  lbl <- paste(format(x[[1]]), collapse = "\n")
  gsub("\n.*$", "...", lbl)
}

# older R does not have a format method for names
format.name <- function(x, ...) {
  deparse(x, backtick = FALSE)
}

#' @rdname internal
#' @param n Number of frames to go up to find the caller environment.
#' @details `ts_caller_env()` returns the environment of the caller function,
#'   `n` levels up the call stack. This is useful for associating error
#'   conditions with the correct call.
#' @return `ts_caller_env()` returns an environment, or `NULL` is called
#'   from the global environment.
#' @export

ts_caller_env <- function(n = 1) {
  parent.frame(n + 1)
}

frame_get <- function(frame, accessor) {
  if (identical(frame, .GlobalEnv)) {
    return(NULL)
  }
  frames <- evalq(sys.frames(), frame)
  for (i in seq_along(frames)) {
    if (identical(frames[[i]], frame)) {
      return(accessor(i))
    }
  }
  NULL
}

#' @rdname internal
#' @param arg Argument to check.
#' @param frame Frame number to inspect, defaults to the caller's frame.
#' @details `ts_check_named_arg()` checks whether an argument was supplied
#'   with a name. If not, it raises an error.
#' @return `ts_check_named_arg()` returns `TRUE` invisibly if the argument
#'   was named, otherwise it raises an error.
#' @export

ts_check_named_arg <- function(arg, frame = -1) {
  arg <- as.character(substitute(arg))
  if (!arg %in% names(sys.call(frame)[-1])) {
    stop(ts_cnd(
      paste0("The `", arg, "` argument must be fully named."),
      call = ts_caller_env()
    ))
  }
  invisible(TRUE)
}

#' @rdname internal
#' @param tree A `ts_tree` object as returned by [ts_tree_new()].
#' @param text Raw vector, the original text used to parse the tree.
#' @param call Environment of the call to associate with the error
#'   condition, defaults to the caller's environment.
#' @details `ts_parse_error_cnd()` creates a parse error condition
#'   associated with a `ts_tree` object and the original text.
#'   The error message includes information about the location of
#'   parse errors in the text. It also has `format()` and `print()`
#'   methods to display the error together with the relevant lines
#'   of the original text.
#' @return `ts_parse_error_cnd()` returns a ts_parse_error condition
#' @export

ts_parse_error_cnd <- function(tree, text, call = ts_caller_env()) {
  call <- frame_get(call, sys.call)
  cnd <- structure(
    list(
      call = call,
      tree = tree,
      text = strsplit(rawToChar(text), "\r?\n")[[1]]
    ),
    class = c("ts_parse_error", "error", "condition")
  )
  cnd$message <- paste0(
    format_ts_parse_error_(cnd),
    collapse = "\n"
  )
  cnd
}

# TODO: taking the first row of the error is not always ideal, because
# ERROR nodes may contain a leading comma before the element that has the
# actual error, and if there is a line break after the comma, then we
# mark the comma as the place of the error, instead of the real error in
# next line.
#
# One workaround is to mark all lines of the ERROR node.
# Another is to ignore the first child of the ERROR node if it is a comma,
# but I am unsure if that's always correct. Sometimes that error may be
# at the comma. Plus maybe we'd need to ignore more than just a first comma.
# So maybe we should mark all rows of the ERROR node.

format_ts_parse_error_ <- function(x, n = 2, ...) {
  file <- attr(x$tree, "file") %||% "<text>"
  nodes <- utils::head(which(x$tree$type == "ERROR" | x$tree$is_missing), n)
  # if two errors start at the same position only show the first
  nodes <- nodes[!duplicated(x$tree$start_byte[nodes])]

  rows <- x$tree$start_row[nodes] + 1L
  cols <- x$tree$start_column[nodes] + 1L
  erows <- x$tree$end_row[nodes] + 1L

  ecols <- ifelse(
    rows == erows,
    x$tree$end_column[nodes] + 1L,
    nchar(x$text[rows])
  )

  lang <- toupper(get_tree_lang(x$tree))
  unlist(mapply(
    format_ts_parse_error_1,
    lang,
    file,
    rows,
    cols,
    ecols,
    MoreArgs = list(text = x$text)
  ))
}

format_ts_parse_error_1 <- function(lang, text, file, row, col, ecol) {
  head <- sprintf("%s parse error `%s`:%d:%d", toupper(lang), file, row, col)

  nchar <- function(x) {
    cli::ansi_nchar(x, type = "width")
  }

  rows <- c(if (row > 1L) row - 1L, row, if (row < length(text)) row + 1L)
  linum <- structure(cli::col_grey(sprintf("%d| ", rows)), names = rows)
  nlinum <- max(nchar(linum))

  # assume a minimum width of 40, otherwise it is hard to format
  width <- max(cli::console_width(), 40L) - nlinum

  prefix <- substr(text[row], 1L, col - 1L)
  error <- substr(text[row], col, ecol)
  suffix <- substr(text[row], ecol + 1L, nchar(text[row]))

  npre <- nchar(prefix)
  nerr <- nchar(error)
  nsuf <- nchar(suffix)
  truncated <- FALSE
  if (npre + nerr + nsuf > width) {
    truncated <- TRUE
    dots <- cli::col_grey("...")
    # need to truncate, first suffix
    if (nsuf > 10) {
      suffix <- paste0(dots, substr(suffix, nchar(suffix) - 7, nchar(suffix)))
      nsuf <- nchar(suffix)
    }
    if (npre + nerr + nsuf > width) {
      # need to truncate, then prefix
      if (npre > 10) {
        prefix <- paste0(substr(prefix, 1, 7), dots)
        npre <- nchar(prefix)
      }
      if (npre + nerr + nsuf > width) {
        # need to truncate error
        err_width <- width - npre - nsuf
        error <- paste0(
          substr(error, 1, err_width %/% 2),
          dots,
          substr(error, nchar(error) - err_width %/% 2 + 4, nchar(error))
        )
      }
    }
  }

  # if we truncated the line, then do not show the context lines
  ctx_before <- ctx_after <- NULL
  if (truncated) {
    linum <- linum[as.character(row)]
  } else {
    ctx_before <- if (row > 1) {
      paste0(linum[as.character(row - 1)], text[row - 1])
    }
    ctx_after <- if (row < length(text)) {
      paste0(linum[as.character(row + 1)], text[row + 1])
    }
  }

  mark <- paste0(
    strrep(" ", nlinum + nchar(prefix)),
    cli::col_red(strrep("^", max(nchar(error), 1)))
  )

  c(
    head,
    ctx_before,
    paste0(linum[as.character(row)], prefix, error, suffix),
    mark,
    ctx_after
  )
}

#' @export

format.ts_parse_error <- function(x, ...) {
  c("<ts_parse_error>", format_ts_parse_error_(x, ...))
}

#' @export

print.ts_parse_error <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

test_that("cnd", {
  v <- 13
  expect_snapshot({
    do <- function() ts_cnd("This is a test error with a value: {v}.")
    do()
  })
})

test_that("ts_caller_arg", {
  do <- function(x) {
    ts_caller_arg(x)
  }
  v <- 13
  expect_snapshot({
    do(1)
    do(v)
  })
})

test_that("as_ts_caller_arg", {
  expect_snapshot(as_ts_caller_arg("foobar"))
})

test_that("as.character.ts_caller_arg", {
  do <- function(x) {
    as.character(ts_caller_arg(x))
  }
  v1 <- 13
  expect_snapshot({
    do(v1)
    do(1 + 1 + 1)
    do(
      foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo +
        foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo +
        foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo
    )
  })
})

test_that("frame_get", {
  expect_null(frame_get(.GlobalEnv, sys.call))
  fake(frame_get, "identical", FALSE)
  expect_null(frame_get(environment(), sys.call()))
})

test_that("ts_check_named_arg", {
  f <- function(foobar = NULL) {
    if (!missing(foobar)) {
      ts_check_named_arg(foobar)
    }
  }
  expect_silent(f(foobar = 42))
  expect_silent(f())

  expect_snapshot(
    error = TRUE,
    {
      f(42)
      f(fooba = 42)
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("ts_parse_error_cnd", {
  skip_if_not_installed("tsjsonc")
  txt <- '{"a": 1, "b": [true, false, error]}'
  tree <- tsjsonc::ts_parse_jsonc(txt, fail_on_parse_error = FALSE)
  expect_snapshot(
    error = TRUE,
    {
      stop(ts_parse_error_cnd(tree = tree, text = charToRaw(txt)))
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("format_ts_parse_error_1", {
  skip_if_not_installed("tsjsonc")
  withr::local_options(cli.width = 40)
  txt <- '{"a": 1, "b": [true, false, "very", error, "this", "is", "ok"]}'
  tree <- tsjsonc::ts_parse_jsonc(txt, fail_on_parse_error = FALSE)
  err <- ts_parse_error_cnd(tree = tree, text = charToRaw(txt))
  expect_snapshot(
    error = TRUE,
    {
      stop(err)
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )

  # truncate error
  txt2 <- '{"a": 1, "b": [true, false, "very", looooooooooooooooooooooong]}'
  tree2 <- tsjsonc::ts_parse_jsonc(txt2, fail_on_parse_error = FALSE)
  err2 <- ts_parse_error_cnd(tree = tree2, text = charToRaw(txt2))
  expect_snapshot(
    error = TRUE,
    {
      stop(err2)
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )

  # show context
  txt3 <- '{"a": 1, \n "b": [true, false, "very", bad]\n}'
  tree3 <- tsjsonc::ts_parse_jsonc(txt3, fail_on_parse_error = FALSE)
  err3 <- ts_parse_error_cnd(tree = tree3, text = charToRaw(txt3))
  expect_snapshot(
    error = TRUE,
    {
      stop(err3)
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )

  # omit context on truncation
  txt4 <- '{"a": 1, \n  "b": [true, false, "very", looooooooooooooooooooooong]\n}'
  tree4 <- tsjsonc::ts_parse_jsonc(txt4, fail_on_parse_error = FALSE)
  err4 <- ts_parse_error_cnd(tree = tree4, text = charToRaw(txt4))
  expect_snapshot(
    error = TRUE,
    {
      stop(err4)
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("format.ts_parse_error, print.ts_parse_error", {
  skip_if_not_installed("tsjsonc")
  txt <- '{"a": 1, "b": [true, false, error]}'
  tree <- tsjsonc::ts_parse_jsonc(txt, fail_on_parse_error = FALSE)
  err <- ts_parse_error_cnd(tree = tree, text = charToRaw(txt))
  expect_snapshot({
    format(err)
    print(err)
  })
})

#' List installed tree-sitter parsers
#'
#' The ts package contains a common interface to several tree-sitter
#' parsers, implemented in other R packages. `ts_list_parsers()` lists
#' the available parsers installed in the system.
#'
#' To see tree-sitter parser packages that are available on CRAN, but not
#' installed on your system, see the packages that depend on ts and have a
#' name with a 'ts' prefix.
# TODO: add URLs here, or pkgsearch  example once ts is on CRAN.
#'
# NOTE: this only works if all suggested packages are installed.
#' Here is an example that includes all tree-sitter parsers at this time:
#'
#' ```{asciicast}
#' ts_list_parsers()
#' ```
#'
#' @param lib_path Library paths to search for installed packages.
#'   Default is [base::.libPaths()].
#' @return A data frame with columns:
#' * `package`: character, the name of the package.
#' * `version`: character, the version of the package.
#' * `title`: character, the title of the package.
#' * `library`: character, the library path where the package is installed.
#' * `loaded`: logical, whether the package is currently loaded.
#' @export
#' @examples
#' ts_list_parsers()

ts_list_parsers <- function(lib_path = .libPaths()) {
  # installed packages
  ipkgs <- list_installed_packages(lib_path)
  ipkgs <- ipkgs[grepl("^ts.", basename(ipkgs))]
  dscs1 <- lapply(ipkgs, function(pkg) {
    suppressWarnings(utils::packageDescription(
      basename(pkg),
      lib.loc = dirname(pkg)
    ))
  })

  # loaded packages
  lpkgs <- loadedNamespaces()
  lpkgs <- lpkgs[grepl("^ts.", lpkgs)]
  dscs2 <- lapply(lpkgs, function(pkg) {
    suppressWarnings(utils::packageDescription(pkg))
  })

  dscs <- Filter(has_ts_parser, c(dscs1, dscs2))
  path <- map_chr(dscs, attr, "file", exact = TRUE)
  dscs <- dscs[!duplicated(path)]
  path <- path[!duplicated(path)]

  pkgpath <- as.character(ifelse(
    basename(path) == "DESCRIPTION",
    dirname(path),
    dirname(dirname(path))
  ))

  tspkgs <- data_frame(
    package = map_chr(dscs, "[[", "Package"),
    version = map_chr(dscs, "[[", "Version"),
    title = map_chr(dscs, "[[", "Title"),
    library = dirname(pkgpath)
  )
  nspath <- map_chr(tspkgs$package, function(pkg) {
    if (pkg %in% loadedNamespaces()) {
      getNamespaceInfo(pkg, "path")
    } else {
      NA_character_
    }
  })
  tspkgs$loaded <- tspkgs$package %in% loadedNamespaces() & nspath == pkgpath

  tspkgs <- tspkgs[order(tspkgs$package, !tspkgs$loaded, tspkgs$version), ]

  class(tspkgs) <- c(
    "ts_parser_list",
    class(tspkgs)
  )
  tspkgs
}

format_rd_parser_list <- function(lst, method = NULL) {
  if (is.null(method)) {
    format_rd_parser_list_no_method(lst)
  } else {
    format_rd_parser_list_method(lst, method)
  }
}

no_ts_package_message <- function() {
  avail <- read.dcf(file.path(doc_path("tsitter"), "ts-packages.dcf"))
  paste0(
    "No tree-sitter parser packages are installed.\n",
    "Available tree-sitter parser packages:\n",
    "\\itemize{\n",
    paste0(
      collapse = "\n",
      "\\item \\strong{\\href{",
      avail[, "URL"],
      "}{",
      avail[, "Package"],
      "}}: ",
      avail[, 'Title'],
      "\\if{text}{ (",
      avail[, "URL"],
      ")}."
    ),
    "\n}\n"
  )
}

format_rd_parser_list_no_method <- function(lst) {
  if (nrow(lst) == 0) {
    return(no_ts_package_message())
  }
  lst <- lst[!duplicated(lst[, c("package", "version")]), ]

  hd <- "Available tree-sitter parsers"
  align <- "lll"
  loaded_hd <- ""
  loaded <- rep("", nrow(lst))
  if (Sys.getenv("IN_PKGDOWN") != "true") {
    hd <- "Installed tree-sitter parsers"
    align <- "llcl"
    loaded_hd <- "\\strong{Loaded} \\tab "
    loaded <- paste0(ifelse(lst$loaded, "yes", "no"), " \\tab ")
  }

  lines <- map_chr(
    seq_len(nrow(lst)),
    function(i) {
      pkg <- lst$package[i]
      ver <- lst$version[i]
      title <- lst$title[i]
      glue(
        "\\link[{pkg}:{pkg}-package]{{ {pkg}} \\tab \\
         {ver} \\tab \\
         {loaded[i]} \\
         {title}.
         "
      )
    }
  )

  paste0(
    "\\subsection{",
    hd,
    "}{\n",
    "\\tabular{",
    align,
    "}{\n",
    "\\strong{Package} \\tab \\strong{Version} \\tab ",
    loaded_hd,
    "\\strong{Title} \\cr\n",
    paste(lines, collapse = "\\cr\n"),
    "\n}\n",
    "}\n"
  )
}

format_rd_parser_list_method <- function(lst, method) {
  if (nrow(lst) == 0) {
    return(no_ts_package_message())
  }
  lst <- lst[!duplicated(lst[, c("package", "version")]), ]

  hd <- "Available tree-sitter parsers"
  loaded_hd <- ""
  loaded <- rep("", nrow(lst))
  align <- "llll"
  if (Sys.getenv("IN_PKGDOWN") != "true") {
    hd <- "Installed tree-sitter parsers"
    align <- "llcll"
    loaded_hd <- "\\strong{Loaded} \\tab "
    loaded <- paste0(ifelse(lst$loaded, "yes", "no"), " \\tab ")
  }

  lines <- map_chr(
    seq_len(nrow(lst)),
    function(i) {
      pkg <- lst$package[i]
      ver <- lst$version[i]
      title <- lst$title[i]
      method <- if (doc_has_method(method, pkg)) {
        cls <- sub("^ts", "ts_tree_", pkg)
        glue(
          "\\code{{\\link[{pkg}:{method}.{cls}]{{{method}(<ts_tree_{pkg}>)}}}}"
        )
      }
      glue(
        "\\strong{{\\link[{pkg}:{pkg}-package]{{{pkg}}}}} \\tab \\
         {ver} \\tab \\
         {loaded[i]} \\
         {title}. \\tab \\
         {method}"
      )
    }
  )
  paste0(
    "\\subsection{",
    hd,
    "}{\n",
    "This is the manual page of the \\code{",
    method,
    "()} S3 generic function.\n",
    "Methods in parser packages may override this generic.\n",
    "For the ones that do see the links to their manual pages in the table.\n",
    "\\tabular{",
    align,
    "}{\n",
    "\\strong{Package} \\tab \\strong{Version} \\tab ",
    loaded_hd,
    "\\strong{Title} \\tab \\strong{Method} \\cr\n",
    paste(lines, collapse = "\\cr\n"),
    "\n}\n",
    "}\n"
  )
}

has_ts_parser <- function(dsc) {
  is.list(dsc) &&
    !is.null(dsc$Imports) &&
    "tsitter" %in% parse_deps("Imports", dsc$Imports)$package
}

list_installed_packages <- function(lib_path = .libPaths()) {
  unlist(lapply(lib_path, dir, full.names = TRUE))
}

parse_deps <- function(type, deps) {
  deps <- str_trim(strsplit(deps, ",")[[1]])
  deps <- lapply(strsplit(deps, "\\("), str_trim)
  deps <- lapply(deps, sub, pattern = "\\)$", replacement = "")
  deps <- deps[vapply(deps, function(x) length(x) != 0, FUN.VALUE = logical(1))]
  res <- data.frame(
    stringsAsFactors = FALSE,
    type = if (length(deps)) type else character(),
    package = vapply(deps, "[", "", 1),
    version = vapply(deps, "[", "", 2)
  )
  res$version <- gsub("\\s+", " ", res$version)
  res$version[is.na(res$version)] <- "*"
  res
}

str_trim <- function(x) {
  sub("^\\s+", "", sub("\\s+$", "", x, useBytes = TRUE), useBytes = TRUE)
}

test_that("ts_list_parsers", {
  skip_if_not_installed("tsjsonc")
  skip_if_not_installed("tstoml")
  skip("noy yet")
  withr::local_options(width = 500)
  loadNamespace("tsjsonc")
  loadNamespace("tstoml")
  expect_snapshot(
    {
      unique(ts_list_parsers()[, 1:3])
    },
  )

  expect_snapshot(
    {
      writeLines(format_rd_parser_list(ts_list_parsers()))
    },
    transform = function(x) {
      gsub("[(][.0-9]+[)]", "(x.y.z)", x)
    }
  )

  expect_snapshot({
    format_rd_parser_list(ts_list_parsers()[integer(), ])
  })
})

#' Print a tree-sitter tree
#'
#' @description
#' Print a `ts_tree` object to the screen.
#'
#' @details
#' Calls \code{\link[tsitter:format.ts_tree]{format.ts_tree()}} to format the
#' ts_tree object, writes the formatted object to the standard output, and
#' returns the original object invisibly.
#'
#' @param x
#' `ts_tree` object to print.
#' @param n
#' Number of lines, or number of selections to print.
#' @param ...
#' Not used currently.
#'
#' @return
#' Invisibly returns the original `ts_tree` object.
#'
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' json <- tsjsonc::ts_parse_jsonc(
#'   '{ "a": 1, "b": [10, 20, 30], "c": { "c1": true, "c2": 100 } }'
#' )
#' print(json)

print.ts_tree <- function(x, n = 10, ...) {
  writeLines(format(x, n = n, ...))
  invisible(x)
}

#' Format tree-sitter trees
#'
#' @description
#' Format a `ts_tree` object for printing.
#'
#' @details
#' This is the engine of \code{\link[tsitter:print.ts_tree]{print.ts_tree()}},
#' possibly useful to obtain a printed representation without doing the
#' actual printing.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' If there are selected nodes in the tree, those will be highlighted
#' in the output. See \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} to
#' select nodes in a tree.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' @param x
#' `ts_tree` object.
#' @param n
#' Number of lines, or number of selections to print.
#' @param ...
#' Currently ignored.
#'
#' @return
#' Character vector of lines to print.
#'
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' json <- tsjsonc::ts_parse_jsonc(
#'   '{ "a": 1, "b": [10, 20, 30], "c": { "c1": true, "c2": 100 } }'
#' )
#' format(json)

format.ts_tree <- function(x, n = 10, ...) {
  sel <- ts_tree_selected_nodes(x, default = FALSE)
  if (length(sel) > 0) {
    format_ts_tree_selection(x, n = n, ...)
  } else {
    format_ts_tree_noselection(x, n = n, ...)
  }
}

format_ts_tree_noselection <- function(x, n = 10, ...) {
  lns <- strsplit(rawToChar(attr(x, "text")), "\r?\n")[[1]]
  nc <- length(lns)
  sc <- min(nc, n)
  lns <- utils::head(lns, sc)
  num <- cli::col_grey(format(seq_len(sc)))

  sel <- ts_tree_selection(x, default = FALSE)

  grey <- cli::col_grey

  sfn <- if (!is.null(attr(x, "file"))) paste0(basename(attr(x, "file")), ", ")
  lang <- sub("^ts_tree_", "", grep("^ts_tree_", class(x), value = TRUE))

  c(
    if (is.null(sel)) {
      grey(glue("# {lang} ({sfn}{nc} line{plural(nc)})"))
    } else {
      grey(glue("# {lang} ({sfn}{nc} line{plural(nc)}, 0 selected elements)"))
    },
    paste0(num, if (sc) cli::col_grey(" | "), lns),
    if (nc > sc) {
      c(
        grey(glue("{cli::symbol$info} {nc-sc} more line{plural(nc-sc)}")),
        grey(glue("{cli::symbol$info} Use `print(n = ...)` to see more lines"))
      )
    }
  )
}

ts_tree_mark_selection <- function(tree, node) {
  unlist(lapply(node, ts_tree_mark_selection1, tree = tree))
}

#' Helper function to decide which AST nodes to highlight for a selection
#' (internal)
#'
#' This function is for packages implementing new parsers based on the ts
#' package. It is very unlikely that you will need to call this function
#' directly.
#'
#' In parsers where AST nodes do not correspond one-to-one to DOM nodes
#' it is useful to highlight multiple AST nodes for a single selected
#' DOM node. This generic function can be overridden in such parsers to
#' return multiple AST node ids for a single selected (DOM) node id.
#'
#' @param tree Tree-sitter tree.
#' @param node Node id, integer scalar.
#' @return Integer vector of node ids to highlight.
#' @export
#' @examples
#' # This is an internal generic for parser implementations, see the
#' # tsjsonc and tstoml packages for examples of methods implementing
#' # custom behavior.

ts_tree_mark_selection1 <- function(tree, node) {
  UseMethod("ts_tree_mark_selection1")
}

#' @rdname ts_tree_mark_selection1
#' @details
#' The default implementation simply returns the input node id.
#' @export

ts_tree_mark_selection1.ts_tree <- function(tree, node) {
  node
}

format_ts_tree_selection <- function(x, n = n, context = 3, ...) {
  lns <- strsplit(rawToChar(attr(x, "text")), "\r?\n")[[1]]
  nlns <- length(lns)
  num <- seq_along(lns)
  sel <- ts_tree_selected_nodes(x, default = FALSE)
  nsel <- length(sel)
  ssel <- min(nsel, n)
  sel <- utils::head(sel, ssel)
  isel <- ts_tree_mark_selection(x, sel)

  # calculate the lines affected by the first ssel selections
  selrows <- rep(FALSE, nlns)
  shwrows <- rep(FALSE, nlns)
  for (sel1 in isel) {
    beg <- x$start_row[sel1] + 1L
    end <- x$end_row[sel1] + 1L
    end <- x$end_row[sel1] + 1L
    if (x$end_column[sel1] == 0) {
      end <- end - 1L
    }
    selrows[beg:end] <- TRUE
    sbeg <- max(1, beg - context)
    send <- min(nlns, end + context)
    shwrows[sbeg:send] <- TRUE
  }

  # now highlight the selected elements
  mark <- rep("  ", nlns)
  for (sel1 in isel) {
    beg <- x$start_row[sel1] + 1L
    end <- x$end_row[sel1] + 1L
    endcol <- x$end_column[sel1]
    if (x$end_column[sel1] == 0) {
      end <- end - 1L
      endcol <- cli::ansi_nchar(lns[end], type = "bytes")
    }
    rows <- beg:end
    mark[rows] <- paste0(cli::bg_cyan(">"), " ")
    # one row only
    if (length(rows) == 1) {
      lns[rows] <- hl(
        lns[rows],
        x$start_column[sel1] + 1L,
        endcol
      )
    } else {
      # first row
      lns[rows[1]] <- hl(lns[rows[1]], x$start_column[sel1] + 1L, end = NULL)
      # middle rows, if any
      if (length(rows) > 2) {
        mid <- middle(rows)
        lns[mid] <- hl(lns[mid])
      }
      # last row
      if (length(rows) >= 2) {
        lns[rows[length(rows)]] <- hl(
          lns[rows[length(rows)]],
          start = NULL,
          endcol
        )
      }
    }
  }

  grey <- cli::col_grey

  # add ... between consecutive lines
  dots <- diff(c(1, num[shwrows])) > 1
  dotlns <- num[shwrows][dots] - 1L
  # add ... to the end
  lastshown <- utils::tail(num[shwrows], 1)
  if (lastshown < nlns) {
    dotlns <- c(dotlns, lastshown + 1L)
  }
  shwrows[dotlns] <- TRUE

  # format the ... lines
  num[shwrows] <- format(num[shwrows])
  num[dotlns] <- "..."
  num <- grey(format(num[shwrows]))
  lns[dotlns] <- ""
  split <- rep(cli::col_grey(" | "), nlns)
  split[dotlns] <- "   "

  slns <- paste0(mark[shwrows], num, split[shwrows], lns[shwrows])

  sfn <- if (!is.null(attr(x, "file"))) paste0(basename(attr(x, "file")), ", ")
  lang <- sub("^ts_tree_", "", grep("^ts_tree_", class(x), value = TRUE))

  c(
    grey(glue(
      "# {lang} ({sfn}{nlns} line{plural(nlns)}, \\
      {nsel} selected element{plural(nsel)})"
    )),
    slns,
    if (nsel > ssel) {
      c(
        grey(glue(
          "{cli::symbol$info} {nsel-ssel} more selected \\
          element{plural(nsel-ssel)}"
        )),
        grey(glue(
          "{cli::symbol$info} Use `print(n = ...)` to see more selected \\
           elements"
        ))
      )
    }
  )
}

plural <- function(x) {
  if (x != 1) {
    "s"
  } else {
    ""
  }
}

# TODO: only vectorized for the default case

hl <- function(txt, start = NULL, end = NULL) {
  if (is.null(start) && is.null(end)) {
    cli::col_cyan(txt)
  } else if (is.null(end) && !is.null(start)) {
    stopifnot(length(txt) == 1)
    paste0(
      if (start > 1) {
        cli::ansi_substr(txt, 1, start - 1)
      },
      cli::col_cyan(cli::ansi_substr(txt, start, cli::ansi_nchar(txt)))
    )
  } else if (is.null(start) && !is.null(end)) {
    stopifnot(length(txt) == 1)
    nc <- cli::ansi_nchar(txt)
    paste0(
      cli::col_cyan(cli::ansi_substr(txt, 1, end)),
      if (end < nc) {
        cli::ansi_substr(txt, end + 1, nc)
      }
    )
  } else {
    stopifnot(length(txt) == 1)
    nc <- cli::ansi_nchar(txt)
    paste0(
      if (start > 1) {
        cli::ansi_substr(txt, 1, start - 1)
      },
      cli::col_cyan(cli::ansi_substr(txt, start, end)),
      if (end < nc) {
        cli::ansi_substr(txt, end + 1, nc)
      }
    )
  }
}

## usethis namespace: start
#' @useDynLib tsitter, .registration = TRUE, .fixes = "c_"
## usethis namespace: end
NULL

## usethis namespace: start
## usethis namespace: end
NULL

#' @name about
#' @title About tsitter
#' @description
#' tsitter is a common interface to [tree-sitter](
#'  https://tree-sitter.github.io/tree-sitter/) parsers, implemented in
#' other R packages. It has a common API to
#'
#' - query,
#' - edit,
#' - format, and
#' - unserialize
#'
#' tree-sitter parse trees.
#'
#' @return
#' Not applicable.
#'
#' @examples
#' # See above please.
#'
#' @details
#'
#' ```{r, child = "tools/man/about.Rmd"}
#' ```
NULL

#' Replace selected elements with a new element in a tree-sitter tree
#'
#' @description
#' Replace all selected elements with a new element.
#'
#' @details
#' If the tree does not have a selection, the new element replaces the
#' whole document.
#'
#' If the tree has an empty selection, the new element is inserted at the
#' position of where the selected elements would be.
#'
#' @param tree
#' A `ts_tree` object.
#' @param new
#' The new element to replace the selected elements with.
#'
#' The type of `new` depends on the parser and the method that implements
#' the insertion. See details in the manual of the specific parser.
#'
#' See details in the manual of the specific parser.
#' @param options
#' A list of options for the update.
#' @param ... Extra arguments for methods.
#'
#' @return
#' The modified `ts_tree` object with the selected elements replaced
#' by the new element.
#'
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc(r"(
#'   {
#'     "name": "example",
#'     "version": "1.0.0",
#'     "dependencies": {
#'       "tsjsonc": "^0.1.0"
#'     }
#'   }
#' )")
#'
#' ts_tree_update(ts_tree_select(tree, "version"), "2.0.0")
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#' # Create a parse tree with tstoml --------------------------------------
#' tree <- tstoml::ts_parse_toml(r"(
#'   [package]
#'   name = "example"
#'   version = "1.0.0"
#'   depdendencies = { tstoml = "0.1.0" }
#' )")
#'
#' ts_tree_update(ts_tree_select(tree, "package", "version"), "2.0.0")

ts_tree_update <- function(tree, new, options, ...) {
  UseMethod("ts_tree_update")
}

`%||%` <- function(l, r) {
  if (is.null(l)) {
    r
  } else {
    l
  }
}

`%&&%` <- function(l, r) {
  if (is.null(l)) {
    NULL
  } else {
    r
  }
}

map_int <- function(.x, .f, ...) {
  vapply(.x, .f, integer(1), ...)
}

map_chr <- function(.x, .f, ...) {
  vapply(.x, .f, character(1), ...)
}

map_lgl <- function(.x, .f, ...) {
  vapply(.x, .f, logical(1), ...)
}

pmap <- function(.l, .f) {
  do.call(mapply, c(list(FUN = .f, SIMPLIFY = FALSE, USE.NAMES = FALSE), .l))
}

imap <- function(.x, .f) {
  idx <- names(.x) %||% seq_along(.x)
  mapply(.f, .x, idx, SIMPLIFY = FALSE, USE.NAMES = FALSE)
}

get_tree_lang <- function(tree) {
  cls <- grep("^ts_tree_", class(tree), value = TRUE)
  if (length(cls) == 0) {
    "<unknown>"
  } else {
    sub("^ts_tree_", "", cls[1])
  }
}

middle <- function(x) {
  if (length(x) <= 2) {
    x[numeric()]
  } else {
    x[-c(1, length(x))]
  }
}

is_named <- function(x) {
  nms <- names(x)
  length(x) == length(nms) && !anyNA(nms) && all(nms != "")
}

get_env <- function(x) {
  ev <- Sys.getenv(x, unset = NA_character_)
  if (is.na(ev)) {
    NULL
  } else {
    ev
  }
}

read_char <- function(fn) {
  paste0(readLines(fn, warn = FALSE, encoding = "UTF-8"), collapse = "\n")
}

test_that("%||%", {
  expect_equal(NULL %||% "foo", "foo")
  expect_equal("foo" %||% stop("nope"), "foo")
})

test_that("map_int", {
  expect_equal(
    map_int(1:3, function(x) x),
    1:3
  )

  expect_equal(
    map_int(c(a = 1, b = 2), function(x) as.integer(x)),
    c(a = 1L, b = 2L)
  )

  expect_equal(
    map_int(integer(), function(x) stop("not called")),
    integer()
  )

  expect_equal(
    map_int(letters[1:3], function(x) 1L),
    c(a = 1L, b = 1L, c = 1L)
  )
})

test_that("map_chr", {
  expect_equal(
    map_chr(1:3, function(x) as.character(x)),
    c("1", "2", "3")
  )

  expect_equal(
    map_chr(c(a = 1, b = 2), function(x) as.character(x)),
    c(a = "1", b = "2")
  )

  expect_equal(
    map_chr(integer(), function(x) stop("not called")),
    character()
  )

  expect_equal(
    map_chr(letters[1:3], function(x) "x"),
    c(a = "x", b = "x", c = "x")
  )
})
test_that("map_lgl", {
  expect_equal(
    map_lgl(1:3, function(x) x > 2),
    c(FALSE, FALSE, TRUE)
  )

  expect_equal(
    map_lgl(c(a = 1, b = 2), function(x) x > 1),
    c(a = FALSE, b = TRUE)
  )

  expect_equal(
    map_lgl(integer(), function(x) stop("not called")),
    logical()
  )
})

test_that("pmap", {
  expect_equal(
    pmap(
      list(1:3, 4:6),
      function(x, y) x + y
    ),
    list(5, 7, 9)
  )
})

test_that("imap", {
  expect_equal(
    imap(
      c(a = 1, b = 2, c = 3),
      function(x, i) paste0(i, x)
    ),
    list("a1", "b2", "c3")
  )

  expect_equal(
    imap(
      c(1, 2, 3),
      function(x, i) paste0(i, x)
    ),
    list("11", "22", "33")
  )
})

test_that("get_tree_lang", {
  skip_if_not_installed("tsjsonc")
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
  expect_equal(get_tree_lang(tree), "jsonc")
  expect_equal(get_tree_lang(list()), "<unknown>")
})

test_that("middle", {
  expect_equal(middle(1:5), 2:4)
  expect_equal(middle(1:2), integer())
  expect_equal(middle(1), integer())
  expect_equal(middle(integer()), integer())
})

test_that("is_named", {
  expect_true(is_named(c(a = 1, b = 2)))
  expect_false(is_named(c(1, 2)))
  x <- c(a = 1, 2)
  names(x) <- c("a", "")
  expect_false(is_named(x))
  x <- c(a = 1)
  x[2] <- 2
  expect_false(is_named(x))
  expect_true(is_named(c()))
})

#' Unserialize selected elements of a tree-sitter tree
#'
#' @description
#' Unserialize the selected elements of a `ts_tree` object, i.e. convert
#' them to R objects.
#'
#' @details
#' If no elements are selected in the tree, then the whole document is
#' unserialized.
#'
#' If the tree has an empty selection, then an empty list is returned.
#'
#' ## The `[[` operator
#'
#' The `[[` operator works similarly to the combination of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
#' `ts_tree_unserialize()`, but it might be more readable.
#'
#' For the details on how the selected elements are mapped to R objects,
#' see the documentation of the methods in the parser packages. The methods
#' in the installed parser packages are linked below.
#'
#' @param tree
#' A `ts_tree` object.
#'
#' @return
#' List of R objects, with one entry for each selected element.
#'
#' @family ts_tree generics
#' @family serialization functions
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')
#'
#' tree
#'
#' ts_tree_unserialize(ts_tree_select(tree, c("b", "c")))
#'
#' ts_tree_unserialize(ts_tree_select(tree, "b"))

ts_tree_unserialize <- function(tree) {
  UseMethod("ts_tree_unserialize")
}

#' Print the document object model (DOM) of a tree-sitter tree
#'
#' @description
#' `ts_tree_dom()` prints the document object model (DOM) tree of a ts_tree
#' object. This tree only includes semantic elements. E.g. for a JSON(C)
#' document it includes objects, arrays and various value types, but not
#' the syntax elements like brackets, commas or colons.
#'
#' @details
#' ## The syntax tree and the DOM tree
#'
#' See \code{\link[tsitter:ts_tree_ast]{ts_tree_ast()}} for the complete
#' tree-sitter syntax tree that includes all nodes, including syntax
#' elements like brackets and commas.
#'
#' @param tree
#' A `ts_tree` object.
#'
#' @return
#' Character vector, the formatted annotated syntax tree, line by
#' line. It has class [cli_tree][cli::tree()], from the cli package. It
#' may contain ANSI escape sequences for coloring and hyperlinks.
#'
#' @family ts_tree exploration
#' @family ts_tree generics
#' @seealso [ts_tree_ast()] to show the annotated syntax tree of a
#'   ts_tree object.
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')
#'
#' tree
#'
#' ts_tree_ast(tree)
#'
#' ts_tree_dom(tree)
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#'
#' # Create a parse tree with tstoml --------------------------------------
#' tree <- tstoml::ts_parse_toml(r"(
#'   title = "TOML Example"
#'   [owner]
#'   name = "Tom Preston-Werner"
#'   dob = 1979-05-27T07:32:00-08:00
#' )")
#'
#' tree
#'
#' ts_tree_ast(tree)
#'
#' ts_tree_dom(tree)

ts_tree_dom <- function(tree) {
  UseMethod("ts_tree_dom")
}

#' @export

ts_tree_dom.default <- function(tree) {
  is_dom_node <- c(1L, which(!is.na(tree$dom_parent)))
  dom <- tree[is_dom_node, ]

  treetab <- data_frame(
    id = as.character(dom$id),
    children = lapply(dom$dom_children, as.character),
    label = paste0(
      dom$dom_type,
      " (",
      dom$id,
      ")",
      ifelse(
        is.na(dom$dom_name),
        "",
        cli::col_grey(paste0(" # ", dom$dom_name))
      )
    )
  )

  tree <- cli::tree(treetab)
  tree
}

#' Format the selected elements of a tree sitter tree for printing
#'
#' @description
#' (Re)format the selected elements of the document represented by a
#' tree-sitter tree, if the tree-sitter parser supports formatting.
#'
#' @details
#' If `tree` does not have a selection, then the whole document is
#' formatted.
#'
#' If `tree` has an empty selection, then it is returned unchanged.
#'
#' Some parsers support options to customize the formatting.
#' See details in the manual of the specific parser.
#'
#' @param tree
#' A `ts_tree` object.
#'
#' @param options
#' A list of options for the formatting.
#'
#' See details in the manual of the specific parser.
#'
#' @param ... Extra arguments for methods.
#'
#' @return
#' A `ts_tree` object representing the reformatted document.
#'
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{ "a":true, "b": [1,2,3] }')
#' tree
#'
#' # Format whole document
#' ts_tree_format(tree)
#'
#' # Format each top element under the document node in one line
#' ts_tree_format(
#'   ts_tree_select(ts_tree_format(tree), TRUE),
#'   options = list(format = "oneline")
#' )
#'
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#' # Create a parse tree with tstoml --------------------------------------
#' tree <- tstoml::ts_parse_toml(r"(
#'   [servers]
#'   alpha = { ip = "127.0.0.1", dc = "eqdc10" }
#'   beta = { ip = "127.0.0.2", dc = "eqdc20" }
#' )")
#'
#' tree
#'
#' ts_tree_format(tree)

ts_tree_format <- function(tree, options, ...) {
  UseMethod("ts_tree_format")
}

test_that("format.ts_tree", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}') %>%
      ts_tree_format()
    print(tree)
    print(tree, n = 1)
    print(tree, n = 2)
  })

  # selections
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}') %>%
      ts_tree_format()
    tree %>% ts_tree_select("a") %>% print()
    tree %>% ts_tree_select("b", TRUE) %>% print()
  })

  # empty selection
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}') %>%
      ts_tree_format()
    tree %>% ts_tree_select("c") %>% print()
  })

  # multiple selected rows
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3], "c": {"d":4}}') %>%
      ts_tree_format()
    tree %>% ts_tree_select(c("b", "c")) %>% print()
  })

  # print limited number of selected elements
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('[1,2,3,4,5,6,7]') %>%
      ts_tree_format()
    tree %>% ts_tree_select(1:5) %>% print(n = 3)
  })
})

test_that("end_column == 0", {
  skip_if_not_installed("tstoml")
  library(magrittr)
  expect_snapshot({
    tree <- tstoml::ts_parse_toml(tstoml::toml_example_text())
    tree %>% ts_tree_select("owner")
  })
})

#' @rdname internal
#' @param s For `ts_collapse()` a character vector to collapse.
#' @param sep Separator string for most elements.
#' @param sep2 Separator string for two elements.
#' @param last Separator string before the last element.
#' @param trunc Integer, maximum number of elements to show before
#'   truncation.
#' @param width Integer, maximum display width of the collapsed string.
#'   If the collapsed string exceeds this width, it will be truncated
#'   with `ellipsis`.
#' @param ellipsis String to indicate truncation.
#' @param style Character, the collapsing style to use. Possible values are
#'   `"both-ends"` (the default), which shows the first few and last few
#'   elements when truncating, and `"head"`, which shows only the first few
#'   elements.
#' @return `ts_collapse()` returns a character scalar, the collapsed string.
#' @details `ts_collapse()` collapses a character vector into a single string,
#'   with options for truncation by number of elements or display width.
#'   It is useful for creating informative error messages.
#' @export
#' @examples
#' ts_collapse(letters[1:3])
#' ts_collapse(letters[1:10], trunc = 5)

ts_collapse <- function(
  s,
  sep = ", ",
  sep2 = sub("^,", "", last),
  last = ", and ",
  trunc = Inf,
  width = Inf,
  ellipsis = "...",
  style = c("both-ends", "head")
) {
  style <- match.arg(style)
  switch(
    style,
    "both-ends" = collapse_both_ends(
      s,
      sep,
      sep2,
      last,
      trunc,
      width,
      ellipsis
    ),
    "head" = collapse_head(s, sep, sep2, last, trunc, width, ellipsis)
  )
}

collapse_head_notrim <- function(x, trunc, sep, sep2, last, ellipsis) {
  lnx <- length(x)

  if (lnx == 1L) {
    return(x)
  }
  if (lnx == 2L) {
    return(paste0(x, collapse = sep2))
  }
  if (lnx <= trunc) {
    # no truncation
    return(paste0(
      paste(x[1:(lnx - 1L)], collapse = sep),
      last,
      x[lnx]
    ))
  } else {
    # truncation, no need for 'last'
    return(paste0(
      paste(x[1:trunc], collapse = sep),
      sep,
      ellipsis
    ))
  }
}

collapse_head <- function(x, sep, sep2, last, trunc, width, ellipsis) {
  trunc <- max(trunc, 1L)
  x <- as.character(x)
  lnx <- length(x)

  # special cases that do not need trimming
  if (lnx == 0L) {
    return("")
  } else if (anyNA(x)) {
    return(NA_character_)
  }

  # easier case, no width trimming
  if (width == Inf) {
    return(collapse_head_notrim(x, trunc, sep, sep2, last, ellipsis))
  }

  # complex case, with width wrapping
  # first we truncate
  tcd <- lnx > trunc
  if (tcd) {
    x <- x[1:trunc]
  }

  # then we calculate the width w/o trimming
  wx <- nchar(x)
  wsep <- nchar(sep, "width")
  wsep2 <- nchar(sep2, "width")
  wlast <- nchar(last, "width")
  well <- nchar(ellipsis, "width")
  if (!tcd) {
    # x[1]
    # x[1] and x[2]
    # x[1], x[2], and x[3]
    nsep <- if (lnx > 2L) lnx - 2L else 0L
    nsep2 <- if (lnx == 2L) 1L else 0L
    nlast <- if (lnx > 2L) 1L else 0L
    wtot <- sum(wx) + nsep * wsep + nsep2 * wsep2 + nlast * wlast
    if (wtot <= width) {
      if (lnx == 1L) {
        return(x)
      } else if (lnx == 2L) {
        return(paste0(x, collapse = sep2))
      } else {
        return(paste0(
          paste(x[1:(lnx - 1L)], collapse = sep),
          last,
          x[lnx]
        ))
      }
    }
  } else {
    # x[1], x[2], x[trunc], ...
    wtot <- sum(wx) + trunc * wsep + well
    if (wtot <= width) {
      return(paste0(
        paste(x, collapse = sep),
        sep,
        ellipsis
      ))
    }
  }

  # we need to find the longest possible truncation for the form
  # x[1], x[2], x[trunc], ...
  # each item is wx + wsep wide, so we search how many fits, with ellipsis
  last <- function(x) if (length(x) >= 1L) x[length(x)] else x[NA_integer_]
  trunc <- last(which(cumsum(wx + wsep) + well <= width))

  # not even one element fits
  if (is.na(trunc)) {
    if (well > width) {
      return(strtrim(ellipsis, width))
    } else if (well == width) {
      return(ellipsis)
    } else if (well + wsep >= width) {
      return(paste0(strtrim(x[1L], width), ellipsis))
    } else {
      return(paste0(
        strtrim(x[1L], max(width - well - wsep, 0L)),
        sep,
        ellipsis
      ))
    }
  }

  return(paste0(
    paste(x[1:trunc], collapse = sep),
    sep,
    ellipsis
  ))
}

collapse_both_ends <- function(x, sep, sep2, last, trunc, width, ellipsis) {
  # we always list at least 5 elements
  trunc <- max(trunc, 5L)
  trunc <- min(trunc, length(x))
  if (length(x) <= 5L || length(x) <= trunc) {
    return(collapse_head(x, sep, sep2, last, trunc = trunc, width, ellipsis))
  }

  # we have at least six elements in the vector
  # 1, 2, 3, ..., 9, and 10
  x <- as.character(c(x[1:(trunc - 2L)], x[length(x) - 1L], x[length(x)]))
  paste0(
    c(x[1:(trunc - 2L)], ellipsis, paste0(x[trunc - 1L], last, x[trunc])),
    collapse = sep
  )
}

trim <- function(x) {
  has_newline <- function(x) any(grepl("\\n", x))
  if (length(x) == 0L || !has_newline(x)) {
    return(x)
  }
  call_with_cleanup(c_trim, x)
}

#' Create tree-sitter tree from file or string
#'
#' @description
#' This is the main function to create a tree-sitter parse tree, using a
#' ts parser implemented in another package.
#'
#' The result is a `ts_tree` object. A `ts_tree` object may be queried,
#' edited, formatted, written to file, etc. using `ts_tree` methods.
#'
#' @details
#' A package that implements a tree-sitter parser provides a function that
#' creates a `ts_language` object for that parser. E.g.
#' tsjsonc has `tsjsonc::ts_language_jsonc()`.
#' You need to use the returned `ts_language` object as the `language`
#' argument of \code{\link[tsitter:ts_tree_new]{ts_tree_new()}}.
#'
#' @param language
#' Language of the file or string, a `ts_language` object, e.g. the return
#' value of `tsjsonc::ts_language_jsonc()`.
#' @param file
#' Path of a file to parse. Use either `file` or `text`, but not both.
#' @param text
#' String to parse. Use either `file` or `text`, but not both.
#' @param ranges
#' Can be used to parse part(s) of the input. It must be a data frame with
#' integer columns `start_row`, `start_col`, `end_row`, `end_col`,
#' `start_byte`, `end_byte`, in this order.
#' @param fail_on_parse_error
#' Logical, whether to error if there are parse errors in the document.
#' Default is `TRUE`.
#' @param ... Additional arguments for methods.
#'
#' @return
#' A `ts_tree` object representing the parse tree of the input. You can
#' use the single bracket \code{\link[tsitter:ts_tree-brackets]{[}}
#' operator to convert it to a data frame.
#'
#' @export
#' @family ts_tree generics
#' @seealso The tree-sitter parser packages typically include shortcuts to
#'   create parse trees from strings and file, e.g.
#'   `tsjsonc::ts_parse_jsonc()` and `tsjsonc::ts_read_jsonc()`.
#'
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#'
#' # JSONC example, needs the tsjsonc package -----------------------------
#' json <- ts_tree_new(
#'   tsjsonc::ts_language_jsonc(),
#'   text = '{ "a": 1, "b": 2 }'
#' )
#'
#' json
#'
#' ts_tree_format(json)
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#'
#' # TOML example, needs the tstoml package -------------------------------
#' toml <- ts_tree_new(
#'   tstoml::ts_language_toml(),
#'   text = '[section]\nkey = "value"\nnumber = 42\n'
#' )
#'
#' toml
#'
#' ts_tree_format(toml)

ts_tree_new <- function(
  language,
  file = NULL,
  text = NULL,
  ranges = NULL,
  fail_on_parse_error = TRUE,
  ...
) {
  UseMethod("ts_tree_new")
}

#' @export

ts_tree_new.ts_language <- function(
  language,
  file = NULL,
  text = NULL,
  ranges = NULL,
  fail_on_parse_error = TRUE,
  ...
) {
  if (is.null(text) + is.null(file) != 1) {
    stop(ts_cnd(
      "Invalid arguments in `ts_tree_new()`: exactly one of `file` \\
       and `text` must be given."
    ))
  }
  if (is.null(text)) {
    text <- readBin(file, "raw", n = file.size(file))
  }
  if (is.character(text)) {
    text <- charToRaw(paste(text, collapse = "\n"))
  }

  tree <- call_with_cleanup(c_parse, text, language, ranges)

  lvls <- seq_len(nrow(tree))
  tree$children <- I(unname(split(
    lvls,
    factor(tree$parent, levels = lvls)
  )))

  # trailing whitespace for each token
  # first we add the leading whitespace to the document token
  # this way printing $code and $tws will print the whole document
  tree$tws <- rep("", nrow(tree))
  if ((lead <- tree$start_byte[1]) > 0) {
    tree$tws[1] <- rawToChar(text[1:lead])
  }

  # then the whitespace of the terminal nodes
  term <- which(!is.na(tree$code))
  from <- tree$end_byte[term] + 1L
  to <- c(tree$start_byte[term][-1], tree$end_byte[1])
  for (i in seq_along(term)) {
    if (from[i] <= to[i]) {
      tree$tws[term[i]] <- rawToChar(text[from[i]:to[i]])
    }
  }

  attr(tree, "text") <- text
  attr(tree, "file") <- if (!is.null(file)) normalizePath(file)
  cls <- sub("^ts_language_", "ts_tree_", class(language)[1])
  class(tree) <- c(cls, "ts_tree", class(tree))

  if (fail_on_parse_error && (tree$has_error[1] || any(tree$is_missing))) {
    stop(ts_parse_error_cnd(tree = tree, text = text))
  }

  language <- structure(
    list(language = language, tree = tree),
    class = class(tree)
  )

  tree
}

test_that("ts_tree_new", {
  skip_if_not_installed("tsjsonc")
  expect_snapshot({
    tree <- ts_tree_new(
      tsjsonc::ts_language_jsonc(),
      text = '{"a": 1, "b": [1,2,3]}'
    )
    tree
  })

  # error if both text and file are NULL or present
  expect_snapshot(
    error = TRUE,
    {
      ts_tree_new(tsjsonc::ts_language_jsonc())
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
  expect_snapshot(
    error = TRUE,
    {
      ts_tree_new(
        tsjsonc::ts_language_jsonc(),
        text = '{"a": 1}',
        file = tempfile()
      )
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )

  # reading from file
  tmp <- file.path(tempfile(), "test.jsonc")
  on.exit(unlink(dirname(tmp), recursive = TRUE), add = TRUE)
  dir.create(dirname(tmp), recursive = TRUE)
  writeLines('{"a": 1, "b": [1,2,3]}', con = tmp)
  expect_snapshot(
    {
      tree <- ts_tree_new(tsjsonc::ts_language_jsonc(), file = tmp)
      tree
    },
    transform = redact_tempfile,
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("ts_tree_new leading whitespace", {
  skip_if_not_installed("tsjsonc")
  expect_snapshot({
    tree <- ts_tree_new(
      tsjsonc::ts_language_jsonc(),
      text = '\n\n   \n{"a": 1, "b": [1,2,3]}'
    )
    tree
  })
})

test_that("ts_tree_new parse error", {
  skip_if_not_installed("tsjsonc")
  expect_snapshot(
    error = TRUE,
    {
      ts_tree_new(
        tsjsonc::ts_language_jsonc(),
        text = '{"a": 1, "b": [1,2,3' # missing closing ]
      )
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

#' Select elements of a tree-sitter tree
#'
#' @description
#' This function is the heart of ts. To edit a tree-sitter tree, you first
#' need to select the parts you want to delete or update.
#'
#' @details
#' The selection process is iterative. Selection expressions (selectors)
#' are applied one by one, and each selector selects nodes from the
#' currently selected nodes. For each selector, it is applied individually
#' to each currently selected node, and the results are concatenated.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' The selection process starts from the root of the DOM tree, the document
#' node (see \code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}), unless
#' `refine = TRUE` is set, in which case it starts from the current
#' selection.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' See the various types of selection expressions below.
#'
#' ## Selectors
#'
#' ### All elements: `TRUE`
#'
#' Selects all child nodes of the current nodes.
#'
#' ### Specific keys: character vector
#'
#' Selects child nodes with the given names from nodes with named children.
#' If a node has no named children, it selects nothing from that node.
#'
#' ### By position: integer vector
#'
#' Selects child nodes by position. Positive indices count from the start,
#' negative indices count from the end. Zero indices are not allowed.
#'
#' ### Matching keys: regular expression
#'
#' A character scalar named `regex` can be used to select child nodes
#' whose names match the given regular expression, from nodes with named
#' children. If a node has no named children, it selects nothing from that
#' node.
#'
#' ### Tree sitter query matches
#'
#' A character scalar named `query` can be used to select nodes matching
#' a tree-sitter query. See \code{\link[tsitter:ts_tree_query]{ts_tree_query()}}
#' for details on tree-sitter queries.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' Instead of a character scalar this can also be a two-element list, where
#' the first element is the query string and the second element is a
#' character vector of capture names to select. In this case only nodes
#' matching the given capture names will be selected.
#'
#' ### Explicit node ids
#'
#' You can use `I(c(...))` to select nodes by their ids directly. This is
#' for advanced use cases only.
#'
#' ## Refining selections
#'
#' If the `refine` argument of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} is `TRUE`, then
#' the selection starts from the already selected elements (all of them
#' simultanously), instead of starting from the document element.
#'
#' ## The `ts_tree_select<-()` replacement function
#'
#' The \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}} replacement
#' function works similarly to the combination of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
#' \code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, but it might be more
#' readable.
#'
#' ## The `[[` and `[[<-` operators
#
#' The `[[` operator works similarly to the combination of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
#' \code{\link[tsitter:ts_tree_unserialize]{ts_tree_unserialize()}}, but it
#' might be more readable.
#'
#' The `[[<-` operator works similarly to the combination of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
#' \code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
#' replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
#' but it might be more readable.
#'
#' @param tree
#' A `ts_tree` object as returned by
#' \code{\link[tsitter:ts_tree_new]{ts_tree_new()}}.
#' @param ...
#' Selection expressions, see details.
#' @param refine
#' Logical, whether to refine the current selection or start
#' a new selection.
#'
#' @return
#' A `ts_tree` object with the selected parts.
#'
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # ----------------------------------------------------------------------
#' # Create a JSONC tree, needs the tsjsonc package
#' json <- ts_tree_new(
#'   tsjsonc::ts_language_jsonc(),
#'   text = '{ "a": 1, "b": 2, "c": { "d": 3, "e": 4 } }'
#' )
#'
#' ts_tree_select(json, "c", "d")
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#'
#' # ----------------------------------------------------------------------
#' # Create a TOML tree, needs the tstoml package
#' toml <- ts_tree_new(
#'   tstoml::ts_language_toml(),
#'   text = tstoml::toml_example_text()
#' )
#'
#' ts_tree_select(toml, "servers", TRUE, "ip")

ts_tree_select <- function(tree, ..., refine = FALSE) {
  slts <- normalize_selectors(tree, list(...))
  if (length(slts) == 1 && is.null(slts[[1]])) {
    attr(tree, "selection") <- NULL
    return(tree)
  }
  current <- if (refine) {
    ts_tree_selection(tree)
  } else {
    ts_tree_selector_default(tree)
  }
  cnodes <- current[[length(current)]]$nodes

  for (idx in seq_along(slts)) {
    slt <- slts[[idx]]
    nxt <- integer()
    for (cur in cnodes) {
      nxt <- unique(c(nxt, ts_tree_select1(tree, cur, slt)))
    }
    current[[length(current) + 1L]] <- list(
      selector = slt,
      nodes = sort(nxt)
    )
    cnodes <- current[[length(current)]]$nodes
  }
  attr(tree, "selection") <- current
  tree
}

normalize_selectors <- function(tree, slts) {
  names(slts) <- names(slts) %||% rep("", length(slts))
  slts <- imap(slts, function(x, nm) {
    if (nm == "regex") {
      x <- structure(
        list(pattern = x),
        class = c("ts_tree_selector_regex", "ts_tree_selector", "list")
      )
    } else if (nm == "query") {
      # we do the query up front, so we don't rerun it for every node
      x <- structure(
        list(
          query = if (is.character(x)) x else x[[1]],
          captures = if (!is.character(x)) x[[2]] else NULL
        ),
        class = c("ts_tree_selector_tsquery", "ts_tree_selector", "list")
      )
      x$nodes <- select_query(tree, x$query, x$captures)
    } else if (inherits(x, "AsIs")) {
      x <- structure(
        list(ids = unclass(x)),
        class = c("ts_tree_selector_ids", "ts_tree_selector", "list")
      )
    }
    x <- if (inherits(x, "ts_tree_selector") || !is.list(x)) list(x) else x
    x
  })
  unlist(slts, recursive = FALSE, use.names = FALSE)
}

#' Select nodes from a tree-sitter tree (internal)
#'
#' This function is for packages implementing new parsers based on the ts
#' package. It is very unlikely that you will need to call this function
#' directly.
#'
#' A parser package may implement methods for this generic to change the
#' behavior of \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} for a
#' certain selector type, or even add new selector types.
#'
#' Each new method should be named as
#' ```
#' ts_tree_select.<ts_tree_class>.<selector_class>
#' ```
#'
#' The ts package implement deault methods for the selector types described
#' in the \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} manual page.
#'
#' @param tree A `ts_tree` object as returned by [ts_tree_new()].
#' @param node Integer scalar, the node id to select from.
#' @param slt A selector object, see details in [ts_tree_select()].
#'
#' @return Must return an integer vector of selected node ids.
#'
#' @export
#' @examples
#' # This is an internal generic for parser implementations, see the
#' # tsjsonc and tstoml packages for examples of methods implementing
#' # selector types.

ts_tree_select1 <- function(tree, node, slt) {
  treesel <- structure(
    list(tree = tree, slt = slt),
    class = paste0(class(tree), ".", class(slt)[1])
  )
  UseMethod("ts_tree_select1", treesel)
}

#' @export

ts_tree_select1.default <- function(tree, node, slt) {
  lang <- toupper(get_tree_lang(tree))
  stop(ts_cnd(
    "Don't know how to select nodes from a `ts_tree` ({lang}) object \\
     using selector of class `{class(slt)[1]}`."
  ))
}

#' @rdname ts_tree_select1
#' @details
#' ## `ts_tree_selector_default` selector
#'
#' Method: `ts_tree_select1.ts_tree.ts_tree_selector_default`
#'
#' This method is used to select the default element(s), when there is no
#' selected element. E.g. when starting a new selection from the root of
#' the DOM tree.
#'
#' The default implementation returns the ids of all children of the
#' document root in the AST, except comments. If there are no such
#' children, it returns the id of the document root of the AST itself
#' (always id 1).
#' @export

ts_tree_select1.ts_tree.ts_tree_selector_default <- function(tree, node, slt) {
  top <- tree$children[[1]]
  top <- top[tree$type[top] != "comment"]
  if (length(top) > 0) top else 1L
}

#' @rdname ts_tree_select1
#' @details
#' ## `NULL` selector
#'
#' Method: `ts_tree_select1.ts_tree.NULL`
#'
#' This method is used for the  `NULL` selector, that is supposed to
#' clear the selection. You probably do not need to override this method.
#' The default implementation returns an empty integer vector.
#' @export

ts_tree_select1.ts_tree.NULL <- function(tree, node, slt) {
  integer()
}

#' @rdname ts_tree_select1
#' @details
#' ## `ts_tree_selector_ids` selector
#'
#' Method: `ts_tree_select1.ts_tree.ts_tree_selector_ids`
#'
#' This method is used to select nodes by their ids directly.
#' You probably do not need to override this method.
#' The default implementation returns the ids stored in the selector.
#'
#' ### Note
#'
#' This behaviour may change in the future to select only nodes
#' in the subtree of the current node.
#' @export

ts_tree_select1.ts_tree.ts_tree_selector_ids <- function(tree, node, slt) {
  # TODO: should we select in subtree of node? Probably.
  slt$ids
}

#' @rdname ts_tree_select1
#' @details
#' ## `ts_tree_selector_tsquery` selector
#'
#' Method: `ts_tree_select1.ts_tree.ts_tree_selector_tsquery`
#'
#' This method is used to select nodes matching a tree-sitter query.
#' You probably do not need to override this method.
#' The default implementation returns the ids stored in the selector.
#'
#' ### Note
#'
#' This behaviour may change in the future to select only nodes
#' in the subtree of the current node.
#' @export

ts_tree_select1.ts_tree.ts_tree_selector_tsquery <- function(tree, node, slt) {
  # TODO: should we select in subtree of node? Probably.
  slt$nodes
}

#' @rdname ts_tree_select1
#' @details
#' ## `character` (character vector) selector
#'
#' Method: `ts_tree_select1.ts_tree.character`
#'
#' This method is used when the selector is a character vector.
#' The default implementation selects DOM children of `node` whose names
#' are in the character vector. If not all children o `node` are named,
#' it returns an empty integer vector. (E.g. in a JSONC document it returns
#' an empty integer vector when nodes is an array.)
#' @export

ts_tree_select1.ts_tree.character <- function(tree, node, slt) {
  chdn <- tree$dom_children[[node]]
  if (!is_named(chdn)) {
    return(integer())
  }
  chdn[names(chdn) %in% slt]
}

#' @rdname ts_tree_select1
#' @details
#' ## `integer` (integer vector) selector
#'
#' Method: `ts_tree_select1.ts_tree.integer`
#'
#' This method is used when the selector is an integer vector.
#' The default implementation selects DOM children of `node` by position.
#' Positive indices count from the start, negative indices count from the
#' end. Zero indices are not allowed and an error is raised if any are
#' used.
#' @export

ts_tree_select1.ts_tree.integer <- function(tree, node, slt) {
  if (any(slt == 0)) {
    stop(ts_cnd("Zero indices are not allowed in tsitter selectors."))
  }
  chdn <- tree$dom_children[[node]]
  slt <- slt[slt <= length(chdn) & slt >= -length(chdn)]

  res <- integer(length(slt))
  pos <- slt >= 0
  if (any(pos)) {
    res[pos] <- chdn[slt[pos]]
  }
  if (any(!pos)) {
    res[!pos] <- rev(rev(chdn)[abs(slt[!pos])])
  }
  res
}

#' @rdname ts_tree_select1
#' @details
#' ## `numeric` (numeric, double vector) selector
#'
#' Method: `ts_tree_select1.ts_tree.numeric`
#'
#' This method is used when the selector is a numeric (double) vector.
#' It currrently coerces the numeric vector to integer and calls the
#' integer method.
#' @export

ts_tree_select1.ts_tree.numeric <- function(tree, node, slt) {
  ts_tree_select1.ts_tree.integer(tree, node, as.integer(slt))
}

#' @rdname ts_tree_select1
#' @details
#' ## `ts_tree_selector_regex` (regular expression) selector
#'
#' Method: `ts_tree_select1.ts_tree.ts_tree_selector_regex`
#'
#' This method is used when the selector is a regular expression.
#' The default implementation selects DOM children of `node` whose names
#' match the regular expression. If not all children o `node` are named,
#' it returns an empty integer vector. (E.g. in a JSONC document it
#' returns an empty integer vector when nodes is an array.)
#' @export

ts_tree_select1.ts_tree.ts_tree_selector_regex <- function(tree, node, slt) {
  chdn <- tree$dom_children[[node]]
  if (!is_named(chdn)) {
    return(integer())
  }
  chdn[grepl(slt$pattern, names(chdn))]
}

#' @rdname ts_tree_select1
#' @details
#' ## `logical` (logical vector) selector
#'
#' Method: `ts_tree_select1.ts_tree.logical`
#'
#' This method is used when the selector is a logical vector.
#' The default implementation only supports scalar `TRUE`, which selects
#' all DOM children of `node`. Other values raise an error.
#' @export

ts_tree_select1.ts_tree.logical <- function(tree, node, slt) {
  if (isTRUE(slt)) {
    tree$dom_children[[node]]
  } else {
    stop(ts_cnd(
      "Invalid logical selector in `ts_tree_select()`: only scalar `TRUE` is \\
       supported."
    ))
  }
}

#' Unserialize parts of a tree-sitter tree
#'
#' @description
#' The `[[` operator works similarly to the combination of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
#' \code{\link[tsitter:ts_tree_unserialize]{ts_tree_unserialize()}}, but it
#' might be more readable.
#'
#' @details
#' The following two expressions are equivalent:
#' ```r
#' ts_tree_unserialize(ts_tree_select(tree, <selectors>))
#' ```
#' and
#' ```r
#' tree[[list(<selectors>)]]
#' ```
#'
#' ## The `[[<-` replacement operator
#'
#' The `[[<-` operator works similarly to the combination of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
#' \code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
#' replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
#' but it might be more readable.
#'
#' @param x
#' A `ts_tree` object.
#' @param i
#' Selection expressions in a list, see details in
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}}.
#' @param ...
#' Additional arguments, passed to
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}}.
#'
#' @return
#' List of R objects, with one entry for each selected element.
#'
#' @family ts_tree generics
#' @family serialization functions
#' @rdname double-bracket-ts-tree
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')
#'
#' tree
#'
#' tree[[list("a")]]
#'
#' # Last two elements of "b"
#' tree[[list("b", -(1:2))]]

`[[.ts_tree` <- function(x, i, ...) {
  if (missing(i)) {
    i <- list()
  }
  ts_tree_unserialize(ts_tree_select(x, i, ...))
}

select_query <- function(tree, query, captures = NULL) {
  mch <- ts_tree_query(tree, query)

  if (!is.null(captures)) {
    bad <- !captures %in% mch$captures$name
    if (any(bad)) {
      stop(ts_cnd(
        "Invalid capture names in `select_query()`: \\
         {ts_collapse(captures[bad])}."
      ))
    }
    mc <- mch$matched_captures[
      mch$matched_captures$name %in% captures,
    ]
  } else {
    mc <- mch$matched_captures
  }

  ids <- if (nrow(mc) == 0) {
    integer()
  } else {
    toml0 <- tree[
      tree$start_byte %in% mc$start_byte & tree$end_byte %in% mc$end_byte,
    ]
    mkeys <- paste0(mc$type, ":", mc$start_byte, ":", mc$end_byte)
    jkeys <- paste0(toml0$type, ":", toml0$start_byte, ":", toml0$end_byte)
    toml0$id[match(mkeys, jkeys)]
  }
  # TODO: should we do this?
  minimize_selection(tree, ids)
}

# remove nodes that are in the subtree of other selected nodes
# start from the last nodes and go up

minimize_selection <- function(tree, ids) {
  ids <- sort(unique(ids))
  sel <- logical(nrow(tree))
  sel[ids] <- TRUE
  for (id in rev(ids)) {
    parent <- tree$parent[id]
    while (!is.na(parent)) {
      if (sel[parent]) {
        sel[id] <- FALSE
        break
      }
      parent <- tree$parent[parent]
    }
  }
  which(sel)
}

# A section is a list of records. Each record has a selector
# and a list of selected nodes.
#
# 1. there is an explicit selection
# 2. otherwise the top element is selected (or elements if many)
# 3. otherwise the document node is selected

#' Helper functions for tree-sitter tree selections (internal)
#'
#' These functions are for packages implementing new parsers based on the ts
#' package. It is very unlikely that you will need to call these functions
#' directly.
#'
#' @param tree A `ts_tree` object as returned by [ts_tree_new()].
#' @param default Logical, whether to return the default selection if there
#'   is no explicit selection, or `NULL`.
#' @return `ts_tree_selection()` returns a list of selection records.
#'
#' @details `ts_tree_selection()` returns the current selection, as a list
#'   of selectors.
#'
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')
#' tree <- ts_tree_select(tree, "b", -1)
#' ts_tree_selection(tree)

ts_tree_selection <- function(tree, default = TRUE) {
  sel <- attr(tree, "selection")
  if (!is.null(sel)) {
    sel
  } else if (default) {
    ts_tree_selector_default(tree)
  } else {
    NULL
  }
}

#' @rdname ts_tree_selection
#' @return `ts_tree_selected_nodes()` returns the ids of the currently
#'   selected nodes.
#' @details `ts_tree_selected_nodes()` returns the ids of the currently
#'   selected nodes.
#' @export

ts_tree_selected_nodes <- function(tree, default = TRUE) {
  sel <- ts_tree_selection(tree, default = default)
  if (is.null(sel)) {
    return(integer())
  } else {
    sel[[length(sel)]]$nodes
  }
}

ts_tree_selector_default <- function(tree) {
  slt <- structure(
    list(),
    class = c("ts_tree_selector_default", "ts_tree_selector", "list")
  )
  list(list(
    selector = slt,
    nodes = ts_tree_select1(tree, NULL, slt)
  ))
}

#' Edit parts of a tree-sitter tree
#'
#' @description
#' The \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}} replacement
#' function works similarly to the combination of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
#' \code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, but
#' it might be more readable.
#'
#' @details
#' The following two expressions are equivalent:
#' ```r
#' tree <- ts_tree_update(ts_tree_select(tree, <selectors>), value)
#' ```
#' and
#' ```r
#' ts_tree_select(tree, <selectors>) <- value
#' ```
#'
#' @param tree
#' A `ts_tree` object as returned by
#' \code{\link[tsitter:ts_tree_new]{ts_tree_new()}}.
#' @param ...
#' Selection expressions, see \code{\link[tsitter:ts_tree_select]{ts_tree_select()}}.
#'
#' @param value
#' An R expression to serialize or
#' \code{\link[tsitter:ts_tree_deleted]{ts_tree_deleted()}}.
#'
#' @return
#' A `ts_tree` object with the selected parts updated.
#'
#' @name select-set
#' @rdname select-set
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')
#
#' tree
#'
#' ts_tree_select(tree, "a") <- 42
#' ts_tree_select(tree, "b", -1) <- ts_tree_deleted()
#'
#' tree

`ts_tree_select<-` <- function(tree, ..., value) {
  UseMethod("ts_tree_select<-")
}

#' @export

`ts_tree_select<-.ts_tree` <- function(tree, ..., value) {
  res <- if (inherits(value, "ts_tree")) {
    value # nocov
  } else if (inherits(value, "ts_tree_action_delete")) {
    ts_tree_delete(ts_tree_select(tree, ..., refine = TRUE))
  } else {
    ts_tree_update(ts_tree_select(tree, ..., refine = TRUE), value)
  }
  attr(res, "selection") <- NULL
  res
}

#' Edit parts of a tree-sitter tree
#'
#' @description
#' The `[[<-` operator works similarly to the combination of
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
#' \code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
#' replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
#' but it might be more readable.
#'
#' @details
#' The following two expressions are equivalent:
#' ```r
#' tree <- ts_tree_update(ts_tree_select(tree, <selectors>), value)
#' ```
#' and
#' ```r
#' tree[[list(<selectors>)]] <- value
#' ```
#'
#' @param x
#' A `ts_tree` object.
#' @param i
#' A list with selection expressions, see
#' \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} for details.
#' @param value
#' An R expression to serialize or
#' \code{\link[tsitter:ts_tree_deleted]{ts_tree_deleted()}}.
#'
#' @return
#' The modified `ts_tree` object.
#' @rdname double-bracket-set-ts-tree
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')
#'
#' tree
#'
#' tree[[list("a")]] <- 42
#' tree[[list("b", -1)]] <- ts_tree_deleted()
#'
#' tree

`[[<-.ts_tree` <- function(x, i, value) {
  # nocov start -- not sure if still is still needed, if [[ are not nested
  if (missing(i)) {
    i <- list()
  }
  # nocov end
  res <- if (inherits(value, "ts_tree")) {
    value # nocov
  } else if (inherits(value, "ts_tree_action_delete")) {
    ts_tree_delete(ts_tree_select(x, i, refine = TRUE))
  } else {
    ts_tree_update(ts_tree_select(x, i, refine = TRUE), value)
  }
  attr(res, "selection") <- NULL
  res
}

#' @rdname select-set
#' @usage NULL
#' @details
#' `ts_tree_deleted()` is a special marker to delete elements from a
#' ts_tree object with `ts_tree_select<-` or the double bracket operator.
#'
#' @return `ts_tree_deleted()` returns a marker object to be used at the right
#'   hand side of the `ts_tree_select<-` or the double bracket replacement
#'   functions, see examples below.
#'
#' @export

ts_tree_deleted <- function() {
  structure(
    list(),
    class = c("ts_tree_action_delete", "ts_tree_action", "list")
  )
}

test_that("ts_tree_write", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tmp <- file.path(tempfile(), "test.jsonc")
  on.exit(unlink(dirname(tmp), recursive = TRUE), add = TRUE)
  dir.create(dirname(tmp), showWarnings = FALSE, recursive = TRUE)
  writeLines('{"a": 1, "b": [true, false]}', con = tmp)

  expect_snapshot({
    tree <- tsjsonc::ts_read_jsonc(tmp)
    tree %>% ts_tree_select("b", 2) %>% ts_tree_delete() %>% ts_tree_write()
    tsjsonc::ts_read_jsonc(tmp)
  })

  # error if no file
  expect_snapshot(
    error = TRUE,
    {
      tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [true, false]}')
      tree %>% ts_tree_write()
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )

  # append \n if needed
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [true, false]}')
    tree %>% ts_tree_write(file = tmp)
    readChar(tmp, nchars = 1000)
  })

  # write to connection
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [true, false]}')
  con <- textConnection(NULL, "w")
  on.exit(close(con), add = TRUE)
  tree %>% ts_tree_write(con)
  expect_snapshot({
    textConnectionValue(con)
  })

  # write to binary connection
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [true, false]}')
  con2 <- rawConnection(raw(0), "wb")
  on.exit(close(con2), add = TRUE)
  tree %>% ts_tree_write(con2)
  expect_snapshot({
    rawToChar(rawConnectionValue(con2))
  })
})

test_that("ts_tree_delete", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
    tree %>% ts_tree_select("a") %>% ts_tree_delete()
    tree %>% ts_tree_select("b", -1L) %>% ts_tree_delete()
  })
})

test_that("ts_tree_select", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 0, "b": [1,2,3]}')
  expect_snapshot({
    tree %>% ts_tree_select("a") %>% ts_tree_unserialize()
    tree %>% ts_tree_select("b", 2) %>% ts_tree_unserialize()
    tree %>% ts_tree_select("b", -1L) %>% ts_tree_unserialize()
    tree %>% ts_tree_select("b", c(1L, 3L)) %>% ts_tree_unserialize()
    tree %>% ts_tree_select("b", TRUE) %>% ts_tree_unserialize()
  })
})

test_that("ts_tree_select NULL", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 0, "b": [1,2,3]}')
  expect_snapshot({
    tree %>% ts_tree_select(NULL)
    tree %>% ts_tree_select("a") %>% ts_tree_select(NULL, refine = TRUE)
  })
})

test_that("ts_tree_select refine", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 0, "b": [1,2,3]}')
  expect_snapshot({
    tree %>%
      ts_tree_select("b") %>%
      ts_tree_select(1, refine = TRUE) %>%
      ts_tree_unserialize()
  })
})

test_that("ts_tree_select regex", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"apple": 1, "banana": 2, "apricot": 3}')
  expect_snapshot({
    tree %>% ts_tree_select(regex = "^a.*") %>% ts_tree_unserialize()
  })
})

test_that("normalize_selectors", {
  expect_snapshot({
    normalize_selectors(NULL, list("a", 1L, TRUE, NULL, I(1), regex = "b"))
  })
})

test_that("ts_tree_select ids", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}') %>% ts_tree_format()
  numbers <- which(tree$type == "number")
  expect_snapshot({
    tree %>% ts_tree_select(I(numbers)) %>% ts_tree_unserialize()
  })
})

test_that("ts_tree_select unknown selector", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
  expect_snapshot(
    error = TRUE,
    {
      tree %>% ts_tree_select(raw(2))
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("ts_tree_select NULL", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
  expect_snapshot({
    tree %>% ts_tree_select("a", NULL)
  })
})

test_that("ts_tree_select keys from non-object", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{ "a": [1,2,3] }')
  expect_snapshot({
    tree %>% ts_tree_select("a", "b")
  })
  expect_snapshot({
    tree %>% ts_tree_select("a", regex = "^b")
  })
})

test_that("ts_tree_select zero index", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
  expect_snapshot(
    error = TRUE,
    {
      tree %>% ts_tree_select("b", 0)
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("ts_tree_select invalid logical selector", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
  expect_snapshot(
    error = TRUE,
    {
      tree %>% ts_tree_select("b", c(TRUE, FALSE))
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )
})

test_that("[[", {
  skip_if_not_installed("tsjsonc")
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [true, false]}')
  expect_snapshot({
    tree[["a"]]
    tree[[list("b", 2)]]
  })
  expect_snapshot({
    tree[[]]
  })
})


test_that("TS query", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc(
    '{"a": {"b": [1,2,3]}, "c": {"b": [4,5,6]}}'
  ) %>%
    ts_tree_format()
  expect_snapshot({
    tree %>% ts_tree_select(query = "(number) @num")
    tree %>% ts_tree_select(query = "(number) @num") %>% ts_tree_unserialize()
  })

  expect_snapshot({
    tree %>% ts_tree_select(query = "(null) @foo")
  })

  # invalid capture name
  expect_snapshot(
    error = TRUE,
    {
      tree %>% ts_tree_select(query = list("(null) @foo", "bar"))
    },
    variant = if (getRversion() <= "3.5.100") "old" else "new"
  )

  # query with a captures parameter
  expect_snapshot({
    tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": 2, "c": 3, "d": 4 }') %>%
      ts_tree_format()
    ts_tree_sexpr(tree)
    tree %>%
      ts_tree_select(
        query = list(
          "((pair (string (string_content) @key) (number) @num)
           (#not-eq? @key \"c\") )",
          "num"
        )
      )
  })
})

test_that("ts_tree_select<-", {
  skip_if_not_installed("tsjsonc")
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
  expect_snapshot({
    ts_tree_select(tree, "a") <- 2
    tree
    ts_tree_select(tree, "b", 2) <- 99
    tree
  })
})

test_that("ts_tree_delete", {
  skip_if_not_installed("tsjsonc")
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
  expect_snapshot({
    ts_tree_select(tree, "a") <- ts_tree_deleted()
    tree
    tree[[list("b", -1L)]] <- ts_tree_deleted()
    tree
  })
})

test_that("ts_tree_select<- can insert", {
  skip_if_not_installed("tsjsonc")
  library(magrittr)
  tree <- tsjsonc::ts_parse_jsonc('{"a": 1, "b": [1,2,3]}')
  expect_snapshot({
    ts_tree_select(tree, "x") <- 42
    tree
    tree[[list("y", "z")]] <- TRUE
    tree %>% print(n = 100)
  })
})

test_that("minimize_selection", {
  skip_if_not_installed("tsjsonc")
  tree <- tsjsonc::ts_parse_jsonc('{"a": null, "b": [1,2,3]}')
  arr <- which(tree$type == "array")
  num <- which(tree$type == "number")
  expect_equal(
    minimize_selection(tree, c(arr, num)),
    arr
  )
})

#' Run tree-sitter queries on tree-sitter trees
#'
#' @description
#' Use \href{https://tree-sitter.github.io/tree-sitter/}{
#' tree-sitter's query language} to find nodes in a tree-sitter tree.
#'
#' @details
#' You probably need to know some details about the specific tree-sitter
#' parser you are using, to write effective queries. See the documentation
#' of the parser package you are using for details about the node types
#' and the query language support. See links below.
#'
#' @param tree
#'  A `ts_tree` object.
#' @param query
#' Character string, the tree-sitter query to run.
#'
#' @return
#' A list with entries `patterns` and `matched_captures`.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' `patterns` contains information about all patterns in the queries and
#' it is a data frame with columns: `id`, `name`, `pattern`, `match_count`.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' `matched_captures` contains information about all matches, and it has
#' columns `id`, `pattern`, `match`, `start_byte`, `end_byte`, `start_row`,
#' `start_column`, `end_row`, `end_column`, `name`, `code`.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' The `pattern` column of `matched_captured` refers to the `id` column of
#' `patterns`.
#'
#' @seealso [ts_tree_select()] to select the nodes matching a query.
#' @export
#' @family ts_tree exploration
#' @family ts_tree generics
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Select all numbers in a JSONC document ------------------------------------
#' json <- tsjsonc::ts_parse_jsonc(
#'   '{ "a": 1, "b": [10, 20, 30], "c": { "c1": true, "c2": 100 } }'
#' )
#' ts_tree_query(json, "(number) @number")

ts_tree_query <- function(tree, query) {
  UseMethod("ts_tree_query")
}

#' @export

ts_tree_query.default <- function(tree, query) {
  qlen <- nchar(query, type = "bytes") + 1L # + \n
  qbeg <- c(1L, cumsum(qlen))
  qnms <- names(query) %||% rep(NA_character_, length(query))
  query1 <- paste0(query, "\n", collapse = "")

  res <- call_with_cleanup(c_code_query, tree, query1)

  qorig <- as.integer(cut(res[[1]][[3]], breaks = qbeg, include.lowest = TRUE))
  list(
    patterns = data_frame(
      id = seq_along(res[[1]][[1]]),
      name = qnms[qorig],
      pattern = res[[1]][[1]],
      match_count = res[[1]][[2]]
    ),
    captures = data_frame(
      id = seq_along(res[[2]]),
      name = res[[2]]
    ),
    matched_captures = data_frame(
      id = map_int(res[[3]], "[[", 3L),
      pattern = map_int(res[[3]], "[[", 1L),
      match = map_int(res[[3]], "[[", 2L),
      type = map_chr(res[[3]], "[[", 12L),
      start_byte = map_int(res[[3]], "[[", 6L),
      end_byte = map_int(res[[3]], "[[", 7L),
      start_row = map_int(res[[3]], "[[", 8L),
      start_column = map_int(res[[3]], "[[", 9L),
      end_row = map_int(res[[3]], "[[", 10L),
      end_column = map_int(res[[3]], "[[", 11L),
      name = map_chr(res[[3]], "[[", 4L),
      code = map_chr(res[[3]], "[[", 5L)
    )
  )
}

#' Delete selected elements from a tree-sitter tree
#'
#' @description
#' Use \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} to select the
#' elements to be deleted, and then call `ts_tree_delete()` to remove them
#' from the tree.
#'
#' @details
#' The formatting of the rest of the document is left as is.
#'
#' If the tree does not have a selection, the tree corresponding to the
#' empty document is returned, i.e. the whole content is deleted.
#'
#' If the tree has a selection, but it is the empty selection, then
#' the tree is returned unchanged.
#'
#' For parsers that support comments, deleting elements that include
#' comments typically delete the comments as well. Other comments are
#' kept as is. See details in the manual of the specific parser.
#'
#' @param tree
#' A `ts_tree` object.
#' @param ... Extra arguments for methods.
#'
#' @return
#' The modified `ts_tree` object with the selected elements removed.
#'
#' @family ts_tree generics
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc(
#'   "{ \"a\": //comment\ntrue, \"b\": [1, 2, 3] }"
#' )
#'
#' tree
#'
#' ts_tree_select(tree, "a")
#'
#' ts_tree_delete(ts_tree_select(tree, "a"))
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#' # Create a parse tree with tstoml --------------------------------------
#' tree <- tstoml::ts_parse_toml(r"(
#'   [servers]
#'   alpha = { ip = "127.0.0.1", dc = "eqdc10" }
#'   beta = { ip = "127.0.0.2", dc = "eqdc20" }
#' )")
#'
#' tree
#'
#' ts_tree_select(tree, "servers", TRUE, "dc")
#' ts_tree_delete(ts_tree_select(tree, "servers", TRUE, "dc"))

ts_tree_delete <- function(tree, ...) {
  UseMethod("ts_tree_delete")
}

#' Show the syntax tree of a tree-sitter tree
#'
#' @description
#' Show the structure of a tree-sitter tree as an S-expression.
#'
#' @details
#' This function returns a nested list representation of the syntax tree,
#' where each node is represented as a list with its type and children.
#'
#' @param tree
#' A `ts_tree` object.
#'
#' @return
#' A string representing the S-expression of the syntax tree.
#'
#' @export
#' @family ts_tree exploration
#' @family ts_tree generics
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc(
#'   "{ \"a\": //comment\ntrue, \"b\": [1, 2, 3] }"
#' )
#'
#' ts_tree_sexpr(tree)
#'
#' @examplesIf requireNamespace("tstoml", quietly = TRUE)
#'
#' # Create a parse tree with tstoml --------------------------------------
#' tree <- tstoml::ts_parse_toml(r"(
#'   [servers]
#'   alpha = { ip = "127.0.0.1", dc = "eqdc10" }
#'   beta = { ip = "127.0.0.2", dc = "eqdc20" }
#' )")
#'
#' ts_tree_sexpr(tree)

ts_tree_sexpr <- function(tree) {
  UseMethod("ts_tree_sexpr")
}

#' @export

ts_tree_sexpr.default <- function(tree) {
  call_with_cleanup(c_s_expr, tree)
}

#' Write a tree-sitter tree to a file
#'
#' @description
#' Writes the document of a ts `ts_tree` object to a file or connection.
#'
#' @details
#' If `tree` was created from a file, then `ts_tree_write()` by default
#' writes it back to the same file. Otherwise, the `file` argument must be
#' specified.
#'
#' To write to a connection, pass a connection object to the `file`
#' argument. If the connection is opened in binary mode, the raw bytes
#' are written using [base::writeBin()]. Otherwise, the raw bytes are
#' converted to characters using the system encoding before writing using
#' [base::rawToChar()].
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' Use `file = stdout()` to write to the standard output, i.e. to the
#' console in an interactive R session.
#'
#' @param tree
#'  A `ts_tree` object as returned by
#' \code{\link[tsitter:ts_tree_new]{ts_tree_new()}}.
#' @param file
#' Character string, connection, or `NULL`. The file or connection
#' to write to. By default it writes to the same file that was used in
#' \code{\link[tsitter:ts_tree_new]{ts_tree_new()}}, if `tree` was read from a
#' file.
#'
#' @return
#' Invisibly returns `NULL`.
#'
#' @family ts_tree generics
#' @export
#' @examplesIf Sys.getenv("IN_PKGDOWN") == "true"
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')
#'
#' # Format and write to file
#' ts_tree_write(ts_tree_format(tree), "example.json")

ts_tree_write <- function(tree, file = NULL) {
  UseMethod("ts_tree_write")
}

#' @export

ts_tree_write.default <- function(tree, file = NULL) {
  file <- file %||% attr(tree, "file")
  if (is.null(file)) {
    lang <- toupper(get_tree_lang(tree))
    stop(ts_cnd(
      "Don't know which file to save {lang} document to. You need to \\
       specify the `file` argument."
    ))
  }

  text <- attr(tree, "text")
  if (length(text) > 0 && text[length(text)] != 0xa) {
    text <- c(text, as.raw(0xa))
  }
  if (inherits(file, "connection")) {
    if (grepl("b", summary(file)$mode)) {
      writeBin(text, con = file)
    } else {
      cat(rawToChar(text), file = file)
    }
  } else {
    writeBin(text, con = file)
  }
  invisible()
}

#' Convert ts_tree object to a data frame
#'
#' @description
#' Create a data frame for the syntax tree of a JSON document, by indexing
#' a ts_tree object with single brackets. This is occasionally useful for
#' exploration and debugging.
#'
#' @details
#' A tree-sitter tree object has at least four classes:
#' * `ts_tree_<parser_name>`, e.g. `ts_tree_tsjsonc`,
#' * `ts_tree`,
#' * `tbl`, from the pillar package, for better printing when converted
#'   to a data frame, and
#' * `data.frame`, since it is a data frame internally.
#'
#' The `ts_tree` class has custom [format()] and [print()] methods, that
#' show (part of) the underlying document, and also the selected elements,
#' if any.
#'
#' </p><p><!-- ------------------------------------------------------>
#'
#' It is sometimes useful to treat a `tree` `ts_tree` object as a data
#' frame, and drop the `ts_tree` classes. This can be done by indexing with
#' single brackets, e.g. `tree[]`. This returns a data frame with one
#' row per token, and various columns with information about the tokens.
#' See details in the 'Value' section or this page.
#'
#' @param x
#' A `ts_tree` object.
#' @param i,j
#' Incides, passed to the regular data.frame indexing method, see
#' \code{\link[base:Extract]{Extract}}.
#' @param drop
#' Passed to the regular data.frame indexing method, see
#' \code{\link[base:Extract]{Extract}}.
#'
#' @return
#' A data frame with one row per token, and columns:
#' * `id`: integer, the id of the token. The (root) document node has id 1.
#' * `parent`: integer, the id of the parent token. The root token has
#'   parent `NA`
#' * `field_name`: character, the field name of the token in its parent.
#' * `type`: character, the type of the token.
#' * `code`: character, the actual code of the token.
#' * `start_byte`, `end_byte`: integer, the byte positions of the token
#'   in the input.
#' * `start_row`, `start_column`, `end_row`, `end_column`: integer, the
#'   position of the token in the input.
#' * `is_missing`: logical, whether the token is a missing token added by
#'   the parser to recover from errors.
#' * `has_error`: logical, whether the token has a parse error.
#' * `children`: list of integer vectors, the ids of the children tokens.
#' * `dom_type`: character, the type of the node in the DOM tree. See
#'   \code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}. Nodes that are not part
#'   of the DOM tree have `NA_character_` here.
#' * `dom_children`: list of integer vectors, the ids of the children in the
#'   DOM tree. See \code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}.
#' * `dom_parent`: integer, the parent of the node in the DOM tree. See
#'   \code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}. Nodes that are not part
#'   of the DOM tree and the document node have have `NA_integer_` here.
#'
#' Other, undocumented columns may also be present, these are considered
#' internal and may change without notice.
#'
#' @rdname ts_tree-brackets
#' @family ts_tree exploration
#' @export
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')
#'
#' tree
#'
#' tree[]

`[.ts_tree` <- function(x, i, j, drop = FALSE) {
  class(x) <- setdiff(class(x), "ts_tree")
  requireNamespace("pillar", quietly = TRUE)
  NextMethod("[")
}

#' @export

`$.ts_tree` <- function(x, name) {
  class(x) <- setdiff(class(x), "ts_tree")
  NextMethod(`$`)
}

redact_tempfile <- function(x) {
  x <- sub(normalizePath(tempdir()), "<tempdir>", x, fixed = TRUE)
  x <- sub(tempdir(), "<tempdir>", x, fixed = TRUE)
  x <- sub(
    normalizePath(tempdir(), winslash = "/"),
    "<tempdir>",
    x,
    fixed = TRUE
  )
  x <- sub("\\R\\", "/R/", x, fixed = TRUE)
  x <- sub("[\\\\/]file[a-zA-Z0-9]+", "/<tempfile>", x)
  x <- sub("[A-Z]:.*Rtmp[a-zA-Z0-9]+[\\\\/]", "<tempdir>/", x)
  x
}

skip_if_not_installed <- function(pkg) {
  if (
    Sys.getenv("NOT_CRAN") != "true" &&
      tolower(Sys.getenv("_R_CHECK_FORCE_SUGGESTS_")) == "false" &&
      !requireNamespace(pkg, quietly = TRUE)
  ) {
    testthat::skip(paste0(pkg, " is not installed"))
  }
}

dglue <- function(..., .envir = parent.frame()) {
  glue(..., .open = "<<", .close = ">>", .envir = .envir)
}

is_rcmd_check <- function() {
  if (Sys.getenv("GITHUB_ACTIONS") == "true") {
    return(FALSE)
  }
  Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") != "" ||
    Sys.getenv("_R_RD_MACROS_PACKAGE_DIR_", "") != ""
}

doc_has_method <- function(method, package) {
  method <- paste0(method, ".ts_tree_", sub("^ts", "", package))
  length(utils::help(
    topic = (method),
    package = (package),
    help_type = "text"
  )) >
    0
}

doc_seealso <- function(method) {
  ts_package <- get_env("R_TS_PACKAGE")
  if (is.null(ts_package)) {
    psrs <- ts_list_parsers()
    psrs <- psrs[!duplicated(psrs$package), ]
    links <- vapply(
      psrs$package,
      function(pkg) {
        if (doc_has_method(method, pkg)) {
          glue(
            "\\code{{\\link[{pkg}:{method}.{pkg}]{{{method}(<ts_tree_{pkg}>)}}}}",
          )
        } else {
          ""
        }
      },
      ""
    )
    links <- links[links != ""]
    if (length(links) > 0) {
      s <- if (length(links) > 1) "s"
      glue("Method{s} in installed package{s}: {ts_collapse(links)}.")
    } else {
      ""
    }
  } else {
    paste0(
      "The generic of this method in the tsitter package: ",
      glue("\\code{{\\link[tsitter:{method}]{{{method}()}}}}"),
      "."
    )
  }
}

doc_insert <- function(key, manpkg = NULL) {
  if (is_rcmd_check()) {
    return("Placeholder.")
  }
  if (!is.null(manpkg)) {
    Sys.setenv("R_TS_PACKAGE" = manpkg)
    on.exit(Sys.unsetenv("R_TS_PACKAGE"), add = TRUE)
  }
  keypcs <- strsplit(key, "::", fixed = TRUE)[[1]]
  if (length(keypcs) == 2) {
    package <- keypcs[1]
    key <- keypcs[2]
  } else {
    package <- "tsitter"
  }
  lib <- dirname(find.package(package))
  output <- doc_create_chunk(key, lib, package, 1L, "<<contents>>")

  mch <- gregexpr("\\\\eval\\{[^\\}]+\\}", output, perl = TRUE)
  regmatches(output, mch)[[1]] <- lapply(
    regmatches(output, mch)[[1]],
    function(x) {
      x <- sub("^\\\\eval\\{", "", x)
      x <- sub("\\}$", "", x)
      eval(parse(text = x))
    }
  )
  output
}

doc_tabs <- function(key) {
  if (is_rcmd_check()) {
    return("Placeholder.")
  }
  package <- if (nzchar(ev <- Sys.getenv("R_TS_PACKAGE"))) {
    ev
  }

  if (is.null(package)) {
    doc_tabs_all(key)
  } else {
    doc_tabs_one(key, package)
  }
}

doc_tabs_all <- function(key) {
  # list all installed ts packages
  psrs <- ts_list_parsers()
  psrs <- psrs[!duplicated(psrs$package), ]

  tsdocpath <- doc_path("tsitter")
  t_tab <- read_char(file.path(tsdocpath, "tab.html"))
  t_div <- read_char(file.path(tsdocpath, "tabs.html"))
  t_btn <- read_char(file.path(tsdocpath, "btn.html"))
  output <- buttons <- tabs <- ""

  for (i in seq_len(nrow(psrs))) {
    language <- sub("^ts", "", psrs$package[i])
    tab <- doc_create_chunk(key, psrs$library[i], psrs$package[i], i, t_tab)
    btn <- dglue(
      t_btn,
      .envir = c(psrs[i, ], list(idx = i, language = language))
    )
    if (!is.null(tab) && nzchar(tab)) {
      tabs <- paste0(tabs, tab, "\n")
      buttons <- paste0(buttons, btn, "\n")
    }
  }
  if (tabs != "") {
    output <- dglue(t_div, .envir = list(tabs = tabs, buttons = buttons))
  }

  output
}

doc_tabs_one <- function(key, package) {
  lib <- dirname(find.package(package))
  doc_create_chunk(key, lib, package, 1, "<<contents>>")
}

doc_path <- function(package) {
  pkgdir <- find.package(package)
  docpath <- file.path(pkgdir, "tsdocs")
  if (!file.exists(docpath)) {
    docpath <- file.path(pkgdir, "inst", "tsdocs")
  }

  docpath
}

doc_create_chunk <- function(key, lib, package, idx, template) {
  file <- paste0(key, ".Rd")
  path <- file.path(doc_path(package), file)
  if (!file.exists(path)) {
    return("")
  }
  x <- read_char(path)
  lns <- strsplit(x, "\n", fixed = TRUE)[[1]]
  rulepos <- which(lns == "# ---")
  lns <- lns[(rulepos + 1):length(lns)]
  lang_data <- list(
    idx = idx,
    package = package,
    language = sub("^ts", "", package),
    contents = paste(lns, collapse = "\n")
  )
  dglue(template, .envir = lang_data)
}

doc_extra <- function() {
  if (is_rcmd_check()) {
    return("Placeholder.")
  }
  tsdocpath <- doc_path("tsitter")
  jspath <- file.path(tsdocpath, "tabs.js")
  js <- read_char(jspath)

  csspath1 <- file.path(doc_path("tsitter"), "w3.css")
  css1 <- read_char(csspath1)
  css <- ""
  if (Sys.getenv("IN_PKGDOWN") == "true") {
    csspath2 <- file.path(doc_path("tsitter"), "pkgdown.css")
    css2 <- read_char(csspath2)
    css <- gsub("%", "\\%", paste0(css1, "\n\n", css2), fixed = TRUE)
    css <- paste0("<style>\n", css, "\n</style>\n")
  } else {
    # in Rd we insert CSS with JS, because tidy does not allow <style>
    # tags in <body>
    js <- paste0(
      "styles = `",
      gsub("%", "\\%", css1, fixed = TRUE),
      "`;\n",
      "var styleSheet = document.createElement('style');\n",
      "styleSheet.textContent = styles;\n",
      "document.head.appendChild(styleSheet);\n",
      js
    )
  }

  dglue("\\if{html}{\\out{<script>\n<<js>>\n</script>\n<<css>>}}")
}

doc_style_tab <- function(idx) {
  paste0(
    "padding: 0.01em 16px;",
    "border: 1px solid #ccc!important;",
    if (idx != 1) {
      "display:none"
    }
  )
}

# ------------------------------------------------------------------------------
# Roxygen2 roclet for generating manual stubs

roxy_tag_parse.roxy_tag_ts <- function(x) {
  lns <- strsplit(x$raw, "\n", fixed = TRUE)[[1]]
  lns1p <- strsplit(lns[1], " ", fixed = TRUE)[[1]]
  file <- lns1p[1]
  title <- paste(lns1p[-1], collapse = " ")

  x$raw <- paste(lns[-1], collapse = "\n")

  pkg <- asNamespace("roxygen2")$roxy_meta_get("current_package")
  asNamespace("roxygen2")$roxy_meta_set("current_package", "x")
  on.exit(
    asNamespace("roxygen2")$roxy_meta_set("current_package", pkg),
    add = TRUE
  )

  x <- asNamespace("roxygen2")$tag_markdown_with_sections(x)
  x$val <- list(
    file = file,
    title = title,
    val = x$val
  )
  x
}

roclet_ts <- function() {
  asNamespace("roxygen2")$roclet("ts")
}

roclet_process.roclet_ts <- function(x, blocks, env, base_path) {
  for (block in blocks) {
    tags <- asNamespace("roxygen2")$block_get_tags(block, "ts")

    for (tag in tags) {
      path <- file.path(
        base_path,
        "inst",
        "tsdocs",
        paste0(tag$val$file, ".Rd")
      )
      value <- paste0(
        "#| title: ",
        tag$val$title,
        "\n# ---\n",
        rd_patch_p(tag$val$val)
      )
      write_if_newer(value, path)
    }
  }
  invisible(list())
}

# Rd2HTML adds a </p> before the \preformatted, this is a hach.
# It might create an un-closed <p>, but that's apparently not a problem.
rd_patch_p <- function(rd) {
  gsub(
    '\\out{<div class="sourceCode r">}}\\preformatted',
    '\\out{<div class="sourceCode r"><p style="display:none">&nbsp;}}\\preformatted',
    rd,
    fixed = TRUE
  )
}

write_if_newer <- function(txt, path) {
  bin <- charToRaw(txt)
  nl <- charToRaw("\n")
  if (length(bin) > 0 && bin[[length(bin)]] != nl) {
    bin <- c(bin, nl)
  }
  if (file.exists(path)) {
    old <- readBin(path, what = "raw", n = file.size(path))
    if (identical(old, bin)) {
      return(invisible(FALSE))
    }
  }
  dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
  cli::cli_alert_info("Manual stub at {.file {basename(path)}}.")
  writeBin(bin, path)
  invisible(TRUE)
}

roclet_output.roclet_ts <- function(x, results, base_path, ...) {
  invisible(NULL)
}

ts_roclet_register <- function() {
  registerS3method(
    "roxy_tag_parse",
    "roxy_tag_ts",
    roxy_tag_parse.roxy_tag_ts,
    asNamespace("roxygen2")
  )
  registerS3method(
    "roclet_process",
    "roclet_ts",
    roclet_process.roclet_ts,
    asNamespace("roxygen2")
  )
  registerS3method(
    "roclet_output",
    "roclet_ts",
    roclet_output.roclet_ts,
    asNamespace("roxygen2")
  )
}

# Compared to glue::glue(), these are fixed:
# - .sep = ""
# - .trim = TRUE
# - .null = character()
# - .literal = TRUE
# - .comment = ""
#
# we also don't allow passing in data as arguments, and `text` is
# a single argument, no need to `paste()` etc.

glue <- function(
  text,
  .envir = parent.frame(),
  .transformer = identity_transformer,
  .open = "{",
  .close = "}",
  .cli = FALSE,
  .trim = TRUE
) {
  text <- paste0(text, collapse = "")

  if (.trim) {
    text <- trim(text)
  }

  f <- function(expr) {
    eval_func <- as.character(.transformer(expr, .envir) %||% character())
  }

  res <- call_with_cleanup(c_glue, text, f, .open, .close, .cli)

  paste0(unlist(res), collapse = "")
}

identity_transformer <- function(text, envir) {
  eval(parse(text = text, keep.source = FALSE), envir)
}

#' Raw bytes of a document of a tree-sitter tree
#'
#' @param x A `ts_tree` object.
#' @return A raw vector containing the bytes of the document of the tree.
#'
#' @export
#' @seealso [as.character.ts_tree()] to get the document as a character scalar.
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')
#'
#' tree
#' as.raw(tree)

as.raw.ts_tree <- function(x) {
  attr(x, "text")
}

#' The document of a tree-sitter tree as a character scalar
#'
#' @param x A `ts_tree` object.
#' @param ... Ignored.
#' @return A character scalar containing the document of the tree.
#'
#' @export
#' @seealso [as.raw.ts_tree()] to get the document as a raw vector.
#' @examplesIf requireNamespace("tsjsonc", quietly = TRUE)
#' # Create a parse tree with tsjsonc -------------------------------------
#' tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')
#'
#' tree
#' as.character(tree)

as.character.ts_tree <- function(x, ...) {
  rawToChar(as.raw(x))
}

# ts_tree_new

    Code
      tree <- ts_tree_new(tsjsonc::ts_language_jsonc(), text = "{\"a\": 1, \"b\": [1,2,3]}")
      tree
    Output
      # jsonc (1 line)
      1 | {"a": 1, "b": [1,2,3]}

# ts_tree_new leading whitespace

    Code
      tree <- ts_tree_new(tsjsonc::ts_language_jsonc(), text = "\n\n   \n{\"a\": 1, \"b\": [1,2,3]}")
      tree
    Output
      # jsonc (4 lines)
      1 | 
      2 | 
      3 |    
      4 | {"a": 1, "b": [1,2,3]}

# ts_tree_select

    Code
      tree %>% ts_tree_select("a") %>% ts_tree_unserialize()
    Output
      [[1]]
      [1] 0
      
    Code
      tree %>% ts_tree_select("b", 2) %>% ts_tree_unserialize()
    Output
      [[1]]
      [1] 2
      
    Code
      tree %>% ts_tree_select("b", -1L) %>% ts_tree_unserialize()
    Output
      [[1]]
      [1] 3
      
    Code
      tree %>% ts_tree_select("b", c(1L, 3L)) %>% ts_tree_unserialize()
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 3
      
    Code
      tree %>% ts_tree_select("b", TRUE) %>% ts_tree_unserialize()
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 2
      
      [[3]]
      [1] 3
      

# ts_tree_select NULL

    Code
      tree %>% ts_tree_select(NULL)
    Output
      # jsonc (1 line)
      1 | {"a": 0, "b": [1,2,3]}
    Code
      tree %>% ts_tree_select("a") %>% ts_tree_select(NULL, refine = TRUE)
    Output
      # jsonc (1 line)
      1 | {"a": 0, "b": [1,2,3]}

---

    Code
      tree %>% ts_tree_select("a", NULL)
    Output
      # jsonc (1 line, 0 selected elements)
      1 | {"a": 1, "b": [1,2,3]}

# ts_tree_select refine

    Code
      tree %>% ts_tree_select("b") %>% ts_tree_select(1, refine = TRUE) %>%
        ts_tree_unserialize()
    Output
      [[1]]
      [1] 1
      

# ts_tree_select regex

    Code
      tree %>% ts_tree_select(regex = "^a.*") %>% ts_tree_unserialize()
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 3
      

# normalize_selectors

    Code
      normalize_selectors(NULL, list("a", 1L, TRUE, NULL, I(1), regex = "b"))
    Output
      [[1]]
      [1] "a"
      
      [[2]]
      [1] 1
      
      [[3]]
      [1] TRUE
      
      [[4]]
      NULL
      
      [[5]]
      $ids
      [1] 1
      
      attr(,"class")
      [1] "ts_tree_selector_ids" "ts_tree_selector"     "list"                
      
      [[6]]
      $pattern
      [1] "b"
      
      attr(,"class")
      [1] "ts_tree_selector_regex" "ts_tree_selector"       "list"                  
      

# ts_tree_select ids

    Code
      tree %>% ts_tree_select(I(numbers)) %>% ts_tree_unserialize()
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 1
      
      [[3]]
      [1] 2
      
      [[4]]
      [1] 3
      

# ts_tree_select keys from non-object

    Code
      tree %>% ts_tree_select("a", "b")
    Output
      # jsonc (1 line, 0 selected elements)
      1 | { "a": [1,2,3] }

---

    Code
      tree %>% ts_tree_select("a", regex = "^b")
    Output
      # jsonc (1 line, 0 selected elements)
      1 | { "a": [1,2,3] }

# [[

    Code
      tree[["a"]]
    Output
      [[1]]
      [1] 1
      
    Code
      tree[[list("b", 2)]]
    Output
      [[1]]
      [1] FALSE
      

---

    Code
      tree[[]]
    Output
      [[1]]
      [[1]]$a
      [1] 1
      
      [[1]]$b
      [[1]]$b[[1]]
      [1] TRUE
      
      [[1]]$b[[2]]
      [1] FALSE
      
      
      

# TS query

    Code
      tree %>% ts_tree_select(query = "(number) @num")
    Output
      # jsonc (16 lines, 6 selected elements)
         1 | {
         2 |     "a": {
         3 |         "b": [
      >  4 |             1,
      >  5 |             2,
      >  6 |             3
         7 |         ]
         8 |     },
         9 |     "c": {
        10 |         "b": [
      > 11 |             4,
      > 12 |             5,
      > 13 |             6
        14 |         ]
        15 |     }
        16 | }
    Code
      tree %>% ts_tree_select(query = "(number) @num") %>% ts_tree_unserialize()
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 2
      
      [[3]]
      [1] 3
      
      [[4]]
      [1] 4
      
      [[5]]
      [1] 5
      
      [[6]]
      [1] 6
      

---

    Code
      tree %>% ts_tree_select(query = "(null) @foo")
    Output
      # jsonc (16 lines, 0 selected elements)
       1 | {
       2 |     "a": {
       3 |         "b": [
       4 |             1,
       5 |             2,
       6 |             3
       7 |         ]
       8 |     },
       9 |     "c": {
      10 |         "b": [
      i 6 more lines
      i Use `print(n = ...)` to see more lines

---

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": 2, \"c\": 3, \"d\": 4 }") %>%
        ts_tree_format()
      ts_tree_sexpr(tree)
    Output
      [1] "(document (object (pair key: (string (string_content)) value: (number)) (pair key: (string (string_content)) value: (number)) (pair key: (string (string_content)) value: (number)) (pair key: (string (string_content)) value: (number))))"
    Code
      tree %>% ts_tree_select(query = list(
        "((pair (string (string_content) @key) (number) @num)\n           (#not-eq? @key \"c\") )",
        "num"))
    Output
      # jsonc (6 lines, 3 selected elements)
        1 | {
      > 2 |     "a": 1,
      > 3 |     "b": 2,
        4 |     "c": 3,
      > 5 |     "d": 4
        6 | }

# ts_tree_select<-

    Code
      ts_tree_select(tree, "a") <- 2
      tree
    Output
      # jsonc (1 line)
      1 | {"a": 2, "b": [1,2,3]}
    Code
      ts_tree_select(tree, "b", 2) <- 99
      tree
    Output
      # jsonc (1 line)
      1 | {"a": 2, "b": [1,99,3]}

# ts_tree_delete

    Code
      ts_tree_select(tree, "a") <- ts_tree_deleted()
      tree
    Output
      # jsonc (1 line)
      1 | {"b": [1,2,3]}
    Code
      tree[[list("b", -1L)]] <- ts_tree_deleted()
      tree
    Output
      # jsonc (1 line)
      1 | {"b": [1,2]}

# ts_tree_select<- can insert

    Code
      ts_tree_select(tree, "x") <- 42
      tree
    Output
      # jsonc (9 lines)
      1 | {
      2 |     "a": 1,
      3 |     "b": [
      4 |         1,
      5 |         2,
      6 |         3
      7 |     ],
      8 |     "x": 42
      9 | }
    Code
      tree[[list("y", "z")]] <- TRUE
      tree %>% print(n = 100)
    Output
      # jsonc (12 lines)
       1 | {
       2 |     "a": 1,
       3 |     "b": [
       4 |         1,
       5 |         2,
       6 |         3
       7 |     ],
       8 |     "x": 42,
       9 |     "y": {
      10 |         "z": true
      11 |     }
      12 | }

# [

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [true, false]}")
      tree
    Output
      # jsonc (1 line)
      1 | {"a": 1, "b": [true, false]}
    Code
      tree[]
    Output
      # A data frame: 24 x 20
            id parent field_name type             code  start_byte end_byte start_row
         <int>  <int> <chr>      <chr>            <chr>      <int>    <int>     <int>
       1     1     NA <NA>       "document"       <NA>           0       28         0
       2     2      1 <NA>       "object"         <NA>           0       28         0
       3     3      2 <NA>       "{"              "{"            0        1         0
       4     4      2 <NA>       "pair"           <NA>           1        7         0
       5     5      4 key        "string"         <NA>           1        4         0
       6     6      5 <NA>       "\""             "\""           1        2         0
       7     7      5 <NA>       "string_content" "a"            2        3         0
       8     8      5 <NA>       "\""             "\""           3        4         0
       9     9      4 <NA>       ":"              ":"            4        5         0
      10    10      4 value      "number"         "1"            6        7         0
      # i 14 more rows
      # i 12 more variables: start_column <int>, end_row <int>, end_column <int>,
      #   is_missing <lgl>, has_error <lgl>, expected <list>, children <I<list>>,
      #   tws <chr>, dom_children <list>, dom_parent <int>, dom_name <chr>,
      #   dom_type <chr>

# get_ranges

    Code
      ts_tree_new(tsjsonc::ts_language_jsonc(), text = txt, ranges = 1:10)
    Condition
      Error:
      ! Invalid ranges, must be a data frame of 6 integer columns
    Code
      ranges2 <- data.frame(a = 0:1, b = 0:1, c = 0:1, d = 0:1, e = 1, f = 0:1)
      ts_tree_new(tsjsonc::ts_language_jsonc(), text = txt, ranges = ranges2)
    Condition
      Error:
      ! Invalid ranges, must be a data frame of 6 integer columns
    Code
      ranges3 <- ranges
      ranges3$start_byte[1] <- 100L
      ts_tree_new(tsjsonc::ts_language_jsonc(), text = txt, ranges = ranges3)
    Condition
      Error:
      ! Invalid ranges for tree-sitter parser

# parse

    Code
      ts_tree_new(structure("jsonc", class = c("ts_language_jsonc", "ts_language")),
      text = "{\"a\": ")
    Condition
      Error:
      ! tsitter `language` must be an external pointer

# ts_tree_write

    Code
      tree <- tsjsonc::ts_read_jsonc(tmp)
      tree %>% ts_tree_select("b", 2) %>% ts_tree_delete() %>% ts_tree_write()
      tsjsonc::ts_read_jsonc(tmp)
    Output
      # jsonc (test.jsonc, 1 line)
      1 | {"a": 1, "b": [true]}

---

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [true, false]}")
      tree %>% ts_tree_write(file = tmp)
      readChar(tmp, nchars = 1000)
    Output
      [1] "{\"a\": 1, \"b\": [true, false]}\n"

---

    Code
      textConnectionValue(con)
    Output
      [1] "{\"a\": 1, \"b\": [true, false]}"

---

    Code
      rawToChar(rawConnectionValue(con2))
    Output
      [1] "{\"a\": 1, \"b\": [true, false]}\n"

# ts_tree_new

    Code
      ts_tree_new(tsjsonc::ts_language_jsonc())
    Condition
      Error in `ts_tree_new.ts_language()`:
      ! Invalid arguments in `ts_tree_new()`: exactly one of `file` and `text` must be given.

---

    Code
      ts_tree_new(tsjsonc::ts_language_jsonc(), text = "{\"a\": 1}", file = tempfile())
    Condition
      Error in `ts_tree_new.ts_language()`:
      ! Invalid arguments in `ts_tree_new()`: exactly one of `file` and `text` must be given.

---

    Code
      tree <- ts_tree_new(tsjsonc::ts_language_jsonc(), file = tmp)
      tree
    Output
      # jsonc (test.jsonc, 1 line)
      1 | {"a": 1, "b": [1,2,3]}

# ts_tree_new parse error

    Code
      ts_tree_new(tsjsonc::ts_language_jsonc(), text = "{\"a\": 1, \"b\": [1,2,3")
    Condition
      Error in `ts_tree_new.ts_language()`:
      ! JSONC parse error `<text>`:1:21
      1| {"a": 1, "b": [1,2,3
                             ^

# cnd

    Code
      do <- (function() ts_cnd("This is a test error with a value: {v}."))
      do()
    Output
      <error in do(): This is a test error with a value: 13.>

# ts_caller_arg

    Code
      do(1)
    Output
      [[1]]
      [1] 1
      
      attr(,"class")
      [1] "ts_caller_arg"
    Code
      do(v)
    Output
      [[1]]
      v
      
      attr(,"class")
      [1] "ts_caller_arg"

# as_ts_caller_arg

    Code
      as_ts_caller_arg("foobar")
    Output
      [[1]]
      [1] "foobar"
      
      attr(,"class")
      [1] "ts_caller_arg"

# as.character.ts_caller_arg

    Code
      do(v1)
    Output
      [1] "v1"
    Code
      do(1 + 1 + 1)
    Output
      [1] "1 + 1 + 1"
    Code
      do(foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo +
        foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo +
        foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo)
    Output
      [1] "foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo_foo + ..."

# format.ts_parse_error, print.ts_parse_error

    Code
      format(err)
    Output
      [1] "<ts_parse_error>"                          
      [2] "JSONC parse error `<text>`:1:29"           
      [3] "1| {\"a\": 1, \"b\": [true, false, error]}"
      [4] "                               ^^^^^^"     
    Code
      print(err)
    Output
      <ts_parse_error>
      JSONC parse error `<text>`:1:29
      1| {"a": 1, "b": [true, false, error]}
                                     ^^^^^^

# ts_tree_dom

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [1,2,3]}")
      ts_tree_dom(tree)
    Output
      document (1)
      \-object (2)
        +-number (10) # a
        \-array (18) # b
          +-number (20)
          +-number (22)
          \-number (24)

# ts_tree_select unknown selector

    Code
      tree %>% ts_tree_select(raw(2))
    Condition
      Error in `ts_tree_select1.default()`:
      ! Don't know how to select nodes from a `ts_tree` (JSONC) object using selector of class `raw`.

# ts_tree_select zero index

    Code
      tree %>% ts_tree_select("b", 0)
    Condition
      Error in `ts_tree_select1.ts_tree.integer()`:
      ! Zero indices are not allowed in tsitter selectors.

# ts_tree_select invalid logical selector

    Code
      tree %>% ts_tree_select("b", c(TRUE, FALSE))
    Condition
      Error in `ts_tree_select1.ts_tree.logical()`:
      ! Invalid logical selector in `ts_tree_select()`: only scalar `TRUE` is supported.

# TS query

    Code
      tree %>% ts_tree_select(query = list("(null) @foo", "bar"))
    Condition
      Error in `select_query()`:
      ! Invalid capture names in `select_query()`: bar.

# ts_tree_sexpr

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [true, false]}")
      ts_tree_sexpr(tree)
    Output
      [1] "(document (object (pair key: (string (string_content)) value: (number)) (pair key: (string (string_content)) value: (array (true) (false)))))"

# ts_check_named_arg

    Code
      f(42)
    Condition
      Error in `f()`:
      ! The `foobar` argument must be fully named.
    Code
      f(fooba = 42)
    Condition
      Error in `f()`:
      ! The `foobar` argument must be fully named.

# ts_parse_error_cnd

    Code
      stop(ts_parse_error_cnd(tree = tree, text = charToRaw(txt)))
    Condition
      Error:
      ! JSONC parse error `<text>`:1:29
      1| {"a": 1, "b": [true, false, error]}
                                     ^^^^^^

# format_ts_parse_error_1

    Code
      stop(err)
    Condition
      Error:
      ! JSONC parse error `<text>`:1:37
      1| {"a": 1...error, ..., "ok"]}
                   ^^^^^^^

---

    Code
      stop(err2)
    Condition
      Error:
      ! JSONC parse error `<text>`:1:37
      1| {"a": 1...loooooooooooo...ooooooong]}
                   ^^^^^^^^^^^^^^^^^^^^^^^^^^

---

    Code
      stop(err3)
    Condition
      Error:
      ! JSONC parse error `<text>`:2:29
      1| {"a": 1, 
      2|  "b": [true, false, "very", bad]
                                     ^^^^
      3| }

---

    Code
      stop(err4)
    Condition
      Error:
      ! JSONC parse error `<text>`:2:30
      2|   "b": ...looooooooooooooooooooooong]
                   ^^^^^^^^^^^^^^^^^^^^^^^^^^^

# get_ranges

    Code
      txt
    Output
      [1] "[xx1,x2,xx3]x"
    Code
      tree <- ts_tree_new(tsjsonc::ts_language_jsonc(), text = txt, ranges = ranges)
      tree
    Output
      # jsonc (1 line)
      1 | [xx1,x2,xx3]x
    Code
      ts_tree_ast(tree)
    Output
      document (1)      1|
      \-array (2)        |
        +-[ (3)          |[
        +-number (4)     |   1
        +-, (5)          |    ,
        +-number (6)     |      2
        +-, (7)          |       ,
        +-number (8)     |          3
        \-] (9)          |           ]

# new_lookahead_sym

    Code
      tree <- ts_tree_new(tsjsonc::ts_language_jsonc(), text = "[{\"a\": }]",
      fail_on_parse_error = FALSE)
      tree[]
    Output
      # A data frame: 14 x 20
            id parent field_name type             code  start_byte end_byte start_row
         <int>  <int> <chr>      <chr>            <chr>      <int>    <int>     <int>
       1     1     NA <NA>       "document"       <NA>           0        9         0
       2     2      1 <NA>       "array"          <NA>           0        9         0
       3     3      2 <NA>       "["              "["            0        1         0
       4     4      2 <NA>       "object"         <NA>           1        8         0
       5     5      4 <NA>       "{"              "{"            1        2         0
       6     6      4 <NA>       "pair"           <NA>           2        6         0
       7     7      6 key        "string"         <NA>           2        5         0
       8     8      7 <NA>       "\""             "\""           2        3         0
       9     9      7 <NA>       "string_content" "a"            3        4         0
      10    10      7 <NA>       "\""             "\""           4        5         0
      11    11      6 <NA>       ":"              ":"            5        6         0
      12    12      6 value      "number"         ""             6        6         0
      13    13      4 <NA>       "}"              "}"            7        8         0
      14    14      2 <NA>       "]"              "]"            8        9         0
      # i 12 more variables: start_column <int>, end_row <int>, end_column <int>,
      #   is_missing <lgl>, has_error <lgl>, expected <list>, children <I<list>>,
      #   tws <chr>, dom_children <list>, dom_parent <int>, dom_name <chr>,
      #   dom_type <chr>
    Code
      tree$is_missing
    Output
       [1] FALSE FALSE FALSE FALSE FALSE FALSE FALSE FALSE FALSE FALSE FALSE  TRUE
      [13] FALSE FALSE
    Code
      exp <- tree$expected[[which(tree$is_missing)]]
      do.call(rbind, exp)
    Output
            symbol name      type
       [1,] 65535  "ERROR"   0   
       [2,] 14     "comment" 0   
       [3,] 1      "{"       1   
       [4,] 5      "["       1   
       [5,] 7      """       1   
       [6,] 16     "_value"  2   
       [7,] 17     "object"  0   
       [8,] 19     "array"   0   
       [9,] 20     "string"  0   
      [10,] 10     "number"  0   
      [11,] 11     "true"    0   
      [12,] 12     "false"   0   
      [13,] 13     "null"    0   

# check_predicate_eq

    Code
      ts_tree_query(tree, "((pair key: (string (string_content) @key)) (#eq? @key \"a\"))")
    Output
      $patterns
      # A data frame: 1 x 4
           id name  pattern                                                            match_count
        <int> <chr> <chr>                                                                    <int>
      1     1 <NA>  "((pair key: (string (string_content) @key)) (#eq? @key \"a\"))\n"           1
      
      $captures
      # A data frame: 1 x 2
           id name 
        <int> <chr>
      1     1 key  
      
      $matched_captures
      # A data frame: 1 x 12
           id pattern match type           start_byte end_byte start_row start_column end_row end_column name  code 
        <int>   <int> <int> <chr>               <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>
      1     1       1     1 string_content          2        3         0            2       0          3 key   a    
      

---

    Code
      ts_tree_query(tree, "((pair key: (string (string_content) @key)) (#not-eq? @key \"a\"))")
    Output
      $patterns
      # A data frame: 1 x 4
           id name  pattern                                                                match_count
        <int> <chr> <chr>                                                                        <int>
      1     1 <NA>  "((pair key: (string (string_content) @key)) (#not-eq? @key \"a\"))\n"           1
      
      $captures
      # A data frame: 1 x 2
           id name 
        <int> <chr>
      1     1 key  
      
      $matched_captures
      # A data frame: 1 x 12
           id pattern match type           start_byte end_byte start_row start_column end_row end_column name  code 
        <int>   <int> <int> <chr>               <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>
      1     1       1     1 string_content         10       11         0           10       0         11 key   b    
      

---

    Code
      ts_tree_query(tree, "((pair key: (string (string_content) @key)\n        value: (string (string_content) @value)) @pair\n        (#eq? @key @value))")
    Output
      $patterns
      # A data frame: 1 x 4
           id name  pattern                                                                                                                             match_count
        <int> <chr> <chr>                                                                                                                                     <int>
      1     1 <NA>  "((pair key: (string (string_content) @key)\n        value: (string (string_content) @value)) @pair\n        (#eq? @key @value))\n"           1
      
      $captures
      # A data frame: 3 x 2
           id name 
        <int> <chr>
      1     1 key  
      2     2 value
      3     3 pair 
      
      $matched_captures
      # A data frame: 3 x 12
           id pattern match type           start_byte end_byte start_row start_column end_row end_column name  code          
        <int>   <int> <int> <chr>               <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>         
      1     3       1     1 pair                    1        9         0            1       0          9 pair  "\"a\": \"a\""
      2     1       1     1 string_content          2        3         0            2       0          3 key   "a"           
      3     2       1     1 string_content          7        8         0            7       0          8 value "a"           
      

---

    Code
      ts_tree_query(tree, "((pair key: (string (string_content) @key)\n        value: (string (string_content) @value))\n        (#not-eq? @key @value))")
    Output
      $patterns
      # A data frame: 1 x 4
           id name  pattern                                                                                                                           match_count
        <int> <chr> <chr>                                                                                                                                   <int>
      1     1 <NA>  "((pair key: (string (string_content) @key)\n        value: (string (string_content) @value))\n        (#not-eq? @key @value))\n"           1
      
      $captures
      # A data frame: 2 x 2
           id name 
        <int> <chr>
      1     1 key  
      2     2 value
      
      $matched_captures
      # A data frame: 2 x 12
           id pattern match type           start_byte end_byte start_row start_column end_row end_column name  code 
        <int>   <int> <int> <chr>               <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>
      1     1       1     1 string_content         12       13         0           12       0         13 key   b    
      2     2       1     1 string_content         17       18         0           17       0         18 value x    
      

# r_grepl

    Code
      ts_tree_query(tree, "((pair key: (string (string_content) @key))\n        (#match? @key \"^a\"))")
    Output
      $patterns
      # A data frame: 1 x 4
           id name  pattern                                                                         match_count
        <int> <chr> <chr>                                                                                 <int>
      1     1 <NA>  "((pair key: (string (string_content) @key))\n        (#match? @key \"^a\"))\n"           1
      
      $captures
      # A data frame: 1 x 2
           id name 
        <int> <chr>
      1     1 key  
      
      $matched_captures
      # A data frame: 1 x 12
           id pattern match type           start_byte end_byte start_row start_column end_row end_column name  code 
        <int>   <int> <int> <chr>               <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>
      1     1       1     1 string_content          2        3         0            2       0          3 key   a    
      

# ts_tree_write

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [true, false]}")
      tree %>% ts_tree_write()
    Condition
      Error in `ts_tree_write.default()`:
      ! Don't know which file to save JSONC document to. You need to specify the `file` argument.

# format.ts_tree

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [1,2,3]}") %>%
        ts_tree_format()
      print(tree)
    Output
      # jsonc (8 lines)
      1 | {
      2 |     "a": 1,
      3 |     "b": [
      4 |         1,
      5 |         2,
      6 |         3
      7 |     ]
      8 | }
    Code
      print(tree, n = 1)
    Output
      # jsonc (8 lines)
      1 | {
      i 7 more lines
      i Use `print(n = ...)` to see more lines
    Code
      print(tree, n = 2)
    Output
      # jsonc (8 lines)
      1 | {
      2 |     "a": 1,
      i 6 more lines
      i Use `print(n = ...)` to see more lines

---

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [1,2,3]}") %>%
        ts_tree_format()
      tree %>% ts_tree_select("a") %>% print()
    Output
      # jsonc (8 lines, 1 selected element)
        1   | {
      > 2   |     "a": 1,
        3   |     "b": [
        4   |         1,
        5   |         2,
        ...   
    Code
      tree %>% ts_tree_select("b", TRUE) %>% print()
    Output
      # jsonc (8 lines, 3 selected elements)
        1 | {
        2 |     "a": 1,
        3 |     "b": [
      > 4 |         1,
      > 5 |         2,
      > 6 |         3
        7 |     ]
        8 | }

---

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [1,2,3]}") %>%
        ts_tree_format()
      tree %>% ts_tree_select("c") %>% print()
    Output
      # jsonc (8 lines, 0 selected elements)
      1 | {
      2 |     "a": 1,
      3 |     "b": [
      4 |         1,
      5 |         2,
      6 |         3
      7 |     ]
      8 | }

---

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [1,2,3], \"c\": {\"d\":4}}") %>%
        ts_tree_format()
      tree %>% ts_tree_select(c("b", "c")) %>% print()
    Output
      # jsonc (11 lines, 2 selected elements)
         1 | {
         2 |     "a": 1,
      >  3 |     "b": [
      >  4 |         1,
      >  5 |         2,
      >  6 |         3
      >  7 |     ],
      >  8 |     "c": {
      >  9 |         "d": 4
      > 10 |     }
        11 | }

---

    Code
      tree <- tsjsonc::ts_parse_jsonc("[1,2,3,4,5,6,7]") %>% ts_tree_format()
      tree %>% ts_tree_select(1:5) %>% print(n = 3)
    Output
      # jsonc (9 lines, 5 selected elements)
        1   | [
      > 2   |     1,
      > 3   |     2,
      > 4   |     3,
        5   |     4,
        6   |     5,
        7   |     6,
        ...   
      i 2 more selected elements
      i Use `print(n = ...)` to see more selected elements

# end_column == 0

    Code
      tree <- tstoml::ts_parse_toml(tstoml::toml_example_text())
      tree %>% ts_tree_select("owner")
    Output
      # toml (23 lines, 1 selected element)
         2  | 
         3  | title = "TOML Example"
         4  | 
      >  5  | [owner]
      >  6  | name = "Tom Preston-Werner"
      >  7  | dob = 1979-05-27T07:32:00-08:00
      >  8  | 
         9  | [database]
        10  | enabled = true
        11  | ports = [ 8000, 8001, 8002 ]
        ...   

# ts_tree_delete

    Code
      tree <- tsjsonc::ts_parse_jsonc("{\"a\": 1, \"b\": [1,2,3]}")
      tree %>% ts_tree_select("a") %>% ts_tree_delete()
    Output
      # jsonc (1 line)
      1 | {"b": [1,2,3]}
    Code
      tree %>% ts_tree_select("b", -1L) %>% ts_tree_delete()
    Output
      # jsonc (1 line)
      1 | {"a": 1, "b": [1,2]}

# ts_list_parsers

    Code
      unique(ts_list_parsers()[, 1:3])
    Output
      # A data frame: 2 x 3
        package version    title          
      * <chr>   <chr>      <chr>          
      1 tsjsonc 0.0.0.9000 Edit JSON Files
      2 tstoml  0.0.0.9000 Edit TOML files

---

    Code
      writeLines(format_rd_parser_list(ts_list_parsers()))
    Output
      \subsection{Installed tree-sitter parsers}{
      \tabular{llcl}{
      \strong{Package} \tab \strong{Version} \tab \strong{Loaded} \tab \strong{Title} \cr
      \link[tsjsonc:tsjsonc-package]{ tsjsonc} \tab 0.0.0.9000 \tab yes \tab  Edit JSON Files.\cr
      \link[tstoml:tstoml-package]{ tstoml} \tab 0.0.0.9000 \tab yes \tab  Edit TOML files.
      }
      }
      

---

    Code
      format_rd_parser_list(ts_list_parsers()[integer(), ])
    Output
      [1] "No tree-sitter parser packages are installed.\nAvailable tree-sitter parser packages:\n\\itemize{\n\\item \\strong{\\href{https://github.com/gaborcsardi/tsjsonc}{tsjsonc}}: Edit JSON Files\\if{text}{ (https://github.com/gaborcsardi/tsjsonc)}.\n\\item \\strong{\\href{https://github.com/gaborcsardi/tstoml}{tstoml}}: Edit TOML Files\\if{text}{ (https://github.com/gaborcsardi/tstoml)}.\n}\n"

#| title: 
# ---
This is the engine of \code{\link[tsitter:print.ts_tree]{print.ts_tree()}},
possibly useful to obtain a printed representation without doing the
actual printing.\if{html}{\out{
</p><p><!--comment-->
}}


If there are selected nodes in the tree, those will be highlighted
in the output. See \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} to
select nodes in a tree.\if{html}{\out{
</p><p><!--comment-->
}}


\eval{tsitter:::doc_tabs("format_examples")}

#| title: 
# ---
If \code{tree} was created from a file, then \code{ts_tree_write()} by default
writes it back to the same file. Otherwise, the \code{file} argument must be
specified.

\eval{tsitter:::doc_tabs("ts_tree_write_details_file")}

To write to a connection, pass a connection object to the \code{file}
argument. If the connection is opened in binary mode, the raw bytes
are written using \code{\link[base:writeBin]{base::writeBin()}}. Otherwise, the raw bytes are
converted to characters using the system encoding before writing using
\code{\link[base:rawToChar]{base::rawToChar()}}.\if{html}{\out{
</p><p><!--comment-->
}}


Use \code{file = stdout()} to write to the standard output, i.e. to the
console in an interactive R session.

#| title: 
# ---
The \verb{[[<-} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
but it might be more readable.

#| title: 
# ---
The selection process is iterative. Selection expressions (selectors)
are applied one by one, and each selector selects nodes from the
currently selected nodes. For each selector, it is applied individually
to each currently selected node, and the results are concatenated.\if{html}{\out{
</p><p><!--comment-->
}}


The selection process starts from the root of the DOM tree, the document
node (see \code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}), unless
\code{refine = TRUE} is set, in which case it starts from the current
selection.\if{html}{\out{
</p><p><!--comment-->
}}


See the various types of selection expressions below.
\subsection{Selectors}{
\subsection{All elements: \code{TRUE}}{

Selects all child nodes of the current nodes.

\eval{tsitter:::doc_tabs("ts_tree_select_true")}
}

\subsection{Specific keys: character vector}{

Selects child nodes with the given names from nodes with named children.
If a node has no named children, it selects nothing from that node.

\eval{tsitter:::doc_tabs("ts_tree_select_character")}
}

\subsection{By position: integer vector}{

Selects child nodes by position. Positive indices count from the start,
negative indices count from the end. Zero indices are not allowed.

\eval{tsitter:::doc_tabs("ts_tree_select_integer")}
}

\subsection{Matching keys: regular expression}{

A character scalar named \code{regex} can be used to select child nodes
whose names match the given regular expression, from nodes with named
children. If a node has no named children, it selects nothing from that
node.

\eval{tsitter:::doc_tabs("ts_tree_select_regex")}
}

\subsection{Tree sitter query matches}{

A character scalar named \code{query} can be used to select nodes matching
a tree-sitter query. See \code{\link[tsitter:ts_tree_query]{ts_tree_query()}}
for details on tree-sitter queries.\if{html}{\out{
</p><p><!--comment-->
}}


Instead of a character scalar this can also be a two-element list, where
the first element is the query string and the second element is a
character vector of capture names to select. In this case only nodes
matching the given capture names will be selected.

\eval{tsitter:::doc_tabs("ts_tree_select_tsquery")}
}

\subsection{Explicit node ids}{

You can use \code{I(c(...))} to select nodes by their ids directly. This is
for advanced use cases only.

\eval{tsitter:::doc_tabs("ts_tree_select_ids")}
}

}

\subsection{Refining selections}{

If the \code{refine} argument of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} is \code{TRUE}, then
the selection starts from the already selected elements (all of them
simultanously), instead of starting from the document element.

\eval{tsitter:::doc_tabs("ts_tree_select_refine")}
}

\subsection{The \verb{ts_tree_select<-()} replacement function}{

The \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}} replacement
function works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, but it might be more
readable.

\eval{tsitter:::doc_tabs("ts_tree_select_set")}
}

\subsection{The \code{[[} and \verb{[[<-} operators}{

The \code{[[} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_unserialize]{ts_tree_unserialize()}}, but it
might be more readable.

\eval{tsitter:::doc_tabs("ts_tree_select_brackets")}

The \verb{[[<-} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
but it might be more readable.

\eval{tsitter:::doc_tabs("ts_tree_select_brackets_set")}
}

#| title: 
# ---
\subsection{The syntax tree and the DOM tree}{

See \code{\link[tsitter:ts_tree_ast]{ts_tree_ast()}} for the complete
tree-sitter syntax tree that includes all nodes, including syntax
elements like brackets and commas.

\eval{tsitter:::doc_tabs("ts_tree_ast_details_syntax_vs_dom")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-select.R
\name{[[<-.ts_tree}
\alias{[[<-.ts_tree}
\title{Edit parts of a tree-sitter tree}
\usage{
\method{[[}{ts_tree}(x, i) <- value
}
\arguments{
\item{x}{A \code{ts_tree} object.}

\item{i}{A list with selection expressions, see
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} for details.}

\item{value}{An R expression to serialize or
\code{\link[tsitter:ts_tree_deleted]{ts_tree_deleted()}}.}
}
\value{
The modified \code{ts_tree} object.
}
\description{
The \verb{[[<-} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
but it might be more readable.
}
\details{
The following two expressions are equivalent:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tree <- ts_tree_update(ts_tree_select(tree, <selectors>), value)
}\if{html}{\out{</div>}}

and

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tree[[list(<selectors>)]] <- value
}\if{html}{\out{</div>}}
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')

tree

tree[[list("a")]] <- 42
tree[[list("b", -1)]] <- ts_tree_deleted()

tree
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.ts_tree}
\alias{print.ts_tree}
\title{Print a tree-sitter tree}
\usage{
\method{print}{ts_tree}(x, n = 10, ...)
}
\arguments{
\item{x}{\code{ts_tree} object to print.}

\item{n}{Number of lines, or number of selections to print.}

\item{...}{Not used currently.}
}
\value{
Invisibly returns the original \code{ts_tree} object.
}
\description{
Print a \code{ts_tree} object to the screen.
}
\details{
Calls \code{\link[tsitter:format.ts_tree]{format.ts_tree()}} to format the
ts_tree object, writes the formatted object to the standard output, and
returns the original object invisibly.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
json <- tsjsonc::ts_parse_jsonc(
  '{ "a": 1, "b": [10, 20, 30], "c": { "c1": true, "c2": 100 } }'
)
print(json)
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

#| title: 
# ---
A data frame with one row per token, and columns:
\itemize{
\item \code{id}: integer, the id of the token. The (root) document node has id 1.
\item \code{parent}: integer, the id of the parent token. The root token has
parent \code{NA}
\item \code{field_name}: character, the field name of the token in its parent.
\item \code{type}: character, the type of the token.
\item \code{code}: character, the actual code of the token.
\item \code{start_byte}, \code{end_byte}: integer, the byte positions of the token
in the input.
\item \code{start_row}, \code{start_column}, \code{end_row}, \code{end_column}: integer, the
position of the token in the input.
\item \code{is_missing}: logical, whether the token is a missing token added by
the parser to recover from errors.
\item \code{has_error}: logical, whether the token has a parse error.
\item \code{children}: list of integer vectors, the ids of the children tokens.
\item \code{dom_type}: character, the type of the node in the DOM tree. See
\code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}. Nodes that are not part
of the DOM tree have \code{NA_character_} here.
\item \code{dom_children}: list of integer vectors, the ids of the children in the
DOM tree. See \code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}.
\item \code{dom_parent}: integer, the parent of the node in the DOM tree. See
\code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}. Nodes that are not part
of the DOM tree and the document node have have \code{NA_integer_} here.
}

Other, undocumented columns may also be present, these are considered
internal and may change without notice.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-query.R
\name{ts_tree_query}
\alias{ts_tree_query}
\title{Run tree-sitter queries on tree-sitter trees}
\usage{
ts_tree_query(tree, query)
}
\arguments{
\item{tree}{A \code{ts_tree} object.}

\item{query}{Character string, the tree-sitter query to run.}
}
\value{
A list with entries \code{patterns} and \code{matched_captures}.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


\code{patterns} contains information about all patterns in the queries and
it is a data frame with columns: \code{id}, \code{name}, \code{pattern}, \code{match_count}.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


\code{matched_captures} contains information about all matches, and it has
columns \code{id}, \code{pattern}, \code{match}, \code{start_byte}, \code{end_byte}, \code{start_row},
\code{start_column}, \code{end_row}, \code{end_column}, \code{name}, \code{code}.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


The \code{pattern} column of \code{matched_captured} refers to the \code{id} column of
\code{patterns}.
}
\description{
Use \href{https://tree-sitter.github.io/tree-sitter/}{
tree-sitter's query language} to find nodes in a tree-sitter tree.
}
\details{
You probably need to know some details about the specific tree-sitter
parser you are using, to write effective queries. See the documentation
of the parser package you are using for details about the node types
and the query language support. See links below.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Select all numbers in a JSONC document ------------------------------------
json <- tsjsonc::ts_parse_jsonc(
  '{ "a": 1, "b": [10, 20, 30], "c": { "c1": true, "c2": 100 } }'
)
ts_tree_query(json, "(number) @number")
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=ts_tree_select]{ts_tree_select()}} to select the nodes matching a query.

Other ts_tree exploration:
\code{\link[=[.ts_tree]{[.ts_tree()}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}}

Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree exploration}
\concept{ts_tree generics}

#| title: 
# ---
A tree-sitter tree object has at least four classes:
\itemize{
\item \verb{ts_tree_<parser_name>}, e.g. \code{ts_tree_tsjsonc},
\item \code{ts_tree},
\item \code{tbl}, from the pillar package, for better printing when converted
to a data frame, and
\item \code{data.frame}, since it is a data frame internally.
}

The \code{ts_tree} class has custom \code{\link[=format]{format()}} and \code{\link[=print]{print()}} methods, that
show (part of) the underlying document, and also the selected elements,
if any.\if{html}{\out{
</p><p><!--comment-->
}}


It is sometimes useful to treat a \code{tree} \code{ts_tree} object as a data
frame, and drop the \code{ts_tree} classes. This can be done by indexing with
single brackets, e.g. \code{tree[]}. This returns a data frame with one
row per token, and various columns with information about the tokens.
See details in the 'Value' section or this page.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-format.R
\name{ts_tree_format}
\alias{ts_tree_format}
\title{Format the selected elements of a tree sitter tree for printing}
\usage{
ts_tree_format(tree, options, ...)
}
\arguments{
\item{tree}{A \code{ts_tree} object.}

\item{options}{A list of options for the formatting.

See details in the manual of the specific parser.}

\item{...}{Extra arguments for methods.}
}
\value{
A \code{ts_tree} object representing the reformatted document.
}
\description{
(Re)format the selected elements of the document represented by a
tree-sitter tree, if the tree-sitter parser supports formatting.
}
\details{
If \code{tree} does not have a selection, then the whole document is
formatted.

If \code{tree} has an empty selection, then it is returned unchanged.

Some parsers support options to customize the formatting.
See details in the manual of the specific parser.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{ "a":true, "b": [1,2,3] }')
tree

# Format whole document
ts_tree_format(tree)

# Format each top element under the document node in one line
ts_tree_format(
  ts_tree_select(ts_tree_format(tree), TRUE),
  options = list(format = "oneline")
)

\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("tstoml", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tstoml --------------------------------------
tree <- tstoml::ts_parse_toml(r"(
  [servers]
  alpha = { ip = "127.0.0.1", dc = "eqdc10" }
  beta = { ip = "127.0.0.2", dc = "eqdc20" }
)")

tree

ts_tree_format(tree)
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

#| title: 
# ---
If \code{tree} does not have a selection, then the whole document is
formatted.

\eval{tsitter:::doc_tabs("ts_tree_format_details_no_selection")}

If \code{tree} has an empty selection, then it is returned unchanged.

\eval{tsitter:::doc_tabs("ts_tree_format_details_empty_selection")}

Some parsers support options to customize the formatting.
See details in the manual of the specific parser.

\eval{tsitter:::doc_tabs("ts_tree_format_details_options")}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-ast.R
\name{ts_tree_ast}
\alias{ts_tree_ast}
\title{Show the annotated syntax tree of a tree-sitter tree}
\usage{
ts_tree_ast(tree)
}
\arguments{
\item{tree}{A \code{ts_tree} object.}
}
\value{
Character vector, the formatted annotated syntax tree, line by
line. It has class \link[cli:tree]{cli_tree}, from the cli package. It
may contain ANSI escape sequences for coloring and hyperlinks.
}
\description{
\code{ts_tree_ast()} prints the annotated syntax tree of a ts_tree object.
This syntax tree contains all tree-sitter nodes, and it shows the
source code associated with each node, along with line numbers.
}
\details{
\subsection{The syntax tree and the DOM tree}{

This syntax tree contains all nodes of the tree-sitter parse tree,
including both named and unnamed nodes and comments. E.g. for a JSON(C)
document it includes the pairs, brackets, braces, commas, colons,
double quotes and string escape sequences as separate nodes.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


See \code{\link[tsitter:ts_tree_dom]{tsitter::ts_tree_dom()}} for a tree that
shows the semantic structure of the parsed document, which may be
different from the syntax tree.
}
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')

tree

ts_tree_ast(tree)

ts_tree_dom(tree)
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("tstoml", quietly = TRUE)) withAutoprint(\{ # examplesIf}

# Create a parse tree with tstoml --------------------------------------
tree <- tstoml::ts_parse_toml(r"(
  title = "TOML Example"
  [owner]
  name = "Tom Preston-Werner"
  dob = 1979-05-27T07:32:00-08:00
)")

tree

ts_tree_ast(tree)

ts_tree_dom(tree)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=ts_tree_dom]{ts_tree_dom()}} to show the document object model (DOM) of a
ts_tree object.

Other ts_tree exploration:
\code{\link[=[.ts_tree]{[.ts_tree()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}}

Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree exploration}
\concept{ts_tree generics}

#| title: 
# ---
A list with entries \code{patterns} and \code{matched_captures}.\if{html}{\out{
</p><p><!--comment-->
}}


\code{patterns} contains information about all patterns in the queries and
it is a data frame with columns: \code{id}, \code{name}, \code{pattern}, \code{match_count}.\if{html}{\out{
</p><p><!--comment-->
}}


\code{matched_captures} contains information about all matches, and it has
columns \code{id}, \code{pattern}, \code{match}, \code{start_byte}, \code{end_byte}, \code{start_row},
\code{start_column}, \code{end_row}, \code{end_column}, \code{name}, \code{code}.\if{html}{\out{
</p><p><!--comment-->
}}


The \code{pattern} column of \code{matched_captured} refers to the \code{id} column of
\code{patterns}.

#| title: 
# ---
Character string, connection, or \code{NULL}. The file or connection
to write to. By default it writes to the same file that was used in
\code{\link[tsitter:ts_tree_new]{ts_tree_new()}}, if \code{tree} was read from a
file.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-sexpr.R
\name{ts_tree_sexpr}
\alias{ts_tree_sexpr}
\title{Show the syntax tree of a tree-sitter tree}
\usage{
ts_tree_sexpr(tree)
}
\arguments{
\item{tree}{A \code{ts_tree} object.}
}
\value{
A string representing the S-expression of the syntax tree.
}
\description{
Show the structure of a tree-sitter tree as an S-expression.
}
\details{
This function returns a nested list representation of the syntax tree,
where each node is represented as a list with its type and children.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc(
  "{ \"a\": //comment\ntrue, \"b\": [1, 2, 3] }"
)

ts_tree_sexpr(tree)
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("tstoml", quietly = TRUE)) withAutoprint(\{ # examplesIf}

# Create a parse tree with tstoml --------------------------------------
tree <- tstoml::ts_parse_toml(r"(
  [servers]
  alpha = { ip = "127.0.0.1", dc = "eqdc10" }
  beta = { ip = "127.0.0.2", dc = "eqdc20" }
)")

ts_tree_sexpr(tree)
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree exploration:
\code{\link[=[.ts_tree]{[.ts_tree()}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_query]{ts_tree_query()}}

Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree exploration}
\concept{ts_tree generics}

#| title: 
# ---
The formatting of the rest of the document is left as is.

\eval{tsitter:::doc_tabs("ts_tree_delete_details_formatting")}

If the tree does not have a selection, the tree corresponding to the
empty document is returned, i.e. the whole content is deleted.

\eval{tsitter:::doc_tabs("ts_tree_delete_details_no_selection")}

If the tree has a selection, but it is the empty selection, then
the tree is returned unchanged.

\eval{tsitter:::doc_tabs("ts_tree_delete_details_empty_selection")}

For parsers that support comments, deleting elements that include
comments typically delete the comments as well. Other comments are
kept as is. See details in the manual of the specific parser.

\eval{tsitter:::doc_tabs("ts_tree_delete_details_comments")}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{ts_tree_mark_selection1}
\alias{ts_tree_mark_selection1}
\alias{ts_tree_mark_selection1.ts_tree}
\title{Helper function to decide which AST nodes to highlight for a selection
(internal)}
\usage{
ts_tree_mark_selection1(tree, node)

\method{ts_tree_mark_selection1}{ts_tree}(tree, node)
}
\arguments{
\item{tree}{Tree-sitter tree.}

\item{node}{Node id, integer scalar.}
}
\value{
Integer vector of node ids to highlight.
}
\description{
This function is for packages implementing new parsers based on the ts
package. It is very unlikely that you will need to call this function
directly.
}
\details{
In parsers where AST nodes do not correspond one-to-one to DOM nodes
it is useful to highlight multiple AST nodes for a single selected
DOM node. This generic function can be overridden in such parsers to
return multiple AST node ids for a single selected (DOM) node id.

The default implementation simply returns the input node id.
}
\examples{
# This is an internal generic for parser implementations, see the
# tsjsonc and tstoml packages for examples of methods implementing
# custom behavior.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-select.R
\name{ts_tree_selection}
\alias{ts_tree_selection}
\alias{ts_tree_selected_nodes}
\title{Helper functions for tree-sitter tree selections (internal)}
\usage{
ts_tree_selection(tree, default = TRUE)

ts_tree_selected_nodes(tree, default = TRUE)
}
\arguments{
\item{tree}{A \code{ts_tree} object as returned by \code{\link[=ts_tree_new]{ts_tree_new()}}.}

\item{default}{Logical, whether to return the default selection if there
is no explicit selection, or \code{NULL}.}
}
\value{
\code{ts_tree_selection()} returns a list of selection records.

\code{ts_tree_selected_nodes()} returns the ids of the currently
selected nodes.
}
\description{
These functions are for packages implementing new parsers based on the ts
package. It is very unlikely that you will need to call these functions
directly.
}
\details{
\code{ts_tree_selection()} returns the current selection, as a list
of selectors.

\code{ts_tree_selected_nodes()} returns the ids of the currently
selected nodes.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')
tree <- ts_tree_select(tree, "b", -1)
ts_tree_selection(tree)
\dontshow{\}) # examplesIf}
}

#| title: 
# ---
Character vector, the formatted annotated syntax tree, line by
line. It has class \link[cli:tree]{cli_tree}, from the cli package. It
may contain ANSI escape sequences for coloring and hyperlinks.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-select.R
\name{[[.ts_tree}
\alias{[[.ts_tree}
\title{Unserialize parts of a tree-sitter tree}
\usage{
\method{[[}{ts_tree}(x, i, ...)
}
\arguments{
\item{x}{A \code{ts_tree} object.}

\item{i}{Selection expressions in a list, see details in
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}}.}

\item{...}{Additional arguments, passed to
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}}.}
}
\value{
List of R objects, with one entry for each selected element.
}
\description{
The \code{[[} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_unserialize]{ts_tree_unserialize()}}, but it
might be more readable.
}
\details{
The following two expressions are equivalent:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ts_tree_unserialize(ts_tree_select(tree, <selectors>))
}\if{html}{\out{</div>}}

and

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tree[[list(<selectors>)]]
}\if{html}{\out{</div>}}
\subsection{The \verb{[[<-} replacement operator}{

The \verb{[[<-} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
but it might be more readable.
}
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')

tree

tree[[list("a")]]

# Last two elements of "b"
tree[[list("b", -(1:2))]]
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}

Other serialization functions:
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}}
}
\concept{serialization functions}
\concept{ts_tree generics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-write.R
\name{ts_tree_write}
\alias{ts_tree_write}
\title{Write a tree-sitter tree to a file}
\usage{
ts_tree_write(tree, file = NULL)
}
\arguments{
\item{tree}{A \code{ts_tree} object as returned by
\code{\link[tsitter:ts_tree_new]{ts_tree_new()}}.}

\item{file}{Character string, connection, or \code{NULL}. The file or connection
to write to. By default it writes to the same file that was used in
\code{\link[tsitter:ts_tree_new]{ts_tree_new()}}, if \code{tree} was read from a
file.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Writes the document of a ts \code{ts_tree} object to a file or connection.
}
\details{
If \code{tree} was created from a file, then \code{ts_tree_write()} by default
writes it back to the same file. Otherwise, the \code{file} argument must be
specified.

To write to a connection, pass a connection object to the \code{file}
argument. If the connection is opened in binary mode, the raw bytes
are written using \code{\link[base:writeBin]{base::writeBin()}}. Otherwise, the raw bytes are
converted to characters using the system encoding before writing using
\code{\link[base:rawToChar]{base::rawToChar()}}.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


Use \code{file = stdout()} to write to the standard output, i.e. to the
console in an interactive R session.
}
\examples{
\dontshow{if (Sys.getenv("IN_PKGDOWN") == "true") withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')

# Format and write to file
ts_tree_write(ts_tree_format(tree), "example.json")
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}}
}
\concept{ts_tree generics}

#| title: 
# ---
The \code{[[} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_unserialize]{ts_tree_unserialize()}}, but it
might be more readable.

#| title: 
# ---
Calls \code{\link[tsitter:format.ts_tree]{format.ts_tree()}} to format the
ts_tree object, writes the formatted object to the standard output, and
returns the original object invisibly.

\eval{tsitter:::doc_tabs("print_examples")}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-select.R
\name{ts_tree_select}
\alias{ts_tree_select}
\title{Select elements of a tree-sitter tree}
\usage{
ts_tree_select(tree, ..., refine = FALSE)
}
\arguments{
\item{tree}{A \code{ts_tree} object as returned by
\code{\link[tsitter:ts_tree_new]{ts_tree_new()}}.}

\item{...}{Selection expressions, see details.}

\item{refine}{Logical, whether to refine the current selection or start
a new selection.}
}
\value{
A \code{ts_tree} object with the selected parts.
}
\description{
This function is the heart of ts. To edit a tree-sitter tree, you first
need to select the parts you want to delete or update.
}
\details{
The selection process is iterative. Selection expressions (selectors)
are applied one by one, and each selector selects nodes from the
currently selected nodes. For each selector, it is applied individually
to each currently selected node, and the results are concatenated.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


The selection process starts from the root of the DOM tree, the document
node (see \code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}), unless
\code{refine = TRUE} is set, in which case it starts from the current
selection.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


See the various types of selection expressions below.
\subsection{Selectors}{
\subsection{All elements: \code{TRUE}}{

Selects all child nodes of the current nodes.
}

\subsection{Specific keys: character vector}{

Selects child nodes with the given names from nodes with named children.
If a node has no named children, it selects nothing from that node.
}

\subsection{By position: integer vector}{

Selects child nodes by position. Positive indices count from the start,
negative indices count from the end. Zero indices are not allowed.
}

\subsection{Matching keys: regular expression}{

A character scalar named \code{regex} can be used to select child nodes
whose names match the given regular expression, from nodes with named
children. If a node has no named children, it selects nothing from that
node.
}

\subsection{Tree sitter query matches}{

A character scalar named \code{query} can be used to select nodes matching
a tree-sitter query. See \code{\link[tsitter:ts_tree_query]{ts_tree_query()}}
for details on tree-sitter queries.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


Instead of a character scalar this can also be a two-element list, where
the first element is the query string and the second element is a
character vector of capture names to select. In this case only nodes
matching the given capture names will be selected.
}

\subsection{Explicit node ids}{

You can use \code{I(c(...))} to select nodes by their ids directly. This is
for advanced use cases only.
}

}

\subsection{Refining selections}{

If the \code{refine} argument of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} is \code{TRUE}, then
the selection starts from the already selected elements (all of them
simultanously), instead of starting from the document element.
}

\subsection{The \verb{ts_tree_select<-()} replacement function}{

The \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}} replacement
function works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, but it might be more
readable.
}

\subsection{The \code{[[} and \verb{[[<-} operators}{

The \code{[[} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_unserialize]{ts_tree_unserialize()}}, but it
might be more readable.

The \verb{[[<-} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
but it might be more readable.
}
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# ----------------------------------------------------------------------
# Create a JSONC tree, needs the tsjsonc package
json <- ts_tree_new(
  tsjsonc::ts_language_jsonc(),
  text = '{ "a": 1, "b": 2, "c": { "d": 3, "e": 4 } }'
)

ts_tree_select(json, "c", "d")
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("tstoml", quietly = TRUE)) withAutoprint(\{ # examplesIf}

# ----------------------------------------------------------------------
# Create a TOML tree, needs the tstoml package
toml <- ts_tree_new(
  tstoml::ts_language_toml(),
  text = tstoml::toml_example_text()
)

ts_tree_select(toml, "servers", TRUE, "ip")
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-unserialize.R
\name{ts_tree_unserialize}
\alias{ts_tree_unserialize}
\title{Unserialize selected elements of a tree-sitter tree}
\usage{
ts_tree_unserialize(tree)
}
\arguments{
\item{tree}{A \code{ts_tree} object.}
}
\value{
List of R objects, with one entry for each selected element.
}
\description{
Unserialize the selected elements of a \code{ts_tree} object, i.e. convert
them to R objects.
}
\details{
If no elements are selected in the tree, then the whole document is
unserialized.

If the tree has an empty selection, then an empty list is returned.
\subsection{The \code{[[} operator}{

The \code{[[} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{ts_tree_unserialize()}, but it might be more readable.

For the details on how the selected elements are mapped to R objects,
see the documentation of the methods in the parser packages. The methods
in the installed parser packages are linked below.
}
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"a": 13, "b": [1, 2, 3], "c": "x"}')

tree

ts_tree_unserialize(ts_tree_select(tree, c("b", "c")))

ts_tree_unserialize(ts_tree_select(tree, "b"))
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}

Other serialization functions:
\code{\link[=[[.ts_tree]{[[.ts_tree()}}
}
\concept{serialization functions}
\concept{ts_tree generics}

#| title: 
# ---
\code{ts_tree_dom()} prints the document object model (DOM) tree of a ts_tree
object. This tree only includes semantic elements. E.g. for a JSON(C)
document it includes objects, arrays and various value types, but not
the syntax elements like brackets, commas or colons.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{format.ts_tree}
\alias{format.ts_tree}
\title{Format tree-sitter trees}
\usage{
\method{format}{ts_tree}(x, n = 10, ...)
}
\arguments{
\item{x}{\code{ts_tree} object.}

\item{n}{Number of lines, or number of selections to print.}

\item{...}{Currently ignored.}
}
\value{
Character vector of lines to print.
}
\description{
Format a \code{ts_tree} object for printing.
}
\details{
This is the engine of \code{\link[tsitter:print.ts_tree]{print.ts_tree()}},
possibly useful to obtain a printed representation without doing the
actual printing.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


If there are selected nodes in the tree, those will be highlighted
in the output. See \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} to
select nodes in a tree.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
json <- tsjsonc::ts_parse_jsonc(
  '{ "a": 1, "b": [10, 20, 30], "c": { "c1": true, "c2": 100 } }'
)
format(json)
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

#| title: 
# ---
If the tree does not have a selection, the new element replaces the
whole document.

\eval{tsitter:::doc_tabs("ts_tree_update_details_no_selection")}

If the tree has an empty selection, the new element is inserted at the
position of where the selected elements would be.

\eval{tsitter:::doc_tabs("ts_tree_update_details_empty_selection")}

#| title: 
# ---
You probably need to know some details about the specific tree-sitter
parser you are using, to write effective queries. See the documentation
of the parser package you are using for details about the node types
and the query language support. See links below.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bracket-ts-tree.R
\name{[.ts_tree}
\alias{[.ts_tree}
\title{Convert ts_tree object to a data frame}
\usage{
\method{[}{ts_tree}(x, i, j, drop = FALSE)
}
\arguments{
\item{x}{A \code{ts_tree} object.}

\item{i, j}{Incides, passed to the regular data.frame indexing method, see
\code{\link[base:Extract]{Extract}}.}

\item{drop}{Passed to the regular data.frame indexing method, see
\code{\link[base:Extract]{Extract}}.}
}
\value{
A data frame with one row per token, and columns:
\itemize{
\item \code{id}: integer, the id of the token. The (root) document node has id 1.
\item \code{parent}: integer, the id of the parent token. The root token has
parent \code{NA}
\item \code{field_name}: character, the field name of the token in its parent.
\item \code{type}: character, the type of the token.
\item \code{code}: character, the actual code of the token.
\item \code{start_byte}, \code{end_byte}: integer, the byte positions of the token
in the input.
\item \code{start_row}, \code{start_column}, \code{end_row}, \code{end_column}: integer, the
position of the token in the input.
\item \code{is_missing}: logical, whether the token is a missing token added by
the parser to recover from errors.
\item \code{has_error}: logical, whether the token has a parse error.
\item \code{children}: list of integer vectors, the ids of the children tokens.
\item \code{dom_type}: character, the type of the node in the DOM tree. See
\code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}. Nodes that are not part
of the DOM tree have \code{NA_character_} here.
\item \code{dom_children}: list of integer vectors, the ids of the children in the
DOM tree. See \code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}.
\item \code{dom_parent}: integer, the parent of the node in the DOM tree. See
\code{\link[tsitter:ts_tree_dom]{ts_tree_dom()}}. Nodes that are not part
of the DOM tree and the document node have have \code{NA_integer_} here.
}

Other, undocumented columns may also be present, these are considered
internal and may change without notice.
}
\description{
Create a data frame for the syntax tree of a JSON document, by indexing
a ts_tree object with single brackets. This is occasionally useful for
exploration and debugging.
}
\details{
A tree-sitter tree object has at least four classes:
\itemize{
\item \verb{ts_tree_<parser_name>}, e.g. \code{ts_tree_tsjsonc},
\item \code{ts_tree},
\item \code{tbl}, from the pillar package, for better printing when converted
to a data frame, and
\item \code{data.frame}, since it is a data frame internally.
}

The \code{ts_tree} class has custom \code{\link[=format]{format()}} and \code{\link[=print]{print()}} methods, that
show (part of) the underlying document, and also the selected elements,
if any.\if{html}{\out{
</p><p><!-- ------------------------------------------------------>
}}


It is sometimes useful to treat a \code{tree} \code{ts_tree} object as a data
frame, and drop the \code{ts_tree} classes. This can be done by indexing with
single brackets, e.g. \code{tree[]}. This returns a data frame with one
row per token, and various columns with information about the tokens.
See details in the 'Value' section or this page.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')

tree

tree[]
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree exploration:
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}}
}
\concept{ts_tree exploration}

#| title: 
# ---
The following two expressions are equivalent:

\if{html}{\out{<div class="sourceCode r"><p style="display:none">&nbsp;}}\preformatted{ts_tree_select(tree, <selectors>) |> ts_tree_unserialize()
}\if{html}{\out{</div>}}

and

\if{html}{\out{<div class="sourceCode r"><p style="display:none">&nbsp;}}\preformatted{tree[[list(<selectors>)]]
}\if{html}{\out{</div>}}

\eval{tsitter:::doc_tabs("ts_tree_select_brackets")}
\subsection{The \verb{[[<-} replacement operator}{

The \verb{[[<-} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, (and also to the
replacement function \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}}),
but it might be more readable.

\eval{tsitter:::doc_tabs("ts_tree_select_brackets_set")}
}

#| title: 
# ---
\subsection{The syntax tree and the DOM tree}{

This syntax tree contains all nodes of the tree-sitter parse tree,
including both named and unnamed nodes and comments. E.g. for a JSON(C)
document it includes the pairs, brackets, braces, commas, colons,
double quotes and string escape sequences as separate nodes.\if{html}{\out{
</p><p><!--comment-->
}}


See \code{\link[tsitter:ts_tree_dom]{tsitter::ts_tree_dom()}} for a tree that
shows the semantic structure of the parsed document, which may be
different from the syntax tree.

\eval{tsitter:::doc_tabs("ts_tree_ast_details_syntax_vs_dom")}
}

% Generated by roxygen2: do not edit by hand % Please edit documentation in R/ts-package.R \name{about} \alias{about} \title{About tsitter} \value{ Not applicable. } \description{ tsitter is a common interface to \href{https://tree-sitter.github.io/tree-sitter/}{tree-sitter} parsers, implemented in other R packages. It has a common API to \itemize{ \item query, \item edit, \item format, and \item unserialize } tree-sitter parse trees. } \details{ In this document I show examples with the \code{tsjsonc} package. \subsection{Create a tree-sitter tree}{ Create a ts_tree (ts_tree_jsonc) object from a string: \if{html}{\out{
    }}\preformatted{txt <- r"(
    // this is a comment
    \{
      "a": \{
        "a1": [1, 2, 3],
        // comment
        "a2": "string"
      \},
      "b": [
        \{
          "b11": true,
          "b12": false
        \},
        \{
          "b21": false,
          "b22": false
        \}
      ]
    \}
    )"

    json <- tsjsonc::ts_parse_jsonc(txt)
    }\if{html}{\out{
}}\if{html}{\out{

    #>

}} Pretty print a ts_tree object: \if{html}{\out{
    }}\preformatted{json
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (19 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |   "a": \{
    #>  5 |     "a1": [1, 2, 3],
    #>  6 |     // comment
    #>  7 |     "a2": "string"
    #>  8 |   \},
    #>  9 |   "b": [
    #> 10 |     \{
    #> ℹ 9 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} } \subsection{Select nodes of a tree}{ Selecting nodes is the basis of editing and querying tree-sitter trees. Select element by objects key: \if{html}{\out{
    }}\preformatted{ts_tree_select(json, "a")
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (19 lines, 1 selected element)
    #>    1  |
    #>    2  | // this is a comment
    #>    3  | \{
    #> >  4  |   "a": \{
    #> >  5  |     "a1": [1, 2, 3],
    #> >  6  |     // comment
    #> >  7  |     "a2": "string"
    #> >  8  |   \},
    #>    9  |   "b": [
    #>   10  |     \{
    #>   11  |       "b11": true,
    #>   ...

}} Select element inside element: \if{html}{\out{
    }}\preformatted{ts_tree_select(json, "a", "a1")
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (19 lines, 1 selected element)
    #>   2   | // this is a comment
    #>   3   | \{
    #>   4   |   "a": \{
    #> > 5   |     "a1": [1, 2, 3],
    #>   6   |     // comment
    #>   7   |     "a2": "string"
    #>   8   |   \},
    #>   ...

}} Select element(s) of an array: \if{html}{\out{
    }}\preformatted{ts_tree_select(json, "a", "a1", 1:2)
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (19 lines, 2 selected elements)
    #>   2   | // this is a comment
    #>   3   | \{
    #>   4   |   "a": \{
    #> > 5   |     "a1": [1, 2, 3],
    #>   6   |     // comment
    #>   7   |     "a2": "string"
    #>   8   |   \},
    #>   ...

}} Select multiple keys from an object: \if{html}{\out{
    }}\preformatted{ts_tree_select(json, "a", c("a1", "a2"))
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (19 lines, 2 selected elements)
    #>    2  | // this is a comment
    #>    3  | \{
    #>    4  |   "a": \{
    #> >  5  |     "a1": [1, 2, 3],
    #>    6  |     // comment
    #> >  7  |     "a2": "string"
    #>    8  |   \},
    #>    9  |   "b": [
    #>   10  |     \{
    #>   ...

}} Select nodes that match a tree-sitter query: \if{html}{\out{
    }}\preformatted{ts_tree_select(json, query = "((pair value: (false) @val))")
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (19 lines, 3 selected elements)
    #>   ...
    #>    9  |   "b": [
    #>   10  |     \{
    #>   11  |       "b11": true,
    #> > 12  |       "b12": false
    #>   13  |     \},
    #>   14  |     \{
    #> > 15  |       "b21": false,
    #> > 16  |       "b22": false
    #>   17  |     \}
    #>   18  |   ]
    #>   19  | \}

}} } \subsection{Delete elements}{ Delete selected elements: \if{html}{\out{
    }}\preformatted{ts_tree_delete(ts_tree_select(json, "a", "a1"))
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (18 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |   "a": \{
    #>  5 |     // comment
    #>  6 |     "a2": "string"
    #>  7 |   \},
    #>  8 |   "b": [
    #>  9 |     \{
    #> 10 |       "b11": true,
    #> ℹ 8 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} } \subsection{Insert elements}{ Insert element into an array: \if{html}{\out{
    }}\preformatted{ts_tree_insert(ts_tree_select(json, "a", "a1"), at = 2, "new")
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (24 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |   "a": \{
    #>  5 |     "a1": [
    #>  6 |         1,
    #>  7 |         2,
    #>  8 |         "new",
    #>  9 |         3
    #> 10 |     ],
    #> ℹ 14 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} Inserting into an array reformats the array. Insert element into an object, at the specified key: \if{html}{\out{
    }}\preformatted{ts_tree_insert(
      ts_tree_select(json, "a"),
      key = "a0",
      at = 0,
      list("new", "element")
    )
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (27 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |   "a": \{
    #>  5 |       "a0": [
    #>  6 |           "new",
    #>  7 |           "element"
    #>  8 |       ],
    #>  9 |       "a1": [
    #> 10 |           1,
    #> ℹ 17 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} } \subsection{Update elements}{ Update existing element: \if{html}{\out{
    }}\preformatted{ts_tree_update(ts_tree_select(json, "a", c("a1", "a2")), "new value")
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (19 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |   "a": \{
    #>  5 |     "a1": "new value",
    #>  6 |     // comment
    #>  7 |     "a2": "new value"
    #>  8 |   \},
    #>  9 |   "b": [
    #> 10 |     \{
    #> ℹ 9 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} Inserts the element if some parents are missing: \if{html}{\out{
    }}\preformatted{json <- ts_parse_jsonc(text = "\{ \\"a\\": \{ \\"b\\": true \} \}")
    json
    }\if{html}{\out{
}}\if{html}{\out{

    #> Error in ts_parse_jsonc(text = "\{ \\"a\\": \{ \\"b\\": true \} \}") :
    #>   could not find function "ts_parse_jsonc"
    #> # jsonc (19 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |   "a": \{
    #>  5 |     "a1": [1, 2, 3],
    #>  6 |     // comment
    #>  7 |     "a2": "string"
    #>  8 |   \},
    #>  9 |   "b": [
    #> 10 |     \{
    #> ℹ 9 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} \if{html}{\out{
    }}\preformatted{ts_tree_update(ts_tree_select(json, "a", "x", "y"), list(1, 2, 3))
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (30 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |   "a": \{
    #>  5 |       "a1": [
    #>  6 |           1,
    #>  7 |           2,
    #>  8 |           3
    #>  9 |       ],
    #> 10 |       // comment
    #> ℹ 20 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} } \subsection{Write out a document}{ Use \code{stdout()} to write it to the screen instread of a file: \if{html}{\out{
    }}\preformatted{ts_tree_write(json, stdout())
    }\if{html}{\out{
}}\if{html}{\out{

    #>
    #> // this is a comment
    #> \{
    #>   "a": \{
    #>     "a1": [1, 2, 3],
    #>     // comment
    #>     "a2": "string"
    #>   \},
    #>   "b": [
    #>     \{
    #>       "b11": true,
    #>       "b12": false
    #>     \},
    #>     \{
    #>       "b21": false,
    #>       "b22": false
    #>     \}
    #>   ]
    #> \}

}} } \subsection{Formatting}{ Format the whole document: \if{html}{\out{
    }}\preformatted{ts_tree_format(json)
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (23 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |     "a": \{
    #>  5 |         "a1": [
    #>  6 |             1,
    #>  7 |             2,
    #>  8 |             3
    #>  9 |         ],
    #> 10 |         // comment
    #> ℹ 13 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} Format part of the document: \if{html}{\out{
    }}\preformatted{ts_tree_format(
      ts_tree_select(json, "a"),
      options = list(format = "compact")
    )
    }\if{html}{\out{
}}\if{html}{\out{

    #> # jsonc (15 lines)
    #>  1 |
    #>  2 | // this is a comment
    #>  3 | \{
    #>  4 |   "a": \{"a1":[1,2,3],"a2":"string"\},
    #>  5 |   "b": [
    #>  6 |     \{
    #>  7 |       "b11": true,
    #>  8 |       "b12": false
    #>  9 |     \},
    #> 10 |     \{
    #> ℹ 5 more lines
    #> ℹ Use `print(n = ...)` to see more lines

}} } \subsection{Unserializing}{ Unserialize a whole document: \if{html}{\out{
    }}\preformatted{ts_tree_unserialize(json)
    }\if{html}{\out{
}}\if{html}{\out{

    #> [[1]]
    #> [[1]]$a
    #> [[1]]$a$a1
    #> [[1]]$a$a1[[1]]
    #> [1] 1
    #>
    #> [[1]]$a$a1[[2]]
    #> [1] 2
    #>
    #> [[1]]$a$a1[[3]]
    #> [1] 3
    #>
    #>
    #> [[1]]$a$a2
    #> [1] "string"
    #>
    #>
    #> [[1]]$b
    #> [[1]]$b[[1]]
    #> [[1]]$b[[1]]$b11
    #> [1] TRUE
    #>
    #> [[1]]$b[[1]]$b12
    #> [1] FALSE
    #>
    #>
    #> [[1]]$b[[2]]
    #> [[1]]$b[[2]]$b21
    #> [1] FALSE
    #>
    #> [[1]]$b[[2]]$b22
    #> [1] FALSE
    #>
    #>
    #>
    #>

}} Note that \code{ts_tree_unserialize()} always returns a list, the first element of the list is the unserialized document. Unserialize part(s) of the document: \if{html}{\out{
    }}\preformatted{ts_tree_unserialize(ts_tree_select(json, "b"))
    }\if{html}{\out{
}}\if{html}{\out{

    #> [[1]]
    #> [[1]][[1]]
    #> [[1]][[1]]$b11
    #> [1] TRUE
    #>
    #> [[1]][[1]]$b12
    #> [1] FALSE
    #>
    #>
    #> [[1]][[2]]
    #> [[1]][[2]]$b21
    #> [1] FALSE
    #>
    #> [[1]][[2]]$b22
    #> [1] FALSE
    #>
    #>
    #>

}} Again, \code{ts_tree_unserialize()} returns a list, with one element for each selected node. } \subsection{Exploring a tree-sitter tree}{ It is often useful to explore the structure of a (JSONC) tree-sitter tree, to help writing the right selection or tree-sitter queries. Print the annotated syntax tree: \if{html}{\out{
    }}\preformatted{ts_tree_ast(json)
    }\if{html}{\out{
}}\if{html}{\out{

    #> document (1)                          2|
    #> ├─comment (2)                          |// this is a comment
    #> └─object (3)                          3|
    #>   ├─\{ (4)                              |\{
    #>   ├─pair (5)                          4|
    #>   │ ├─string (6)                       |
    #>   │ │ ├─" (7)                          |  "
    #>   │ │ ├─string_content (8)             |   a
    #>   │ │ └─" (9)                          |    "
    #>   │ ├─: (10)                           |     :
    #>   │ └─object (11)                      |
    #>   │   ├─\{ (12)                         |       \{
    #>   │   ├─pair (13)                     5|
    #>   │   │ ├─string (14)                  |
    #>   │   │ │ ├─" (15)                     |    "
    #>   │   │ │ ├─string_content (16)        |     a1
    #>   │   │ │ └─" (17)                     |       "
    #>   │   │ ├─: (18)                       |        :
    #>   │   │ └─array (19)                   |
    #>   │   │   ├─[ (20)                     |          [
    #>   │   │   ├─number (21)                |           1
    #>   │   │   ├─, (22)                     |            ,
    #>   │   │   ├─number (23)                |              2
    #>   │   │   ├─, (24)                     |               ,
    #>   │   │   ├─number (25)                |                 3
    #>   │   │   └─] (26)                     |                  ]
    #>   │   ├─, (27)                         |                   ,
    #>   │   ├─comment (28)                  6|    // comment
    #>   │   ├─pair (29)                     7|
    #>   │   │ ├─string (30)                  |
    #>   │   │ │ ├─" (31)                     |    "
    #>   │   │ │ ├─string_content (32)        |     a2
    #>   │   │ │ └─" (33)                     |       "
    #>   │   │ ├─: (34)                       |        :
    #>   │   │ └─string (35)                  |
    #>   │   │   ├─" (36)                     |          "
    #>   │   │   ├─string_content (37)        |           string
    #>   │   │   └─" (38)                     |                 "
    #>   │   └─\} (39)                        8|  \}
    #>   ├─, (40)                             |   ,
    #>   ├─pair (41)                         9|
    #>   │ ├─string (42)                      |
    #>   │ │ ├─" (43)                         |  "
    #>   │ │ ├─string_content (44)            |   b
    #>   │ │ └─" (45)                         |    "
    #>   │ ├─: (46)                           |     :
    #>   │ └─array (47)                       |
    #>   │   ├─[ (48)                         |       [
    #>   │   ├─object (49)                  10|
    #>   │   │ ├─\{ (50)                       |    \{
    #>   │   │ ├─pair (51)                  11|
    #>   │   │ │ ├─string (52)                |
    #>   │   │ │ │ ├─" (53)                   |      "
    #>   │   │ │ │ ├─string_content (54)      |       b11
    #>   │   │ │ │ └─" (55)                   |          "
    #>   │   │ │ ├─: (56)                     |           :
    #>   │   │ │ └─true (57)                  |             true
    #>   │   │ ├─, (58)                       |                 ,
    #>   │   │ ├─pair (59)                  12|
    #>   │   │ │ ├─string (60)                |
    #>   │   │ │ │ ├─" (61)                   |      "
    #>   │   │ │ │ ├─string_content (62)      |       b12
    #>   │   │ │ │ └─" (63)                   |          "
    #>   │   │ │ ├─: (64)                     |           :
    #>   │   │ │ └─false (65)                 |             false
    #>   │   │ └─\} (66)                     13|    \}
    #>   │   ├─, (67)                         |     ,
    #>   │   ├─object (68)                  14|
    #>   │   │ ├─\{ (69)                       |    \{
    #>   │   │ ├─pair (70)                  15|
    #>   │   │ │ ├─string (71)                |
    #>   │   │ │ │ ├─" (72)                   |      "
    #>   │   │ │ │ ├─string_content (73)      |       b21
    #>   │   │ │ │ └─" (74)                   |          "
    #>   │   │ │ ├─: (75)                     |           :
    #>   │   │ │ └─false (76)                 |             false
    #>   │   │ ├─, (77)                       |                  ,
    #>   │   │ ├─pair (78)                  16|
    #>   │   │ │ ├─string (79)                |
    #>   │   │ │ │ ├─" (80)                   |      "
    #>   │   │ │ │ ├─string_content (81)      |       b22
    #>   │   │ │ │ └─" (82)                   |          "
    #>   │   │ │ ├─: (83)                     |           :
    #>   │   │ │ └─false (84)                 |             false
    #>   │   │ └─\} (85)                     17|    \}
    #>   │   └─] (86)                       18|  ]
    #>   └─\} (87)                           19|\}

}} Print the document object model: \if{html}{\out{
    }}\preformatted{ts_tree_dom(json)
    }\if{html}{\out{
}}\if{html}{\out{

    #> document (1)
    #> └─object (3)
    #>   ├─object (11) # a
    #>   │ ├─array (19) # a1
    #>   │ │ ├─number (21)
    #>   │ │ ├─number (23)
    #>   │ │ └─number (25)
    #>   │ └─string (35) # a2
    #>   └─array (47) # b
    #>     ├─object (49)
    #>     │ ├─true (57) # b11
    #>     │ └─false (65) # b12
    #>     └─object (68)
    #>       ├─false (76) # b21
    #>       └─false (84) # b22

}} Print the structural summary of a tree: \if{html}{\out{
    }}\preformatted{ts_tree_sexpr(json)
    }\if{html}{\out{
}}\if{html}{\out{

    #> [1] "(document (comment) (object (pair key: (string (string_content)) value: (ob
    #> ject (pair key: (string (string_content)) value: (array (number) (number) (numbe
    #> r))) (comment) (pair key: (string (string_content)) value: (string (string_conte
    #> nt))))) (pair key: (string (string_content)) value: (array (object (pair key: (s
    #> tring (string_content)) value: (true)) (pair key: (string (string_content)) valu
    #> e: (false))) (object (pair key: (string (string_content)) value: (false)) (pair
    #> key: (string (string_content)) value: (false)))))))"

}} } } \examples{ # See above please. }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-list-parsers.R
\name{ts_list_parsers}
\alias{ts_list_parsers}
\title{List installed tree-sitter parsers}
\usage{
ts_list_parsers(lib_path = .libPaths())
}
\arguments{
\item{lib_path}{Library paths to search for installed packages.
Default is \code{\link[base:.libPaths]{base::.libPaths()}}.}
}
\value{
A data frame with columns:
\itemize{
\item \code{package}: character, the name of the package.
\item \code{version}: character, the version of the package.
\item \code{title}: character, the title of the package.
\item \code{library}: character, the library path where the package is installed.
\item \code{loaded}: logical, whether the package is currently loaded.
}
}
\description{
The ts package contains a common interface to several tree-sitter
parsers, implemented in other R packages. \code{ts_list_parsers()} lists
the available parsers installed in the system.
}
\details{
To see tree-sitter parser packages that are available on CRAN, but not
installed on your system, see the packages that depend on ts and have a
name with a 'ts' prefix.

Here is an example that includes all tree-sitter parsers at this time:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ts_list_parsers()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #B9C0CB;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #999999;"># A data frame: 2 × 5</span>                                                           
#>   package version    title           library                    loaded          
#> <span style="color: #c2c2c2;">*</span> <span style="font-style: italic;color: #999999;">&lt;chr&gt;</span>   <span style="font-style: italic;color: #999999;">&lt;chr&gt;</span>      <span style="font-style: italic;color: #999999;">&lt;chr&gt;</span>           <span style="font-style: italic;color: #999999;">&lt;chr&gt;</span>                      <span style="font-style: italic;color: #999999;">&lt;lgl&gt;</span>           
#> <span style="color: #c2c2c2;">1</span> tsjsonc 0.0.0.9000 Edit JSON Files /Users/gaborcsardi/Librar… FALSE           
#> <span style="color: #c2c2c2;">2</span> tstoml  0.0.0.9000 Edit TOML files /Users/gaborcsardi/Librar… FALSE           
</pre></div>
}}
}
\examples{
ts_list_parsers()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-select.R
\name{ts_tree_select1}
\alias{ts_tree_select1}
\alias{ts_tree_select1.ts_tree.ts_tree_selector_default}
\alias{ts_tree_select1.ts_tree.NULL}
\alias{ts_tree_select1.ts_tree.ts_tree_selector_ids}
\alias{ts_tree_select1.ts_tree.ts_tree_selector_tsquery}
\alias{ts_tree_select1.ts_tree.character}
\alias{ts_tree_select1.ts_tree.integer}
\alias{ts_tree_select1.ts_tree.numeric}
\alias{ts_tree_select1.ts_tree.ts_tree_selector_regex}
\alias{ts_tree_select1.ts_tree.logical}
\title{Select nodes from a tree-sitter tree (internal)}
\usage{
ts_tree_select1(tree, node, slt)

\method{ts_tree_select1}{ts_tree.ts_tree_selector_default}(tree, node, slt)

\method{ts_tree_select1}{ts_tree.NULL}(tree, node, slt)

\method{ts_tree_select1}{ts_tree.ts_tree_selector_ids}(tree, node, slt)

\method{ts_tree_select1}{ts_tree.ts_tree_selector_tsquery}(tree, node, slt)

\method{ts_tree_select1}{ts_tree.character}(tree, node, slt)

\method{ts_tree_select1}{ts_tree.integer}(tree, node, slt)

\method{ts_tree_select1}{ts_tree.numeric}(tree, node, slt)

\method{ts_tree_select1}{ts_tree.ts_tree_selector_regex}(tree, node, slt)

\method{ts_tree_select1}{ts_tree.logical}(tree, node, slt)
}
\arguments{
\item{tree}{A \code{ts_tree} object as returned by \code{\link[=ts_tree_new]{ts_tree_new()}}.}

\item{node}{Integer scalar, the node id to select from.}

\item{slt}{A selector object, see details in \code{\link[=ts_tree_select]{ts_tree_select()}}.}
}
\value{
Must return an integer vector of selected node ids.
}
\description{
This function is for packages implementing new parsers based on the ts
package. It is very unlikely that you will need to call this function
directly.
}
\details{
A parser package may implement methods for this generic to change the
behavior of \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} for a
certain selector type, or even add new selector types.

Each new method should be named as

\if{html}{\out{<div class="sourceCode">}}\preformatted{ts_tree_select.<ts_tree_class>.<selector_class>
}\if{html}{\out{</div>}}

The ts package implement deault methods for the selector types described
in the \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} manual page.

\subsection{\code{ts_tree_selector_default} selector}{

Method: \code{ts_tree_select1.ts_tree.ts_tree_selector_default}

This method is used to select the default element(s), when there is no
selected element. E.g. when starting a new selection from the root of
the DOM tree.

The default implementation returns the ids of all children of the
document root in the AST, except comments. If there are no such
children, it returns the id of the document root of the AST itself
(always id 1).
}

\subsection{\code{NULL} selector}{

Method: \code{ts_tree_select1.ts_tree.NULL}

This method is used for the  \code{NULL} selector, that is supposed to
clear the selection. You probably do not need to override this method.
The default implementation returns an empty integer vector.
}

\subsection{\code{ts_tree_selector_ids} selector}{

Method: \code{ts_tree_select1.ts_tree.ts_tree_selector_ids}

This method is used to select nodes by their ids directly.
You probably do not need to override this method.
The default implementation returns the ids stored in the selector.
\subsection{Note}{

This behaviour may change in the future to select only nodes
in the subtree of the current node.
}

}

\subsection{\code{ts_tree_selector_tsquery} selector}{

Method: \code{ts_tree_select1.ts_tree.ts_tree_selector_tsquery}

This method is used to select nodes matching a tree-sitter query.
You probably do not need to override this method.
The default implementation returns the ids stored in the selector.
\subsection{Note}{

This behaviour may change in the future to select only nodes
in the subtree of the current node.
}

}

\subsection{\code{character} (character vector) selector}{

Method: \code{ts_tree_select1.ts_tree.character}

This method is used when the selector is a character vector.
The default implementation selects DOM children of \code{node} whose names
are in the character vector. If not all children o \code{node} are named,
it returns an empty integer vector. (E.g. in a JSONC document it returns
an empty integer vector when nodes is an array.)
}

\subsection{\code{integer} (integer vector) selector}{

Method: \code{ts_tree_select1.ts_tree.integer}

This method is used when the selector is an integer vector.
The default implementation selects DOM children of \code{node} by position.
Positive indices count from the start, negative indices count from the
end. Zero indices are not allowed and an error is raised if any are
used.
}

\subsection{\code{numeric} (numeric, double vector) selector}{

Method: \code{ts_tree_select1.ts_tree.numeric}

This method is used when the selector is a numeric (double) vector.
It currrently coerces the numeric vector to integer and calls the
integer method.
}

\subsection{\code{ts_tree_selector_regex} (regular expression) selector}{

Method: \code{ts_tree_select1.ts_tree.ts_tree_selector_regex}

This method is used when the selector is a regular expression.
The default implementation selects DOM children of \code{node} whose names
match the regular expression. If not all children o \code{node} are named,
it returns an empty integer vector. (E.g. in a JSONC document it
returns an empty integer vector when nodes is an array.)
}

\subsection{\code{logical} (logical vector) selector}{

Method: \code{ts_tree_select1.ts_tree.logical}

This method is used when the selector is a logical vector.
The default implementation only supports scalar \code{TRUE}, which selects
all DOM children of \code{node}. Other values raise an error.
}
}
\examples{
# This is an internal generic for parser implementations, see the
# tsjsonc and tstoml packages for examples of methods implementing
# selector types.
}

#| title: 
# ---
\code{ts_tree_ast()} prints the annotated syntax tree of a ts_tree object.
This syntax tree contains all tree-sitter nodes, and it shows the
source code associated with each node, along with line numbers.

#| title: 
# ---
The following two expressions are equivalent:

\if{html}{\out{<div class="sourceCode r"><p style="display:none">&nbsp;}}\preformatted{tree <- ts_tree_select(tree, <selectors>) |> ts_tree_update(value)
}\if{html}{\out{</div>}}

and

\if{html}{\out{<div class="sourceCode r"><p style="display:none">&nbsp;}}\preformatted{ts_tree_select(tree, <selectors>) <- value
}\if{html}{\out{</div>}}

\eval{tsitter:::doc_tabs("ts_tree_select_set")}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-delete.R
\name{ts_tree_delete}
\alias{ts_tree_delete}
\title{Delete selected elements from a tree-sitter tree}
\usage{
ts_tree_delete(tree, ...)
}
\arguments{
\item{tree}{A \code{ts_tree} object.}

\item{...}{Extra arguments for methods.}
}
\value{
The modified \code{ts_tree} object with the selected elements removed.
}
\description{
Use \code{\link[tsitter:ts_tree_select]{ts_tree_select()}} to select the
elements to be deleted, and then call \code{ts_tree_delete()} to remove them
from the tree.
}
\details{
The formatting of the rest of the document is left as is.

If the tree does not have a selection, the tree corresponding to the
empty document is returned, i.e. the whole content is deleted.

If the tree has a selection, but it is the empty selection, then
the tree is returned unchanged.

For parsers that support comments, deleting elements that include
comments typically delete the comments as well. Other comments are
kept as is. See details in the manual of the specific parser.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc(
  "{ \"a\": //comment\ntrue, \"b\": [1, 2, 3] }"
)

tree

ts_tree_select(tree, "a")

ts_tree_delete(ts_tree_select(tree, "a"))
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("tstoml", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tstoml --------------------------------------
tree <- tstoml::ts_parse_toml(r"(
  [servers]
  alpha = { ip = "127.0.0.1", dc = "eqdc10" }
  beta = { ip = "127.0.0.2", dc = "eqdc20" }
)")

tree

ts_tree_select(tree, "servers", TRUE, "dc")
ts_tree_delete(ts_tree_select(tree, "servers", TRUE, "dc"))
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

#| title: 
# ---
The \code{\link[tsitter:ts_tree_select<-]{ts_tree_select<-()}} replacement
function works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{\link[tsitter:ts_tree_update]{ts_tree_update()}}, but
it might be more readable.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapse.R, R/error.R
\name{ts_collapse}
\alias{ts_collapse}
\alias{ts_cnd}
\alias{ts_caller_arg}
\alias{as_ts_caller_arg}
\alias{as.character.ts_caller_arg}
\alias{ts_caller_env}
\alias{ts_check_named_arg}
\alias{ts_parse_error_cnd}
\title{Utility functions for ts language implementations (internal)}
\usage{
ts_collapse(
  s,
  sep = ", ",
  sep2 = sub("^,", "", last),
  last = ", and ",
  trunc = Inf,
  width = Inf,
  ellipsis = "...",
  style = c("both-ends", "head")
)

ts_cnd(..., class = NULL, call = ts_caller_env(), .envir = parent.frame())

ts_caller_arg(arg)

as_ts_caller_arg(x)

\method{as.character}{ts_caller_arg}(x, ...)

ts_caller_env(n = 1)

ts_check_named_arg(arg, frame = -1)

ts_parse_error_cnd(tree, text, call = ts_caller_env())
}
\arguments{
\item{s}{For \code{ts_collapse()} a character vector to collapse.}

\item{sep}{Separator string for most elements.}

\item{sep2}{Separator string for two elements.}

\item{last}{Separator string before the last element.}

\item{trunc}{Integer, maximum number of elements to show before
truncation.}

\item{width}{Integer, maximum display width of the collapsed string.
If the collapsed string exceeds this width, it will be truncated
with \code{ellipsis}.}

\item{ellipsis}{String to indicate truncation.}

\item{style}{Character, the collapsing style to use. Possible values are
\code{"both-ends"} (the default), which shows the first few and last few
elements when truncating, and \code{"head"}, which shows only the first few
elements.}

\item{...}{Arguments collapsed and interpolated into a condition message.}

\item{class}{A character vector of classes for the condition.}

\item{call}{Environment of the call to associate with the error
condition, defaults to the caller's environment.}

\item{.envir}{The environment in which to evaluate the \code{...} expressions,
defaults to the parent frame.}

\item{arg}{Argument to check.}

\item{x}{A ts_caller_arg object.}

\item{n}{Number of frames to go up to find the caller environment.}

\item{frame}{Frame number to inspect, defaults to the caller's frame.}

\item{tree}{A \code{ts_tree} object as returned by \code{\link[=ts_tree_new]{ts_tree_new()}}.}

\item{text}{Raw vector, the original text used to parse the tree.}
}
\value{
\code{ts_collapse()} returns a character scalar, the collapsed string.

\code{ts_cnd()} returns an error condition object.

\code{ts_caller_arg()} returns the captured expression as a
ts_caller_arg object.

\code{as_ts_caller_arg()} returns a ts_caller_arg object.

\code{as.character.ts_caller_arg()} returns a short string
representation of the caller argument, a character scalar.

\code{ts_caller_env()} returns an environment, or \code{NULL} is called
from the global environment.

\code{ts_check_named_arg()} returns \code{TRUE} invisibly if the argument
was named, otherwise it raises an error.

\code{ts_parse_error_cnd()} returns a ts_parse_error condition
}
\description{
These functions are for packages implementing new parsers based on the ts
package. It is very unlikely that you will need to call these functions
directly.
}
\details{
\code{ts_collapse()} collapses a character vector into a single string,
with options for truncation by number of elements or display width.
It is useful for creating informative error messages.

\code{ts_cnd()} creates a condition object. It interpolates its
arguments into a single message.

\code{ts_caller_arg()} captures the expression used as an argument
to a function, for use in error messages.

\code{as_ts_caller_arg()} converts an object into a caller argument
object. This is useful when referring to parts of the caller argument
in downstream error messages.

\code{as.character.ts_caller_arg()} formats a caller argument
object as a short string for use in error messages. Multi-line
expressions are truncated after the first line.

\code{ts_caller_env()} returns the environment of the caller function,
\code{n} levels up the call stack. This is useful for associating error
conditions with the correct call.

\code{ts_check_named_arg()} checks whether an argument was supplied
with a name. If not, it raises an error.

\code{ts_parse_error_cnd()} creates a parse error condition
associated with a \code{ts_tree} object and the original text.
The error message includes information about the location of
parse errors in the text. It also has \code{format()} and \code{print()}
methods to display the error together with the relevant lines
of the original text.
}
\examples{
ts_collapse(letters[1:3])
ts_collapse(letters[1:10], trunc = 5)
}
\keyword{internal}

#| title: 
# ---
Can be used to parse part(s) of the input. It must be a data frame with
integer columns \code{start_row}, \code{start_col}, \code{end_row}, \code{end_col},
\code{start_byte}, \code{end_byte}, in this order.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as.character.ts_tree}
\alias{as.character.ts_tree}
\title{The document of a tree-sitter tree as a character scalar}
\usage{
\method{as.character}{ts_tree}(x, ...)
}
\arguments{
\item{x}{A \code{ts_tree} object.}

\item{...}{Ignored.}
}
\value{
A character scalar containing the document of the tree.
}
\description{
The document of a tree-sitter tree as a character scalar
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')

tree
as.character(tree)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=as.raw.ts_tree]{as.raw.ts_tree()}} to get the document as a raw vector.
}

#| title: 
# ---
It is not always possible to insert a new element into a selected
element. For example in a JSONC document you can only insert a new
element into an array or an object, but not into scalar elements.
If the insertion is not possible, an error is raised.

\eval{tsitter:::doc_tabs("ts_tree_insert_details_errors")}

If \code{tree} does not have a selection, the new element is inserted into
at the top level.

\eval{tsitter:::doc_tabs("ts_tree_insert_details_no_selection")}

If \code{tree} has an empty selection, then it is returned unchanged, i.e.
no new element is inserted.

\eval{tsitter:::doc_tabs("ts_tree_insert_details_empty_selection")}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-insert.R
\name{ts_tree_insert}
\alias{ts_tree_insert}
\title{Insert a new element into a tree-sitter tree}
\usage{
ts_tree_insert(tree, new, key, at, options, ...)
}
\arguments{
\item{tree}{A \code{ts_tree} object.}

\item{new}{The new element to insert.

The type of \code{new} depends on the parser and the method that implements
the insertion. See details in the manual of the specific parser.}

\item{key}{The key of the new element, if inserting into a keyed element.

For example a JSON(C) object or a TOML table are keyed elements.}

\item{at}{The position to insert the new element at.

The interpretation of this argument depends on the method that
implements the insertion. Typically the followings are supported:
\itemize{
\item \code{0} inserts at the beginning.
\item \code{Inf} inserts at the end.
\item A positive integer \code{n} inserts \emph{after} the \code{n}-th element.
\item A character scalar inserts \emph{after} the element with the given key,
in keyed elements.
}

See the details in the manual of the specific parser.}

\item{options}{A list of options for the insertion.

See details in the manual of the specific parser.}

\item{...}{Extra arguments for methods.}
}
\value{
A \code{ts_tree} object representing the modified parse tree.
}
\description{
Insert a new element into each selected element.
}
\details{
It is not always possible to insert a new element into a selected
element. For example in a JSONC document you can only insert a new
element into an array or an object, but not into scalar elements.
If the insertion is not possible, an error is raised.

If \code{tree} does not have a selection, the new element is inserted into
at the top level.

If \code{tree} has an empty selection, then it is returned unchanged, i.e.
no new element is inserted.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{ "a": true, "b": [1, 2, 3] }')

ts_tree_insert(ts_tree_select(tree, "b"), 4, at = Inf)
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("tstoml", quietly = TRUE)) withAutoprint(\{ # examplesIf}

# Create a parse tree with tstoml --------------------------------------
tree <- tstoml::ts_parse_toml(r"(
  [servers]
  alpha = { ip = "127.0.0.1", dc = "eqdc10" }
  beta = { ip = "127.0.0.2", dc = "eqdc20" }
)")

ts_tree_insert(
  ts_tree_select(tree, "servers", TRUE),
  key = "active",
  TRUE
)
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

#| title: 
# ---
If no elements are selected in the tree, then the whole document is
unserialized.

\eval{tsitter:::doc_tabs("ts_tree_unserialize_details_no_selection")}

If the tree has an empty selection, then an empty list is returned.

\eval{tsitter:::doc_tabs("ts_tree_unserialize_details_empty_selection")}
\subsection{The \code{[[} operator}{

The \code{[[} operator works similarly to the combination of
\code{\link[tsitter:ts_tree_select]{ts_tree_select()}} and
\code{ts_tree_unserialize()}, but it might be more readable.

\eval{tsitter:::doc_tabs("ts_tree_select_brackets")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-update.R
\name{ts_tree_update}
\alias{ts_tree_update}
\title{Replace selected elements with a new element in a tree-sitter tree}
\usage{
ts_tree_update(tree, new, options, ...)
}
\arguments{
\item{tree}{A \code{ts_tree} object.}

\item{new}{The new element to replace the selected elements with.

The type of \code{new} depends on the parser and the method that implements
the insertion. See details in the manual of the specific parser.

See details in the manual of the specific parser.}

\item{options}{A list of options for the update.}

\item{...}{Extra arguments for methods.}
}
\value{
The modified \code{ts_tree} object with the selected elements replaced
by the new element.
}
\description{
Replace all selected elements with a new element.
}
\details{
If the tree does not have a selection, the new element replaces the
whole document.

If the tree has an empty selection, the new element is inserted at the
position of where the selected elements would be.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc(r"(
  {
    "name": "example",
    "version": "1.0.0",
    "dependencies": {
      "tsjsonc": "^0.1.0"
    }
  }
)")

ts_tree_update(ts_tree_select(tree, "version"), "2.0.0")
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("tstoml", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tstoml --------------------------------------
tree <- tstoml::ts_parse_toml(r"(
  [package]
  name = "example"
  version = "1.0.0"
  depdendencies = { tstoml = "0.1.0" }
)")

ts_tree_update(ts_tree_select(tree, "package", "version"), "2.0.0")
\dontshow{\}) # examplesIf}
}
\seealso{
Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_new]{ts_tree_new()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts-tree-new.R
\name{ts_tree_new}
\alias{ts_tree_new}
\title{Create tree-sitter tree from file or string}
\usage{
ts_tree_new(
  language,
  file = NULL,
  text = NULL,
  ranges = NULL,
  fail_on_parse_error = TRUE,
  ...
)
}
\arguments{
\item{language}{Language of the file or string, a \code{ts_language} object, e.g. the return
value of \code{tsjsonc::ts_language_jsonc()}.}

\item{file}{Path of a file to parse. Use either \code{file} or \code{text}, but not both.}

\item{text}{String to parse. Use either \code{file} or \code{text}, but not both.}

\item{ranges}{Can be used to parse part(s) of the input. It must be a data frame with
integer columns \code{start_row}, \code{start_col}, \code{end_row}, \code{end_col},
\code{start_byte}, \code{end_byte}, in this order.}

\item{fail_on_parse_error}{Logical, whether to error if there are parse errors in the document.
Default is \code{TRUE}.}

\item{...}{Additional arguments for methods.}
}
\value{
A \code{ts_tree} object representing the parse tree of the input. You can
use the single bracket \code{\link[tsitter:ts_tree-brackets]{[}}
operator to convert it to a data frame.
}
\description{
This is the main function to create a tree-sitter parse tree, using a
ts parser implemented in another package.

The result is a \code{ts_tree} object. A \code{ts_tree} object may be queried,
edited, formatted, written to file, etc. using \code{ts_tree} methods.
}
\details{
A package that implements a tree-sitter parser provides a function that
creates a \code{ts_language} object for that parser. E.g.
tsjsonc has \code{tsjsonc::ts_language_jsonc()}.
You need to use the returned \code{ts_language} object as the \code{language}
argument of \code{\link[tsitter:ts_tree_new]{ts_tree_new()}}.
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}

# JSONC example, needs the tsjsonc package -----------------------------
json <- ts_tree_new(
  tsjsonc::ts_language_jsonc(),
  text = '{ "a": 1, "b": 2 }'
)

json

ts_tree_format(json)
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("tstoml", quietly = TRUE)) withAutoprint(\{ # examplesIf}

# TOML example, needs the tstoml package -------------------------------
toml <- ts_tree_new(
  tstoml::ts_language_toml(),
  text = '[section]\nkey = "value"\nnumber = 42\n'
)

toml

ts_tree_format(toml)
\dontshow{\}) # examplesIf}
}
\seealso{
The tree-sitter parser packages typically include shortcuts to
create parse trees from strings and file, e.g.
\code{tsjsonc::ts_parse_jsonc()} and \code{tsjsonc::ts_read_jsonc()}.

Other ts_tree generics:
\code{\link[=[[.ts_tree]{[[.ts_tree()}},
\code{\link[=[[<-.ts_tree]{[[<-.ts_tree()}},
\code{\link[=format.ts_tree]{format.ts_tree()}},
\code{\link[=print.ts_tree]{print.ts_tree()}},
\code{\link{select-set}},
\code{\link[=ts_tree_ast]{ts_tree_ast()}},
\code{\link[=ts_tree_delete]{ts_tree_delete()}},
\code{\link[=ts_tree_dom]{ts_tree_dom()}},
\code{\link[=ts_tree_format]{ts_tree_format()}},
\code{\link[=ts_tree_insert]{ts_tree_insert()}},
\code{\link[=ts_tree_query]{ts_tree_query()}},
\code{\link[=ts_tree_select]{ts_tree_select()}},
\code{\link[=ts_tree_sexpr]{ts_tree_sexpr()}},
\code{\link[=ts_tree_unserialize]{ts_tree_unserialize()}},
\code{\link[=ts_tree_update]{ts_tree_update()}},
\code{\link[=ts_tree_write]{ts_tree_write()}}
}
\concept{ts_tree generics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as.raw.ts_tree}
\alias{as.raw.ts_tree}
\title{Raw bytes of a document of a tree-sitter tree}
\usage{
\method{as.raw}{ts_tree}(x)
}
\arguments{
\item{x}{A \code{ts_tree} object.}
}
\value{
A raw vector containing the bytes of the document of the tree.
}
\description{
Raw bytes of a document of a tree-sitter tree
}
\examples{
\dontshow{if (requireNamespace("tsjsonc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Create a parse tree with tsjsonc -------------------------------------
tree <- tsjsonc::ts_parse_jsonc('{"foo": 42, "bar": [1, 2, 3]}')

tree
as.raw(tree)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=as.character.ts_tree]{as.character.ts_tree()}} to get the document as a character scalar.
}
